@@ -1,9 +1,9 @@
 /* src/cacaoh/headers.c - functions for header generation
 
-   Copyright (C) 1996-2005 R. Grafl, A. Krall, C. Kruegel, C. Oates,
-   R. Obermaisser, M. Platter, M. Probst, S. Ring, E. Steiner,
-   C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich, J. Wenninger,
-   Institut f. Computersprachen - TU Wien
+   Copyright (C) 1996-2005, 2006 R. Grafl, A. Krall, C. Kruegel,
+   C. Oates, R. Obermaisser, M. Platter, M. Probst, S. Ring,
+   E. Steiner, C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich,
+   J. Wenninger, Institut f. Computersprachen - TU Wien
 
    This file is part of CACAO.
 
@@ -19,31 +19,34 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-   02111-1307, USA.
+   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
+   02110-1301, USA.
 
-   Contact: cacao@complang.tuwien.ac.at
+   Contact: cacao@cacaojvm.org
 
    Authors: Reinhard Grafl
 
    Changes: Mark Probst
             Philipp Tomsich
             Christian Thalinger
+			Edwin Steiner
 
-   $Id: headers.c,v 3.107 2005/11/22 23:40:30 twisti Exp $
+   $Id: headers.c 4938 2006-05-22 09:06:44Z twisti $
 
 */
 
 
+#include "config.h"
+
 #include <assert.h>
 #include <ctype.h>
 #include <stdarg.h>
 #include <stdlib.h>
 #include <string.h>
 
-#include "config.h"
+#include "vm/types.h"
 
-#if defined(USE_THREADS) && defined(NATIVE_THREADS)
+#if defined(ENABLE_THREADS)
 # if defined(__DARWIN__)
 #  include <signal.h>
 # endif
@@ -60,7 +63,6 @@
 #include "vm/class.h"
 #include "vm/global.h"
 #include "vm/method.h"
-#include "vm/tables.h"
 #include "vm/loader.h"
 #include "vm/options.h"
 #include "vm/stringlocal.h"
@@ -70,13 +72,13 @@
 #if defined(ENABLE_INTRP)
 /* dummy interpreter stack to keep the compiler happy */
 
-u1 intrp_main_stack[1];
+u1 *intrp_main_stack;
 #endif
 
 
 /* for raising exceptions from native methods *********************************/
 
-#if !defined(USE_THREADS) || !defined(NATIVE_THREADS)
+#if !defined(ENABLE_THREADS)
 java_objectheader *_no_threads_exceptionptr = NULL;
 #endif
 
@@ -97,21 +99,17 @@ java_objectheader *native_new_and_init_int(classinfo *c, s4 i) { return NULL; }
 java_objectheader *native_new_and_init_throwable(classinfo *c, java_lang_Throwable *t) { return NULL; }
 
 
-java_objectheader *asm_calljavafunction(methodinfo *m, void *arg1, void *arg2,
-										void *arg3, void *arg4) { return NULL; }
+java_objectheader *vm_call_method(methodinfo *m, java_objectheader *o, ...)
+{ return NULL; }
 
-void asm_call_jit_compiler(void) {}
+void vm_abort(const char *text, ...)
+{
+	abort();
+}
 
 /* code patching functions */
 void patcher_builtin_arraycheckcast(u1 *sp) {}
 
-s4 asm_builtin_f2i(float a) { return 0; }
-s8 asm_builtin_f2l(float a) { return 0; }
-s4 asm_builtin_d2i(double a) { return 0; }
-s8 asm_builtin_d2l(double a) { return 0; }
-
-void use_class_as_object(void) {}
-
 #if defined(__MIPS__)
 long compare_and_swap(long *p, long oldval, long newval)
 {
@@ -125,7 +123,11 @@ long compare_and_swap(long *p, long oldval, long newval)
 
 
 u1 *createcompilerstub(methodinfo *m) { return NULL; }
-u1 *codegen_createnativestub(functionptr f, methodinfo *m) { return NULL; }
+#if defined(ENABLE_INTRP)
+u1 *intrp_createcompilerstub(methodinfo *m) { return NULL; }
+#endif
+
+codeinfo *codegen_createnativestub(functionptr f, methodinfo *m) { return NULL; }
 
 void removecompilerstub(u1 *stub) {}
 void removenativestub(u1 *stub) {}
@@ -136,7 +138,20 @@ u1* asm_initialize_thread_stack(void *func, u1 *stack) { return NULL; }
 void *asm_switchstackandcall(void *stack, void *func, void **stacktopsave, void * p) { return NULL; }
 
 void asm_handle_builtin_exception(classinfo *c) {}
+
+
+#if defined(ENABLE_JIT)
 void asm_getclassvalues_atomic(vftbl_t *super, vftbl_t *sub, castinfo *out) {}
+#endif
+
+#if defined(ENABLE_INTRP)
+void intrp_asm_getclassvalues_atomic(vftbl_t *super, vftbl_t *sub, castinfo *out) {}
+#endif
+
+u8 asm_get_cycle_count(void)
+{
+	return 0;
+}
 
 
 void *Java_java_lang_VMObject_clone(void *env, void *clazz, void * this)
@@ -163,6 +178,8 @@ void typeinfo_print(FILE *file,typeinfo *info,int indent) {}
 void exceptions_print_exception(java_objectheader *xptr) {}
 void stacktrace_print_trace(java_objectheader *xptr) {}
 
+void code_free_code_of_method(methodinfo *m) {}
+
 
 /* exception functions ********************************************************/
 
@@ -239,7 +256,7 @@ java_objectheader *new_exception_message(const char *classname, const char *mess
 java_objectheader *new_exception_utfmessage(const char *classname, utf *message)
 {
 	fprintf(stderr, "%s: ", classname);
-	utf_display(message);
+	utf_display_printable_ascii(message);
 	fputc('\n', stderr);
 
 	exit(1);
@@ -269,7 +286,7 @@ java_objectheader *new_classformaterror(classinfo *c, const char *message, ...)
 {
 	va_list ap;
 
-	utf_display(c->name);
+	utf_display_printable_ascii(c->name);
 	fprintf(stderr, ": ");
 
 	va_start(ap, message);
@@ -286,10 +303,20 @@ java_objectheader *new_classformaterror(classinfo *c, const char *message, ...)
 }
 
 
+void exceptions_throw_classformaterror(classinfo *c, const char *message, ...)
+{
+	va_list ap;
+
+	va_start(ap, message);
+	(void) new_classformaterror(c, message, ap);
+	va_end(ap);
+}
+
+
 java_objectheader *new_classnotfoundexception(utf *name)
 {
 	fprintf(stderr, "java.lang.ClassNotFoundException: ");
-	utf_fprint(stderr, name);
+	utf_fprint_printable_ascii(stderr, name);
 	fputc('\n', stderr);
 
 	exit(1);
@@ -303,7 +330,7 @@ java_objectheader *new_classnotfoundexception(utf *name)
 java_objectheader *new_noclassdeffounderror(utf *name)
 {
 	fprintf(stderr, "java.lang.NoClassDefFoundError: ");
-	utf_fprint(stderr, name);
+	utf_fprint_printable_ascii(stderr, name);
 	fputc('\n', stderr);
 
 	exit(1);
@@ -314,14 +341,30 @@ java_objectheader *new_noclassdeffounderror(utf *name)
 }
 
 
+java_objectheader *exceptions_new_linkageerror(const char *message,
+											   classinfo *c)
+{
+	fprintf(stderr, "java.lang.LinkageError: %s",message);
+	if (c) {
+		utf_fprint_printable_ascii_classname(stderr, c->name);
+	}
+	fputc('\n', stderr);
+
+	exit(1);
+
+	/* keep compiler happy */
+
+	return NULL;
+}
+
 java_objectheader *exceptions_new_nosuchmethoderror(classinfo *c,
 													utf *name, utf *desc)
 {
 	fprintf(stderr, "java.lang.NoSuchMethodError: ");
-	utf_fprint(stderr, c->name);
+	utf_fprint_printable_ascii(stderr, c->name);
 	fprintf(stderr, ".");
-	utf_fprint(stderr, name);
-	utf_fprint(stderr, desc);
+	utf_fprint_printable_ascii(stderr, name);
+	utf_fprint_printable_ascii(stderr, desc);
 	fputc('\n', stderr);
 
 	exit(1);
@@ -356,7 +399,7 @@ java_objectheader *new_unsupportedclassversionerror(classinfo *c, const char *me
 
 	fprintf(stderr, "%s: ", string_java_lang_UnsupportedClassVersionError);
 
-	utf_display(c->name);
+	utf_display_printable_ascii(c->name);
 	fprintf(stderr, ": ");
 
 	va_start(ap, message);
@@ -393,6 +436,12 @@ java_objectheader *new_negativearraysizeexception(void)
 }
 
 
+void exceptions_throw_negativearraysizeexception(void)
+{
+	(void) new_negativearraysizeexception();
+}
+
+
 java_objectheader *new_nullpointerexception(void)
 {
 	fprintf(stderr, "%s", string_java_lang_NullPointerException);
@@ -403,24 +452,34 @@ java_objectheader *new_nullpointerexception(void)
 	return NULL;
 }
 
+
+void exceptions_throw_nullpointerexception(void)
+{
+	(void) new_nullpointerexception();
+}
+
+
 void classnotfoundexception_to_noclassdeffounderror(void)
 {
 }
 
 /* machine dependent stuff ****************************************************/
 
-#if defined(USE_THREADS) && defined(NATIVE_THREADS)
-threadcritnode asm_criticalsections;
+#if defined(ENABLE_THREADS)
+critical_section_node_t asm_criticalsections;
 void thread_restartcriticalsection(ucontext_t *uc) {}
 #endif
 
 void md_param_alloc(methoddesc *md) {}
 
 
+#if defined(ENABLE_INTRP)
+void print_dynamic_super_statistics(void) {}
+#endif
+
+
 /************************ global variables **********************/
 
-chain *nativemethod_chain;              /* chain with native methods          */
-chain *nativeclass_chain;               /* chain with processed classes       */
 chain *ident_chain;     /* chain with method and field names in current class */
 FILE *file = NULL;
 static u4 outputsize;
@@ -444,7 +503,7 @@ void printID(utf *u)
 	char *utf_ptr = u->text;
 	int i;
 
-	for (i = 0; i < utf_strlen(u); i++) 
+	for (i = 0; i < utf_get_number_of_u2s(u); i++) 
 		printIDpart(utf_nextu2(&utf_ptr));
 }
 
@@ -603,7 +662,7 @@ static void printfields(classinfo *c)
 			fprintf(file, "   ");
 			printtype(f->descriptor->text);
 			fprintf(file, " ");
-			utf_fprint(file, f->name);
+			utf_fprint_printable_ascii(file, f->name);
 
 			/* rename multiple fieldnames */
 			if ((ident_count = searchidentchain_utf(f->name)))
@@ -629,11 +688,11 @@ void printmethod(methodinfo *m)
 
 	/* create remarks */
 	fprintf(file, "\n/*\n * Class:     ");
-	utf_fprint(file, m->class->name);
+	utf_fprint_printable_ascii(file, m->class->name);
 	fprintf(file, "\n * Method:    ");
-	utf_fprint(file, m->name);
+	utf_fprint_printable_ascii(file, m->name);
 	fprintf(file, "\n * Signature: ");
-	utf_fprint(file, m->descriptor);
+	utf_fprint_printable_ascii(file, m->descriptor);
 	fprintf(file, "\n */\n");
 
 	/* create prototype */ 			
@@ -646,7 +705,14 @@ void printmethod(methodinfo *m)
 
 	fprintf(file, "_");
 	printID(m->name);
-	if (m->nativelyoverloaded) printOverloadPart(m->descriptor);
+
+	/* ATTENTION: We use the methodinfo's isleafmethod variable as
+	   nativelyoverloaded, so we can save some space during
+	   runtime. */
+
+	if (m->isleafmethod)
+		printOverloadPart(m->descriptor);
+
 	fprintf(file, "(JNIEnv *env");
 	
 	utf_ptr = m->descriptor->text + 1;
@@ -678,7 +744,7 @@ void gen_header_filename(char *buffer, utf *u)
 {
 	s4 i;
   
-	for (i = 0; i < utf_strlen(u); i++) {
+	for (i = 0; i < utf_get_number_of_u2s(u); i++) {
 		if ((u->text[i] == '/') || (u->text[i] == '$')) {
 			buffer[i] = '_';  /* convert '$' and '/' to '_' */
 
@@ -686,7 +752,7 @@ void gen_header_filename(char *buffer, utf *u)
 			buffer[i] = u->text[i];
 		}
 	}
-	buffer[utf_strlen(u)] = '\0';
+	buffer[utf_get_number_of_u2s(u)] = '\0';
 }
 
 
@@ -699,13 +765,14 @@ void headerfile_generate(classinfo *c, char *opt_directory)
 	char uclassname[1024];
 	u2 i;
 	methodinfo *m;	      		
-	u2 i2;
+	u2 j;
 	methodinfo *m2;
-	u2 nativelyoverloaded;	      		
-		      
-	/* store class in chain */		      
-	chain_addlast(nativeclass_chain, c);
-			    	
+	bool nativelyoverloaded;
+
+	/* prevent compiler warnings */
+
+	nativelyoverloaded = false;
+
 	/* open headerfile for class */
 	gen_header_filename(classname, c->name);
 
@@ -737,7 +804,7 @@ void headerfile_generate(classinfo *c, char *opt_directory)
 
 	/* create structure for direct access to objects */	
 	fprintf(file, "/* Structure information for class: ");
-	utf_fprint(file, c->name);
+	utf_fprint_printable_ascii(file, c->name);
 	fprintf(file, " */\n\n");
 	fprintf(file, "typedef struct ");
 	printID(c->name);							
@@ -758,34 +825,41 @@ void headerfile_generate(classinfo *c, char *opt_directory)
 	/* create method-prototypes */
 		      		
 	/* find overloaded methods */
-	for (i = 0; i < c->methodscount; i++) {
 
+	for (i = 0; i < c->methodscount; i++) {
 		m = &(c->methods[i]);
 
-		if (!(m->flags & ACC_NATIVE)) continue;
-		if (!m->nativelyoverloaded) {
-			nativelyoverloaded=false;
-			for (i2=i+1;i2<c->methodscount; i2++) {
-				m2 = &(c->methods[i2]);
-				if (!(m2->flags & ACC_NATIVE)) continue;
-				if (m->name==m2->name) {
-					m2->nativelyoverloaded=true;
-					nativelyoverloaded=true;
+		if (!(m->flags & ACC_NATIVE))
+			continue;
+
+		/* We use the methodinfo's isleafmethod variable as
+		   nativelyoverloaded, so we can save some space during
+		   runtime. */
+
+		if (!m->isleafmethod) {
+			nativelyoverloaded = false;
+
+			for (j = i + 1; j < c->methodscount; j++) {
+				m2 = &(c->methods[j]);
+
+				if (!(m2->flags & ACC_NATIVE))
+					continue;
+
+				if (m->name == m2->name) {
+					m2->isleafmethod = true;
+					nativelyoverloaded = true;
 				}
 			}
-			m->nativelyoverloaded=nativelyoverloaded;
 		}
 
+		m->isleafmethod = nativelyoverloaded;
 	}
 
 	for (i = 0; i < c->methodscount; i++) {
-
 		m = &(c->methods[i]);
 
-		if (m->flags & ACC_NATIVE) {
-			chain_addlast(nativemethod_chain, m);
+		if (m->flags & ACC_NATIVE)
 			printmethod(m);
-		}
 	}
 
 	chain_free(ident_chain);
@@ -825,4 +899,5 @@ void print_classname(classinfo *clazz)
  * c-basic-offset: 4
  * tab-width: 4
  * End:
+ * vim:noexpandtab:sw=4:ts=4:
  */
