@@ -1,10 +1,10 @@
 /* src/native/jvmti.c - implementation of the Java Virtual Machine Tool 
                         Interface functions
 
-   Copyright (C) 1996-2005 R. Grafl, A. Krall, C. Kruegel, C. Oates,
-   R. Obermaisser, M. Platter, M. Probst, S. Ring, E. Steiner,
-   C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich, J. Wenninger,
-   Institut f. Computersprachen - TU Wien
+   Copyright (C) 1996-2005, 2006 R. Grafl, A. Krall, C. Kruegel,
+   C. Oates, R. Obermaisser, M. Platter, M. Probst, S. Ring,
+   E. Steiner, C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich,
+   J. Wenninger, Institut f. Computersprachen - TU Wien
 
    This file is part of CACAO.
 
@@ -20,34 +20,39 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-   02111-1307, USA.
+   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
+   02110-1301, USA.
 
-   Contact: cacao@complang.tuwien.ac.at
+   Contact: cacao@cacaojvm.org
 
    Author: Martin Platter
 
    Changes:             
 
    
-   $Id: jvmti.h,v 1.3 2005/11/04 16:58:36 motse Exp $
+   $Id: jvmti.h 4944 2006-05-23 15:31:19Z motse $
 
 */
-#ifndef JVMTI_H
-#define JVMTI_H
+#ifndef _JVMTI_H
+#define _JVMTI_H
 
 #include "native/jni.h"
-#include "native/include/java_lang_String.h"
 #include <sys/types.h>
 
+
+#define JVMTI_VERSION_INTERFACE_JNI   0x00000000
+#define JVMTI_VERSION_INTERFACE_JVMTI 0x30000000
+#define JVMTI_VERSION_MASK_INTERFACE_TYPE 0x70000000
 #define JVMTI_VERSION_1_0 0x30010000
+#define JVMTI_VERSION     JVMTI_VERSION_1_0 
+
 
 typedef jobject jthread;
 typedef jobject jthreadGroup;
 typedef jlong jlocation;
 struct _jrawMonitorID;
 typedef struct _jrawMonitorID *jrawMonitorID;
-typedef struct jvmtiEnv_struct jvmtiEnv;
+typedef struct jvmtiEnv_struct *jvmtiEnv;
 
 typedef enum {
     JVMTI_ERROR_NONE = 0, /* No error has occurred. This is the error code that is 
@@ -1177,48 +1182,6 @@ struct jvmtiEnv_struct {
 #define JVMTI_CLASS_STATUS_ARRAY  16  
 #define JVMTI_CLASS_STATUS_PRIMITIVE  32 
 
-/* cacao specific */
-
-typedef struct {
-	jvmtiEnv *jvmti_env;
-	jthread thread;
-	jmethodID method;
-	jlocation location;
-	jclass klass;
-	jobject object;
-	jfieldID field;
-	char signature_type;
-	jvalue value;
-	jboolean b;
-	void* address;
-	void** new_address_ptr;
-	jmethodID catch_method;
-	jlocation catch_location;
-	char* name;
-	jobject protection_domain;
-	jint jint1;
-	jint jint2;
-	unsigned char* class_data;
-	jint* new_class_data_len;
-	unsigned char** new_class_data;
-	jvmtiAddrLocationMap* map;
-	void* compile_info;
-	jlong jlong;
-} genericEventData;
-
-#if defined(USE_THREADS) && defined(NATIVE_THREADS)
-struct _jrawMonitorID {
-	java_lang_String *name;
-};
-#endif
-
-jvmtiEnv* new_jvmtienv();
-void set_jvmti_phase(jvmtiPhase p);
-pid_t debuggee;
-jvmtiEnv* remotedbgjvmtienv;
-jvmtiEventCallbacks jvmti_jdwp_EventCallbacks;
-void agentload(char* opt_arg);
-void agentunload();
 #endif
 
 
