@@ -1,9 +1,9 @@
 /* src/native/vm/Field.c - java/lang/reflect/Field
 
-   Copyright (C) 1996-2005 R. Grafl, A. Krall, C. Kruegel, C. Oates,
-   R. Obermaisser, M. Platter, M. Probst, S. Ring, E. Steiner,
-   C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich, J. Wenninger,
-   Institut f. Computersprachen - TU Wien
+   Copyright (C) 1996-2005, 2006 R. Grafl, A. Krall, C. Kruegel,
+   C. Oates, R. Obermaisser, M. Platter, M. Probst, S. Ring,
+   E. Steiner, C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich,
+   J. Wenninger, Institut f. Computersprachen - TU Wien
 
    This file is part of CACAO.
 
@@ -19,24 +19,25 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-   02111-1307, USA.
+   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
+   02110-1301, USA.
 
-   Contact: cacao@complang.tuwien.ac.at
+   Contact: cacao@cacaojvm.org
 
    Authors: Roman Obermaiser
 
    Changes: Joseph Wenninger
             Christian Thalinger
 
-   $Id: Field.c,v 3.37 2005/11/15 23:43:17 twisti Exp $
+   $Id: Field.c 4807 2006-04-21 13:08:00Z edwin $
 
 */
 
 
+#include "config.h"
+
 #include <assert.h>
 
-#include "config.h"
 #include "vm/types.h"
 
 #include "native/jni.h"
@@ -60,7 +61,6 @@
 #include "vm/loader.h"
 #include "vm/resolve.h"
 #include "vm/stringlocal.h"
-#include "vm/tables.h"
 #include "vm/utf8.h"
 #include "vm/jit/stacktrace.h"
 
@@ -70,7 +70,15 @@
 
 /* cacao_get_field_address *****************************************************
 
-   XXX
+   Return the address of a field of an object.
+
+   IN:
+      this.........the field (a java.lang.reflect.Field object)
+	  o............the object of which to get the field
+
+   RETURN VALUE:
+      a pointer to the field, or
+	  NULL if an exception has been thrown
 
 *******************************************************************************/
 
@@ -93,7 +101,9 @@ static void *cacao_get_field_address(java_lang_reflect_Field *this,
 		if (this->flag == false) {
 			/* get the calling class */
 
-			oa = cacao_createClassContextArray();
+			oa = stacktrace_getClassContext();
+			if (!oa)
+				return NULL;
 
 			/* this function is always called like this:
 
@@ -116,8 +126,9 @@ static void *cacao_get_field_address(java_lang_reflect_Field *this,
 	if (f->flags & ACC_STATIC) {
 		/* initialize class if required */
 
-		if (!initialize_class(c))
-			return NULL;
+		if (!(c->state & CLASS_INITIALIZED))
+			if (!initialize_class(c))
+				return NULL;
 
 		/* return value address */
 
@@ -127,7 +138,7 @@ static void *cacao_get_field_address(java_lang_reflect_Field *this,
 		/* obj is required for not-static fields */
 
 		if (o == NULL) {
-			*exceptionptr = new_nullpointerexception();
+			exceptions_throw_nullpointerexception();
 			return NULL;
 		}
 	
@@ -137,7 +148,7 @@ static void *cacao_get_field_address(java_lang_reflect_Field *this,
 
 	/* exception path */
 
-	*exceptionptr = new_illegalargumentexception();
+	exceptions_throw_illegalargumentexception();
 
 	return NULL;
 }
@@ -298,7 +309,7 @@ JNIEXPORT s4 JNICALL Java_java_lang_reflect_Field_getBoolean(JNIEnv *env, java_l
 	case PRIMITIVETYPE_BOOLEAN:
 		return (s4) *((s4 *) addr);
 	default:
-		*exceptionptr = new_illegalargumentexception();
+		exceptions_throw_illegalargumentexception();
 		return 0;
 	}
 }
@@ -331,7 +342,7 @@ JNIEXPORT s4 JNICALL Java_java_lang_reflect_Field_getByte(JNIEnv *env, java_lang
 	case PRIMITIVETYPE_BYTE:
 		return (s4) *((s4 *) addr);
 	default:
-		*exceptionptr = new_illegalargumentexception();
+		exceptions_throw_illegalargumentexception();
 		return 0;
 	}
 }
@@ -364,7 +375,7 @@ JNIEXPORT s4 JNICALL Java_java_lang_reflect_Field_getChar(JNIEnv *env, java_lang
 	case PRIMITIVETYPE_CHAR:
 		return (s4) *((s4 *) addr);
 	default:
-		*exceptionptr = new_illegalargumentexception();
+		exceptions_throw_illegalargumentexception();
 		return 0;
 	}
 }
@@ -398,7 +409,7 @@ JNIEXPORT s4 JNICALL Java_java_lang_reflect_Field_getShort(JNIEnv *env, java_lan
 	case PRIMITIVETYPE_SHORT:
 		return (s4) *((s4 *) addr);
 	default:
-		*exceptionptr = new_illegalargumentexception();
+		exceptions_throw_illegalargumentexception();
 		return 0;
 	}
 }
@@ -434,7 +445,7 @@ JNIEXPORT s4 JNICALL Java_java_lang_reflect_Field_getInt(JNIEnv *env , java_lang
 	case PRIMITIVETYPE_INT:
 		return (s4) *((s4 *) addr);
 	default:
-		*exceptionptr = new_illegalargumentexception();
+		exceptions_throw_illegalargumentexception();
 		return 0;
 	}
 }
@@ -472,7 +483,7 @@ JNIEXPORT s8 JNICALL Java_java_lang_reflect_Field_getLong(JNIEnv *env, java_lang
 	case PRIMITIVETYPE_LONG:
 		return (s8) *((s8 *) addr);
 	default:
-		*exceptionptr = new_illegalargumentexception();
+		exceptions_throw_illegalargumentexception();
 		return 0;
 	}
 }
@@ -512,7 +523,7 @@ JNIEXPORT float JNICALL Java_java_lang_reflect_Field_getFloat(JNIEnv *env, java_
 	case PRIMITIVETYPE_FLOAT:
 		return (float) *((float *) addr);
 	default:
-		*exceptionptr = new_illegalargumentexception();
+		exceptions_throw_illegalargumentexception();
 		return 0;
 	}
 }
@@ -554,7 +565,7 @@ JNIEXPORT double JNICALL Java_java_lang_reflect_Field_getDouble(JNIEnv *env , ja
 	case PRIMITIVETYPE_DOUBLE:
 		return (double) *((double *) addr);
 	default:
-		*exceptionptr = new_illegalargumentexception();
+		exceptions_throw_illegalargumentexception();
 		return 0;
 	}
 }
@@ -608,7 +619,7 @@ JNIEXPORT void JNICALL Java_java_lang_reflect_Field_set(JNIEnv *env, java_lang_r
 			val = ((java_lang_Boolean *) value)->value;
 			break;
 		default:
-			*exceptionptr = new_illegalargumentexception();
+			exceptions_throw_illegalargumentexception();
 			return;
 		}
 
@@ -627,7 +638,7 @@ JNIEXPORT void JNICALL Java_java_lang_reflect_Field_set(JNIEnv *env, java_lang_r
 			val = ((java_lang_Byte *) value)->value;
 			break;
 		default:	
-			*exceptionptr = new_illegalargumentexception();
+			exceptions_throw_illegalargumentexception();
 			return;
 		}
 
@@ -646,7 +657,7 @@ JNIEXPORT void JNICALL Java_java_lang_reflect_Field_set(JNIEnv *env, java_lang_r
 			val = ((java_lang_Character *) value)->value;
 			break;
 		default:
-			*exceptionptr = new_illegalargumentexception();
+			exceptions_throw_illegalargumentexception();
 			return;
 		}
 
@@ -670,7 +681,7 @@ JNIEXPORT void JNICALL Java_java_lang_reflect_Field_set(JNIEnv *env, java_lang_r
 			val = ((java_lang_Short *) value)->value;
 			break;
 		default:
-			*exceptionptr = new_illegalargumentexception();
+			exceptions_throw_illegalargumentexception();
 			return;
 		}
 
@@ -700,7 +711,7 @@ JNIEXPORT void JNICALL Java_java_lang_reflect_Field_set(JNIEnv *env, java_lang_r
 			val = ((java_lang_Integer *) value)->value;
 			break;
 		default:
-			*exceptionptr = new_illegalargumentexception();
+			exceptions_throw_illegalargumentexception();
 			return;
 		}
 
@@ -733,7 +744,7 @@ JNIEXPORT void JNICALL Java_java_lang_reflect_Field_set(JNIEnv *env, java_lang_r
 			val = ((java_lang_Long *) value)->value;
 			break;
 		default:
-			*exceptionptr = new_illegalargumentexception();
+			exceptions_throw_illegalargumentexception();
 			return;
 		}
 
@@ -769,7 +780,7 @@ JNIEXPORT void JNICALL Java_java_lang_reflect_Field_set(JNIEnv *env, java_lang_r
 			val = ((java_lang_Float *) value)->value;
 			break;
 		default:
-			*exceptionptr = new_illegalargumentexception();
+			exceptions_throw_illegalargumentexception();
 			return;
 		}
 
@@ -808,7 +819,7 @@ JNIEXPORT void JNICALL Java_java_lang_reflect_Field_set(JNIEnv *env, java_lang_r
 			val = ((java_lang_Double *) value)->value;
 			break;
 		default:
-			*exceptionptr = new_illegalargumentexception();
+			exceptions_throw_illegalargumentexception();
 			return;
 		}
 
@@ -829,7 +840,7 @@ JNIEXPORT void JNICALL Java_java_lang_reflect_Field_set(JNIEnv *env, java_lang_r
 
 	/* raise exception */
 
-	*exceptionptr = new_illegalargumentexception();
+	exceptions_throw_illegalargumentexception();
 }
 
 
@@ -861,7 +872,7 @@ JNIEXPORT void JNICALL Java_java_lang_reflect_Field_setBoolean(JNIEnv *env, java
 		*((s4 *) addr) = value;
 		break;
 	default:
-		*exceptionptr = new_illegalargumentexception();
+		exceptions_throw_illegalargumentexception();
 	}
 
 	return;
@@ -907,7 +918,7 @@ JNIEXPORT void JNICALL Java_java_lang_reflect_Field_setByte(JNIEnv *env, java_la
 		*((double *) addr) = value;
 		break;
 	default:
-		*exceptionptr = new_illegalargumentexception();
+		exceptions_throw_illegalargumentexception();
 	}
 
 	return;
@@ -952,7 +963,7 @@ JNIEXPORT void JNICALL Java_java_lang_reflect_Field_setChar(JNIEnv *env, java_la
 		*((double *) addr) = value;
 		break;
 	default:
-		*exceptionptr = new_illegalargumentexception();
+		exceptions_throw_illegalargumentexception();
 	}
 
 	return;
@@ -997,7 +1008,7 @@ JNIEXPORT void JNICALL Java_java_lang_reflect_Field_setShort(JNIEnv *env, java_l
 		*((double *) addr) = value;
 		break;
 	default:
-		*exceptionptr = new_illegalargumentexception();
+		exceptions_throw_illegalargumentexception();
 	}
 
 	return;
@@ -1041,7 +1052,7 @@ JNIEXPORT void JNICALL Java_java_lang_reflect_Field_setInt(JNIEnv *env, java_lan
 		*((double *) addr) = value;
 		break;
 	default:
-		*exceptionptr = new_illegalargumentexception();
+		exceptions_throw_illegalargumentexception();
 	}
 
 	return;
@@ -1082,7 +1093,7 @@ JNIEXPORT void JNICALL Java_java_lang_reflect_Field_setLong(JNIEnv *env, java_la
 		*((double *) addr) = value;
 		break;
 	default:
-		*exceptionptr = new_illegalargumentexception();
+		exceptions_throw_illegalargumentexception();
 	}
 
 	return;
@@ -1120,7 +1131,7 @@ JNIEXPORT void JNICALL Java_java_lang_reflect_Field_setFloat(JNIEnv *env, java_l
 		*((double *) addr) = value;
 		break;
 	default:
-		*exceptionptr = new_illegalargumentexception();
+		exceptions_throw_illegalargumentexception();
 	}
 
 	return;
@@ -1155,7 +1166,7 @@ JNIEXPORT void JNICALL Java_java_lang_reflect_Field_setDouble(JNIEnv *env, java_
 		*((double *) addr) = value;
 		break;
 	default:
-		*exceptionptr = new_illegalargumentexception();
+		exceptions_throw_illegalargumentexception();
 	}
 
 	return;
@@ -1182,25 +1193,24 @@ JNIEXPORT java_lang_Class* JNICALL Java_java_lang_reflect_Field_getType(JNIEnv *
 	if (!resolve_class_from_typedesc(desc, true, false, &ret))
 		return NULL;
 	
-	if (!use_class_as_object(ret))
-		return NULL;
-
 	return (java_lang_Class *) ret;
 }
 
 
 /*
- * Class:     java_lang_reflect_Field
- * Method:    getModifiers
+ * Class:     java/lang/reflect/Field
+ * Method:    getModifiersInternal
  * Signature: ()I
  */
-JNIEXPORT s4 JNICALL Java_java_lang_reflect_Field_getModifiers(JNIEnv *env, java_lang_reflect_Field *this)
+JNIEXPORT s4 JNICALL Java_java_lang_reflect_Field_getModifiersInternal(JNIEnv *env, java_lang_reflect_Field *this)
 {
 	classinfo *c;
+	fieldinfo *f;
 
 	c = (classinfo *) this->declaringClass;
+	f = &(c->fields[this->slot]);
 
-	return c->fields[this->slot].flags;
+	return f->flags;
 }
 
 
