@@ -1,9 +1,9 @@
 /* src/vm/builtin.c - functions for unsupported operations
 
-   Copyright (C) 1996-2005 R. Grafl, A. Krall, C. Kruegel, C. Oates,
-   R. Obermaisser, M. Platter, M. Probst, S. Ring, E. Steiner,
-   C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich, J. Wenninger,
-   Institut f. Computersprachen - TU Wien
+   Copyright (C) 1996-2005, 2006 R. Grafl, A. Krall, C. Kruegel,
+   C. Oates, R. Obermaisser, M. Platter, M. Probst, S. Ring,
+   E. Steiner, C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich,
+   J. Wenninger, Institut f. Computersprachen - TU Wien
 
    This file is part of CACAO.
 
@@ -19,16 +19,17 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-   02111-1307, USA.
+   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
+   02110-1301, USA.
 
-   Contact: cacao@complang.tuwien.ac.at
+   Contact: cacao@cacaojvm.org
 
    Authors: Reinhard Grafl
             Andreas Krall
             Mark Probst
 
    Changes: Christian Thalinger
+            Edwin Steiner
 
    Contains C functions for JavaVM Instructions that cannot be
    translated to machine language directly. Consequently, the
@@ -36,16 +37,18 @@
    calls instead of machine instructions, using the C calling
    convention.
 
-   $Id: builtin.c,v 3.148 2005/11/11 11:17:33 twisti Exp $
+   $Id: builtin.c 4960 2006-05-26 12:19:43Z edwin $
 
 */
 
 
+#include "config.h"
+
 #include <assert.h>
+#include <math.h>
 #include <stdlib.h>
 #include <string.h>
 
-#include "config.h"
 #include "vm/types.h"
 
 #include "arch.h"
@@ -60,27 +63,24 @@
 #include "native/include/java_lang_Object.h"          /* required by VMObject */
 #include "native/include/java_lang_VMObject.h"
 
-#if defined(USE_THREADS)
-# if defined(NATIVE_THREADS)
-#  include "threads/native/threads.h"
-# else
-#  include "threads/green/threads.h"
-#  include "threads/green/locks.h"
-# endif
+#if defined(ENABLE_THREADS)
+# include "threads/native/threads.h"
 #endif
 
 #include "toolbox/logging.h"
 #include "toolbox/util.h"
 #include "vm/builtin.h"
+#include "vm/class.h"
 #include "vm/exceptions.h"
 #include "vm/global.h"
 #include "vm/initialize.h"
 #include "vm/loader.h"
 #include "vm/options.h"
 #include "vm/stringlocal.h"
-#include "vm/tables.h"
 #include "vm/jit/asmpart.h"
 #include "vm/jit/patcher.h"
+#include "vm/rt-timing.h"
+#include "vm/cycles-stats.h"
 
 
 /* include builtin tables *****************************************************/
@@ -88,6 +88,11 @@
 #include "vm/builtintable.inc"
 
 
+CYCLES_STATS_DECLARE(builtin_monitorenter,100,2)
+CYCLES_STATS_DECLARE(builtin_monitorexit ,100,2)
+CYCLES_STATS_DECLARE(builtin_new         ,100,5)
+CYCLES_STATS_DECLARE(builtin_overhead    , 80,1)
+
 /* builtintable_init ***********************************************************
 
    Parse the descriptors of builtin functions and create the parsed
@@ -240,7 +245,7 @@ static void builtintable_sort_automatic(void)
 
 /* builtin_init ****************************************************************
 
-   XXX
+   Initialize the global table of builtin functions.
 
 *******************************************************************************/
 
@@ -343,7 +348,7 @@ s4 builtin_isanysubclass(classinfo *sub, classinfo *super)
 			(sub->vftbl->interfacetable[-super->index] != NULL);
 
 	} else {
-		asm_getclassvalues_atomic(super->vftbl, sub->vftbl, &classvalues);
+		ASM_GETCLASSVALUES_ATOMIC(super->vftbl, sub->vftbl, &classvalues);
 
 		res = (u4) (classvalues.sub_baseval - classvalues.super_baseval) <=
 			(u4) classvalues.super_diffval;
@@ -362,7 +367,7 @@ s4 builtin_isanysubclass_vftbl(vftbl_t *sub, vftbl_t *super)
 	if (sub == super)
 		return 1;
 
-	asm_getclassvalues_atomic(super, sub, &classvalues);
+	ASM_GETCLASSVALUES_ATOMIC(super, sub, &classvalues);
 
 	if ((base = classvalues.super_baseval) <= 0) {
 		/* super is an interface */
@@ -497,12 +502,16 @@ s4 builtin_arrayinstanceof(java_objectheader *o, classinfo *targetclass)
 }
 
 
-/************************** exception functions *******************************
+/* builtin_throw_exception *****************************************************
 
-******************************************************************************/
+   Sets the exceptionptr with the thrown exception and prints some
+   debugging information.  Called from asm_vm_call_method.
 
-java_objectheader *builtin_throw_exception(java_objectheader *xptr)
+*******************************************************************************/
+
+void *builtin_throw_exception(java_objectheader *xptr)
 {
+#if !defined(NDEBUG)
     java_lang_Throwable *t;
 	char                *logtext;
 	s4                   logtextlen;
@@ -517,12 +526,17 @@ java_objectheader *builtin_throw_exception(java_objectheader *xptr)
 
 		if (t) {
 			logtextlen +=
-				utf_strlen(xptr->vftbl->class->name) +
-				strlen(": ") +
-				javastring_strlen((java_objectheader *) t->detailMessage);
-
-		} else
+				utf_bytes(xptr->vftbl->class->name);
+			if (t->detailMessage) {
+				logtextlen += strlen(": ") +
+					u2_utflength(t->detailMessage->value->data 
+									+ t->detailMessage->offset,
+						     	 t->detailMessage->count);
+			}
+		} 
+		else {
 			logtextlen += strlen("(nil)");
+		}
 
 		/* allocate memory */
 
@@ -533,8 +547,7 @@ java_objectheader *builtin_throw_exception(java_objectheader *xptr)
 		strcpy(logtext, "Builtin exception thrown: ");
 
 		if (t) {
-			utf_sprint_classname(logtext + strlen(logtext),
-								 xptr->vftbl->class->name);
+			utf_cat_classname(logtext, xptr->vftbl->class->name);
 
 			if (t->detailMessage) {
 				char *buf;
@@ -542,7 +555,7 @@ java_objectheader *builtin_throw_exception(java_objectheader *xptr)
 				buf = javastring_tochar((java_objectheader *) t->detailMessage);
 				strcat(logtext, ": ");
 				strcat(logtext, buf);
-				MFREE(buf, char, strlen(buf));
+				MFREE(buf, char, strlen(buf) + 1);
 			}
 
 		} else {
@@ -555,12 +568,17 @@ java_objectheader *builtin_throw_exception(java_objectheader *xptr)
 
 		dump_release(dumpsize);
 	}
+#endif /* !defined(NDEBUG) */
+
+	/* actually set the exception */
 
 	*exceptionptr = xptr;
 
-	return xptr;
-}
+	/* Return a NULL pointer.  This is required for vm_call_method to
+	   check for an exception.  This is for convenience. */
 
+	return NULL;
+}
 
 
 /* builtin_canstore ************************************************************
@@ -580,14 +598,14 @@ s4 builtin_canstore(java_objectarray *oa, java_objectheader *o)
 	vftbl_t         *valuevftbl;
 	s4               base;
 	castinfo         classvalues;
-	
+
 	if (!o)
 		return 1;
 
 	/* The following is guaranteed (by verifier checks):
 	 *
-	 *     *) a->...vftbl->arraydesc != NULL
-	 *     *) a->...vftbl->arraydesc->componentvftbl != NULL
+	 *     *) oa->...vftbl->arraydesc != NULL
+	 *     *) oa->...vftbl->arraydesc->componentvftbl != NULL
 	 *     *) o->vftbl is not an interface vftbl
 	 */
 	
@@ -598,13 +616,13 @@ s4 builtin_canstore(java_objectarray *oa, java_objectheader *o)
 	if ((desc->dimension - 1) == 0) {
 		s4 res;
 
-		/* {a is a one-dimensional array} */
-		/* {a is an array of references} */
+		/* {oa is a one-dimensional array} */
+		/* {oa is an array of references} */
 		
 		if (valuevftbl == componentvftbl)
 			return 1;
 
-		asm_getclassvalues_atomic(componentvftbl, valuevftbl, &classvalues);
+		ASM_GETCLASSVALUES_ATOMIC(componentvftbl, valuevftbl, &classvalues);
 
 		if ((base = classvalues.super_baseval) <= 0)
 			/* an array of interface references */
@@ -617,7 +635,7 @@ s4 builtin_canstore(java_objectarray *oa, java_objectheader *o)
 		return res;
 	}
 
-	/* {a has dimension > 1} */
+	/* {oa has dimension > 1} */
 	/* {componentvftbl->arraydesc != NULL} */
 
 	/* check if o is an array */
@@ -660,7 +678,7 @@ s4 builtin_canstore_onedim (java_objectarray *a, java_objectheader *o)
 	if (valuevftbl == elementvftbl)
 		return 1;
 
-	asm_getclassvalues_atomic(elementvftbl, valuevftbl, &classvalues);
+	ASM_GETCLASSVALUES_ATOMIC(elementvftbl, valuevftbl, &classvalues);
 
 	if ((base = classvalues.super_baseval) <= 0)
 		/* an array of interface references */
@@ -702,7 +720,7 @@ s4 builtin_canstore_onedim_class(java_objectarray *a, java_objectheader *o)
 	if (valuevftbl == elementvftbl)
 		return 1;
 
-	asm_getclassvalues_atomic(elementvftbl, valuevftbl, &classvalues);
+	ASM_GETCLASSVALUES_ATOMIC(elementvftbl, valuevftbl, &classvalues);
 
 	res = (unsigned) (classvalues.sub_baseval - classvalues.super_baseval)
 		<= (unsigned) classvalues.super_diffval;
@@ -723,10 +741,19 @@ s4 builtin_canstore_onedim_class(java_objectarray *a, java_objectheader *o)
 java_objectheader *builtin_new(classinfo *c)
 {
 	java_objectheader *o;
+#if defined(ENABLE_RT_TIMING)
+	struct timespec time_start, time_end;
+#endif
+#if defined(ENABLE_CYCLES_STATS)
+	u8 cycles_start, cycles_end;
+#endif
+
+	RT_TIMING_GET_TIME(time_start);
+	CYCLES_STATS_GET(cycles_start);
 
 	/* is the class loaded */
 
-	assert(c->loaded);
+	assert(c->state & CLASS_LOADED);
 
 	/* check if we can instantiate this class */
 
@@ -738,29 +765,39 @@ java_objectheader *builtin_new(classinfo *c)
 	}
 
 	/* is the class linked */
-	if (!c->linked)
+
+	if (!(c->state & CLASS_LINKED))
 		if (!link_class(c))
 			return NULL;
 
-	if (!c->initialized) {
+	if (!(c->state & CLASS_INITIALIZED)) {
+#if !defined(NDEBUG)
 		if (initverbose)
 			log_message_class("Initialize class (from builtin_new): ", c);
+#endif
 
 		if (!initialize_class(c))
 			return NULL;
 	}
 
-	o = heap_allocate(c->instancesize, true, c->finalizer);
+	o = heap_allocate(c->instancesize, c->flags & ACC_CLASS_HAS_POINTERS,
+					  c->finalizer);
 
 	if (!o)
 		return NULL;
 
 	o->vftbl = c->vftbl;
 
-#if defined(USE_THREADS) && defined(NATIVE_THREADS)
-	initObjectLock(o);
+#if defined(ENABLE_THREADS)
+	lock_init_object_lock(o);
 #endif
 
+	CYCLES_STATS_GET(cycles_end);
+	RT_TIMING_GET_TIME(time_end);
+
+	CYCLES_STATS_COUNT(builtin_new,cycles_end - cycles_start);
+	RT_TIMING_TIME_DIFF(time_start, time_end, RT_TIMING_NEW_OBJECT);
+
 	return o;
 }
 
@@ -781,13 +818,18 @@ java_arrayheader *builtin_newarray(s4 size, classinfo *arrayclass)
 	s4                componentsize;
 	s4                actualsize;
 	java_arrayheader *a;
+#if defined(ENABLE_RT_TIMING)
+	struct timespec time_start, time_end;
+#endif
+
+	RT_TIMING_GET_TIME(time_start);
 
 	desc          = arrayclass->vftbl->arraydesc;
 	dataoffset    = desc->dataoffset;
 	componentsize = desc->componentsize;
 
 	if (size < 0) {
-		*exceptionptr = new_negativearraysizeexception();
+		exceptions_throw_negativearraysizeexception();
 		return NULL;
 	}
 
@@ -805,12 +847,15 @@ java_arrayheader *builtin_newarray(s4 size, classinfo *arrayclass)
 
 	a->objheader.vftbl = arrayclass->vftbl;
 
-#if defined(USE_THREADS) && defined(NATIVE_THREADS)
-	initObjectLock(&a->objheader);
+#if defined(ENABLE_THREADS)
+	lock_init_object_lock(&a->objheader);
 #endif
 
 	a->size = size;
 
+	RT_TIMING_GET_TIME(time_end);
+	RT_TIMING_TIME_DIFF(time_start, time_end, RT_TIMING_NEW_ARRAY);
+
 	return a;
 }
 
@@ -830,11 +875,11 @@ java_objectarray *builtin_anewarray(s4 size, classinfo *componentclass)
 	
 	/* is class loaded */
 
-	assert(componentclass->loaded);
+	assert(componentclass->state & CLASS_LOADED);
 
 	/* is class linked */
 
-	if (!componentclass->linked)
+	if (!(componentclass->state & CLASS_LINKED))
 		if (!link_class(componentclass))
 			return NULL;
 
@@ -1067,7 +1112,7 @@ java_arrayheader *builtin_multianewarray(int n, classinfo *arrayclass,
 #endif
 
 		if (size < 0) {
-			*exceptionptr = new_negativearraysizeexception();
+			exceptions_throw_negativearraysizeexception();
 			return NULL;
 		}
 	}
@@ -1086,7 +1131,9 @@ java_arrayheader *builtin_multianewarray(int n, classinfo *arrayclass,
 	
 *****************************************************************************/
 
-s4 methodindent = 0;
+#if !defined(NDEBUG)
+static s4 methodindent = 0;
+static u4 callcount = 0;
 
 java_objectheader *builtin_trace_exception(java_objectheader *xptr,
 										   methodinfo *m,
@@ -1096,139 +1143,140 @@ java_objectheader *builtin_trace_exception(java_objectheader *xptr,
 	char *logtext;
 	s4    logtextlen;
 	s4    dumpsize;
+	codeinfo *code;
 
-	if (opt_verbose || runverbose || opt_verboseexception) {
-		/* when running with verbosecall we remove one indent level */
+	if (opt_verbosecall && indent)
+		methodindent--;
 
-		if (runverbose && indent)
-			methodindent--;
+	/* calculate message length */
 
-		/* calculate message length */
+	if (xptr) {
+		logtextlen =
+			strlen("Exception ") + utf_bytes(xptr->vftbl->class->name);
+	} 
+	else {
+		logtextlen = strlen("Some Throwable");
+	}
 
-		if (xptr) {
-			logtextlen =
-				strlen("Exception ") +
-				utf_strlen(xptr->vftbl->class->name);
+	logtextlen += strlen(" thrown in ");
 
-		} else {
-			logtextlen = strlen("Some Throwable");
-		}
-
-		logtextlen += strlen(" thrown in ");
-
-		if (m) {
-			logtextlen +=
-				utf_strlen(m->class->name) +
-				strlen(".") +
-				utf_strlen(m->name) +
-				utf_strlen(m->descriptor) +
-				strlen("(NOSYNC,NATIVE");
+	if (m) {
+		logtextlen +=
+			utf_bytes(m->class->name) +
+			strlen(".") +
+			utf_bytes(m->name) +
+			utf_bytes(m->descriptor) +
+			strlen("(NOSYNC,NATIVE");
 
 #if SIZEOF_VOID_P == 8
-			logtextlen +=
-				strlen(")(0x123456789abcdef0) at position 0x123456789abcdef0 (");
+		logtextlen +=
+			strlen(")(0x123456789abcdef0) at position 0x123456789abcdef0 (");
 #else
-			logtextlen += strlen(")(0x12345678) at position 0x12345678 (");
+		logtextlen += strlen(")(0x12345678) at position 0x12345678 (");
+#endif
+
+		if (m->class->sourcefile == NULL)
+			logtextlen += strlen("<NO CLASSFILE INFORMATION>");
+		else
+			logtextlen += utf_bytes(m->class->sourcefile);
+
+		logtextlen += strlen(":65536)");
+
+	} 
+	else {
+		logtextlen += strlen("call_java_method");
+	}
+
+	logtextlen += strlen("0");
+
+	/* allocate memory */
+
+	dumpsize = dump_size();
+
+	logtext = DMNEW(char, logtextlen);
+
+	if (xptr) {
+		strcpy(logtext, "Exception ");
+		utf_cat_classname(logtext, xptr->vftbl->class->name);
+
+	} else {
+		strcpy(logtext, "Some Throwable");
+	}
+
+	strcat(logtext, " thrown in ");
+
+	if (m) {
+		utf_cat_classname(logtext, m->class->name);
+		strcat(logtext, ".");
+		utf_cat(logtext, m->name);
+		utf_cat(logtext, m->descriptor);
+
+		if (m->flags & ACC_SYNCHRONIZED)
+			strcat(logtext, "(SYNC");
+		else
+			strcat(logtext, "(NOSYNC");
+
+		if (m->flags & ACC_NATIVE) {
+			strcat(logtext, ",NATIVE");
+
+			code = m->code;
+
+#if SIZEOF_VOID_P == 8
+			sprintf(logtext + strlen(logtext),
+					")(0x%016lx) at position 0x%016lx",
+					(ptrint) code->entrypoint, (ptrint) pos);
+#else
+			sprintf(logtext + strlen(logtext),
+					")(0x%08x) at position 0x%08x",
+					(ptrint) code->entrypoint, (ptrint) pos);
+#endif
+
+		} else {
+
+			/* XXX preliminary: This should get the actual codeinfo */
+			/* in which the exception happened.                     */
+			code = m->code;
+			
+#if SIZEOF_VOID_P == 8
+			sprintf(logtext + strlen(logtext),
+					")(0x%016lx) at position 0x%016lx (",
+					(ptrint) code->entrypoint, (ptrint) pos);
+#else
+			sprintf(logtext + strlen(logtext),
+					")(0x%08x) at position 0x%08x (",
+					(ptrint) code->entrypoint, (ptrint) pos);
 #endif
 
 			if (m->class->sourcefile == NULL)
-				logtextlen += strlen("<NO CLASSFILE INFORMATION>");
+				strcat(logtext, "<NO CLASSFILE INFORMATION>");
 			else
-				logtextlen += utf_strlen(m->class->sourcefile);
+				utf_cat(logtext, m->class->sourcefile);
 
-			logtextlen += strlen(":65536)");
-
-		} else
-			logtextlen += strlen("call_java_method");
-
-		logtextlen += strlen("0");
-
-		/* allocate memory */
-
-		dumpsize = dump_size();
-
-		logtext = DMNEW(char, logtextlen);
-
-		if (xptr) {
-			strcpy(logtext, "Exception ");
-			utf_strcat_classname(logtext, xptr->vftbl->class->name);
-
-		} else {
-			strcpy(logtext, "Some Throwable");
+			sprintf(logtext + strlen(logtext), ":%d)", 0);
 		}
 
-		strcat(logtext, " thrown in ");
+	} else
+		strcat(logtext, "call_java_method");
 
-		if (m) {
-			utf_strcat_classname(logtext, m->class->name);
-			strcat(logtext, ".");
-			utf_strcat(logtext, m->name);
-			utf_strcat(logtext, m->descriptor);
+	log_text(logtext);
 
-			if (m->flags & ACC_SYNCHRONIZED)
-				strcat(logtext, "(SYNC");
-			else
-				strcat(logtext, "(NOSYNC");
+	/* release memory */
 
-			if (m->flags & ACC_NATIVE) {
-				strcat(logtext, ",NATIVE");
-
-#if SIZEOF_VOID_P == 8
-				sprintf(logtext + strlen(logtext),
-						")(0x%016lx) at position 0x%016lx",
-						(ptrint) m->entrypoint, (ptrint) pos);
-#else
-				sprintf(logtext + strlen(logtext),
-						")(0x%08x) at position 0x%08x",
-						(ptrint) m->entrypoint, (ptrint) pos);
-#endif
-
-			} else {
-#if SIZEOF_VOID_P == 8
-				sprintf(logtext + strlen(logtext),
-						")(0x%016lx) at position 0x%016lx (",
-						(ptrint) m->entrypoint, (ptrint) pos);
-#else
-				sprintf(logtext + strlen(logtext),
-						")(0x%08x) at position 0x%08x (",
-						(ptrint) m->entrypoint, (ptrint) pos);
-#endif
-
-				if (m->class->sourcefile == NULL)
-					strcat(logtext, "<NO CLASSFILE INFORMATION>");
-				else
-					utf_strcat(logtext, m->class->sourcefile);
-
-				sprintf(logtext + strlen(logtext), ":%d)", 0);
-			}
-
-		} else
-			strcat(logtext, "call_java_method");
-
-		log_text(logtext);
-
-		/* release memory */
-
-		dump_release(dumpsize);
-
-		/* print stacktrace for exception */
-
-		if (opt_verboseexception) {
-			exceptions_print_exception(xptr);
-			stacktrace_print_trace(xptr);
-		}
-	}
+	dump_release(dumpsize);
 
 	return xptr;
 }
+#endif /* !defined(NDEBUG) */
 
 
 /* builtin_trace_args **********************************************************
 
-   XXX
+   Print method call with arguments for -verbose:call.
 
 *******************************************************************************/
 
+#if !defined(NDEBUG)
+
 #ifdef TRACE_ARGS_NUM
 void builtin_trace_args(s8 a0, s8 a1,
 #if TRACE_ARGS_NUM >= 4
@@ -1254,11 +1302,14 @@ void builtin_trace_args(s8 a0, s8 a1,
 	/* calculate message length */
 
 	logtextlen =
-		6 + methodindent + strlen("called: ") +
-		utf_strlen(m->class->name) +
+		strlen("4294967295 ") +
+		strlen("-2147483647-") +        /* INT_MAX should be sufficient       */
+		methodindent +
+		strlen("called: ") +
+		utf_bytes(m->class->name) +
 		strlen(".") +
-		utf_strlen(m->name) +
-		utf_strlen(m->descriptor) +
+		utf_bytes(m->name) +
+		utf_bytes(m->descriptor) +
 		strlen(" SYNCHRONIZED") + strlen("(") + strlen(")");
 
 	/* add maximal argument length */
@@ -1271,7 +1322,11 @@ void builtin_trace_args(s8 a0, s8 a1,
 
 	logtext = DMNEW(char, logtextlen);
 
-	sprintf(logtext,"-%d-",methodindent);
+	callcount++;
+
+	sprintf(logtext, "%10d ", callcount);
+	sprintf(logtext + strlen(logtext), "-%d-", methodindent);
+
 	pos = strlen(logtext);
 
 	for (i = 0; i < methodindent; i++)
@@ -1279,10 +1334,10 @@ void builtin_trace_args(s8 a0, s8 a1,
 
 	strcpy(logtext + pos, "called: ");
 
-	utf_strcat_classname(logtext, m->class->name);
+	utf_cat_classname(logtext, m->class->name);
 	strcat(logtext, ".");
-	utf_strcat(logtext, m->name);
-	utf_strcat(logtext, m->descriptor);
+	utf_cat(logtext, m->name);
+	utf_cat(logtext, m->descriptor);
 
 	if (m->flags & ACC_PUBLIC)       strcat(logtext, " PUBLIC");
 	if (m->flags & ACC_PRIVATE)      strcat(logtext, " PRIVATE");
@@ -1477,34 +1532,44 @@ void builtin_trace_args(s8 a0, s8 a1,
 	methodindent++;
 }
 #endif
+#endif /* !defined(NDEBUG) */
 
 
 /* builtin_displaymethodstop ***************************************************
 
-   XXX
+   Print method exit for -verbose:call.
 
 *******************************************************************************/
 
+#if !defined(NDEBUG)
 void builtin_displaymethodstop(methodinfo *m, s8 l, double d, float f)
 {
-	methoddesc *md;
-	char       *logtext;
-	s4          logtextlen;
-	s4          dumpsize;
-	s4          i;
-	s4          pos;
-	imm_union   imu;
+	methoddesc  	  *md;
+	char        	  *logtext;
+	s4          	   logtextlen;
+	s4          	   dumpsize;
+	s4          	   i;
+	s4                 pos;
+	java_objectheader *o;
+	java_lang_String  *s;
+	classinfo         *c;
+	s4                 len;
+	utf               *u;
+	imm_union          imu;
 
 	md = m->parseddesc;
 
 	/* calculate message length */
 
 	logtextlen =
-		6 + methodindent + strlen("finished: ") +
-		utf_strlen(m->class->name) +
+		strlen("4294967295 ") +
+		strlen("-2147483647-") +        /* INT_MAX should be sufficient       */
+		methodindent +
+		strlen("finished: ") +
+		utf_bytes(m->class->name) +
 		strlen(".") +
-		utf_strlen(m->name) +
-		utf_strlen(m->descriptor) +
+		utf_bytes(m->name) +
+		utf_bytes(m->descriptor) +
 		strlen(" SYNCHRONIZED") + strlen("(") + strlen(")");
 
 	/* add maximal argument length */
@@ -1526,17 +1591,19 @@ void builtin_displaymethodstop(methodinfo *m, s8 l, double d, float f)
 
 	/* generate the message */
 
-	sprintf(logtext,"-%d-",methodindent);
+	sprintf(logtext, "           ");
+	sprintf(logtext + strlen(logtext), "-%d-", methodindent);
+
 	pos = strlen(logtext);
 
 	for (i = 0; i < methodindent; i++)
 		logtext[pos++] = '\t';
 
 	strcpy(logtext + pos, "finished: ");
-	utf_strcat_classname(logtext, m->class->name);
+	utf_cat_classname(logtext, m->class->name);
 	strcat(logtext, ".");
-	utf_strcat(logtext, m->name);
-	utf_strcat(logtext, m->descriptor);
+	utf_cat(logtext, m->name);
+	utf_cat(logtext, m->descriptor);
 
 	switch (md->returntype.type) {
 	case TYPE_INT:
@@ -1553,6 +1620,62 @@ void builtin_displaymethodstop(methodinfo *m, s8 l, double d, float f)
 
 	case TYPE_ADR:
 		sprintf(logtext + strlen(logtext), "->%p", (void *) (ptrint) l);
+
+		/* check return argument for java.lang.Class or java.lang.String */
+
+		o = (java_objectheader *) (ptrint) l;
+
+		if (o != NULL) {
+			if (o->vftbl->class == class_java_lang_String) {
+				/* get java.lang.String object and the length of the
+                   string */
+
+				s = (java_lang_String *) (ptrint) l;
+
+				u = javastring_toutf(s, false);
+
+				len = strlen(", String = \"") + utf_bytes(u) + strlen("\"");
+
+				/* realloc memory for string length */
+
+				DMREALLOC(logtext, char, logtextlen, logtextlen + len);
+
+				/* convert to utf8 string and strcat it to the logtext */
+
+				strcat(logtext, ", String = \"");
+				utf_cat(logtext, u);
+				strcat(logtext, "\"");
+
+			} else {
+				if (o->vftbl->class == class_java_lang_Class) {
+					/* if the object returned is a java.lang.Class
+					   cast it to classinfo structure and get the name
+					   of the class */
+
+					c = (classinfo *) (ptrint) l;
+
+					u = c->name;
+
+				} else {
+					/* if the object returned is not a java.lang.String or
+					   a java.lang.Class just print the name of the class */
+
+					u = o->vftbl->class->name;
+				}
+
+				len = strlen(", Class = \"") + utf_bytes(u) + strlen("\"");
+
+				/* realloc memory for string length */
+
+				DMREALLOC(logtext, char, logtextlen, logtextlen + len);
+
+				/* strcat to the logtext */
+
+				strcat(logtext, ", Class = \"");
+				utf_cat_classname(logtext, u);
+				strcat(logtext, "\"");
+			}
+		}
 		break;
 
 	case TYPE_FLT:
@@ -1576,158 +1699,66 @@ void builtin_displaymethodstop(methodinfo *m, s8 l, double d, float f)
 
 	dump_release(dumpsize);
 }
+#endif /* !defined(NDEBUG) */
 
 
-/****************************************************************************
-			 SYNCHRONIZATION FUNCTIONS
-*****************************************************************************/
-
-#if defined(USE_THREADS) && !defined(NATIVE_THREADS)
-/*
- * Lock the mutex of an object.
- */
-void internal_lock_mutex_for_object(java_objectheader *object)
+#if defined(ENABLE_CYCLES_STATS)
+void builtin_print_cycles_stats(FILE *file)
 {
-	mutexHashEntry *entry;
-	int hashValue;
+	fprintf(file,"builtin cylce count statistics:\n");
 
-	assert(object != 0);
+	CYCLES_STATS_PRINT_OVERHEAD(builtin_overhead,file);
+	CYCLES_STATS_PRINT(builtin_monitorenter,file);
+	CYCLES_STATS_PRINT(builtin_monitorexit ,file);
+	CYCLES_STATS_PRINT(builtin_new         ,file);
 
-	hashValue = MUTEX_HASH_VALUE(object);
-	entry = &mutexHashTable[hashValue];
-
-	if (entry->object != 0) {
-		if (entry->mutex.count == 0 && entry->conditionCount == 0) {
-			entry->object = 0;
-			entry->mutex.holder = 0;
-			entry->mutex.count = 0;
-			entry->mutex.muxWaiters = 0;
-
-		} else {
-			while (entry->next != 0 && entry->object != object)
-				entry = entry->next;
-
-			if (entry->object != object) {
-				entry->next = firstFreeOverflowEntry;
-				firstFreeOverflowEntry = firstFreeOverflowEntry->next;
-
-				entry = entry->next;
-				entry->object = 0;
-				entry->next = 0;
-				assert(entry->conditionCount == 0);
-			}
-		}
-
-	} else {
-		entry->mutex.holder = 0;
-		entry->mutex.count = 0;
-		entry->mutex.muxWaiters = 0;
-	}
-
-	if (entry->object == 0)
-		entry->object = object;
-	
-	internal_lock_mutex(&entry->mutex);
+	fprintf(file,"\n");
 }
-#endif
+#endif /* defined(ENABLE_CYCLES_STATS) */
 
-
-#if defined(USE_THREADS) && !defined(NATIVE_THREADS)
-/*
- * Unlocks the mutex of an object.
- */
-void internal_unlock_mutex_for_object (java_objectheader *object)
-{
-	int hashValue;
-	mutexHashEntry *entry;
-
-	hashValue = MUTEX_HASH_VALUE(object);
-	entry = &mutexHashTable[hashValue];
-
-	if (entry->object == object) {
-		internal_unlock_mutex(&entry->mutex);
-
-	} else {
-		while (entry->next != 0 && entry->next->object != object)
-			entry = entry->next;
-
-		assert(entry->next != 0);
-
-		internal_unlock_mutex(&entry->next->mutex);
-
-		if (entry->next->mutex.count == 0 && entry->conditionCount == 0) {
-			mutexHashEntry *unlinked = entry->next;
-
-			entry->next = unlinked->next;
-			unlinked->next = firstFreeOverflowEntry;
-			firstFreeOverflowEntry = unlinked;
-		}
-	}
-}
-#endif
-
-
-#if defined(USE_THREADS)
+#if defined(ENABLE_THREADS)
 void builtin_monitorenter(java_objectheader *o)
 {
-#if !defined(NATIVE_THREADS)
-	int hashValue;
-
-	++blockInts;
-
-	hashValue = MUTEX_HASH_VALUE(o);
-	if (mutexHashTable[hashValue].object == o 
-		&& mutexHashTable[hashValue].mutex.holder == currentThread)
-		++mutexHashTable[hashValue].mutex.count;
-	else
-		internal_lock_mutex_for_object(o);
-
-	--blockInts;
-#else
-	monitorEnter((threadobject *) THREADOBJECT, o);
+#if defined(ENABLE_CYCLES_STATS)
+	u8 cycles_start, cycles_overhead, cycles_end;
 #endif
+
+	CYCLES_STATS_GET(cycles_start);
+	CYCLES_STATS_GET(cycles_overhead);
+
+	lock_monitor_enter((threadobject *) THREADOBJECT, o);
+
+	CYCLES_STATS_GET(cycles_end);
+	CYCLES_STATS_COUNT(builtin_monitorenter, cycles_end - cycles_overhead);
+	CYCLES_STATS_COUNT(builtin_overhead    , cycles_overhead - cycles_start);
 }
 #endif
 
 
-#if defined(USE_THREADS)
+#if defined(ENABLE_THREADS)
 /*
  * Locks the class object - needed for static synchronized methods.
- * The use_class_as_object call is needed in order to circumvent a
- * possible deadlock with builtin_monitorenter called by another
- * thread calling use_class_as_object.
  */
 void builtin_staticmonitorenter(classinfo *c)
 {
-	use_class_as_object(c);
-	builtin_monitorenter(&c->header);
+	builtin_monitorenter(&c->object.header);
 }
 #endif
 
 
-#if defined(USE_THREADS)
+#if defined(ENABLE_THREADS)
 void builtin_monitorexit(java_objectheader *o)
 {
-#if !defined(NATIVE_THREADS)
-	int hashValue;
-
-	++blockInts;
-
-	hashValue = MUTEX_HASH_VALUE(o);
-	if (mutexHashTable[hashValue].object == o) {
-		if (mutexHashTable[hashValue].mutex.count == 1
-			&& mutexHashTable[hashValue].mutex.muxWaiters != 0)
-			internal_unlock_mutex_for_object(o);
-		else
-			--mutexHashTable[hashValue].mutex.count;
-
-	} else
-		internal_unlock_mutex_for_object(o);
-
-	--blockInts;
-#else
-	monitorExit((threadobject *) THREADOBJECT, o);
+#if defined(ENABLE_CYCLES_STATS)
+	u8 cycles_start, cycles_end;
 #endif
+
+	CYCLES_STATS_GET(cycles_start);
+
+	lock_monitor_exit((threadobject *) THREADOBJECT, o);
+
+	CYCLES_STATS_GET(cycles_end);
+	CYCLES_STATS_COUNT(builtin_monitorexit, cycles_end - cycles_start);
 }
 #endif
 
@@ -1745,17 +1776,35 @@ void builtin_monitorexit(java_objectheader *o)
 
 ******************************************************************************/
 
-s4 builtin_idiv(s4 a, s4 b) { return a / b; }
-s4 builtin_irem(s4 a, s4 b) { return a % b; }
+#if !SUPPORT_DIVISION
+s4 builtin_idiv(s4 a, s4 b)
+{
+	s4 c;
+
+	c = a / b;
+
+	return c;
+}
+
+s4 builtin_irem(s4 a, s4 b)
+{
+	s4 c;
+
+	c = a % b;
+
+	return c;
+}
+#endif /* !SUPPORT_DIVISION */
 
 
-/************** Functions for long arithmetics *******************************
+/* functions for long arithmetics **********************************************
 
-	On systems where 64 bit Integers are not supported by the CPU, these
-	functions are needed.
+   On systems where 64 bit Integers are not supported by the CPU,
+   these functions are needed.
 
 ******************************************************************************/
 
+#if !(SUPPORT_LONG && SUPPORT_LONG_ADD)
 s8 builtin_ladd(s8 a, s8 b)
 {
 	s8 c;
@@ -1769,7 +1818,7 @@ s8 builtin_ladd(s8 a, s8 b)
 	return c;
 }
 
-s8 builtin_lsub(s8 a, s8 b) 
+s8 builtin_lsub(s8 a, s8 b)
 {
 	s8 c;
 
@@ -1782,7 +1831,23 @@ s8 builtin_lsub(s8 a, s8 b)
 	return c;
 }
 
-s8 builtin_lmul(s8 a, s8 b) 
+s8 builtin_lneg(s8 a)
+{
+	s8 c;
+
+#if U8_AVAILABLE
+	c = -a;
+#else
+	c = builtin_i2l(0);
+#endif
+
+	return c;
+}
+#endif /* !(SUPPORT_LONG && SUPPORT_LONG_ADD) */
+
+
+#if !(SUPPORT_LONG && SUPPORT_LONG_MUL)
+s8 builtin_lmul(s8 a, s8 b)
 {
 	s8 c;
 
@@ -1794,8 +1859,11 @@ s8 builtin_lmul(s8 a, s8 b)
 
 	return c;
 }
+#endif /* !(SUPPORT_LONG && SUPPORT_LONG_MUL) */
 
-s8 builtin_ldiv(s8 a, s8 b) 
+
+#if !(SUPPORT_DIVISION && SUPPORT_LONG && SUPPORT_LONG_DIV)
+s8 builtin_ldiv(s8 a, s8 b)
 {
 	s8 c;
 
@@ -1808,7 +1876,7 @@ s8 builtin_ldiv(s8 a, s8 b)
 	return c;
 }
 
-s8 builtin_lrem(s8 a, s8 b) 
+s8 builtin_lrem(s8 a, s8 b)
 {
 	s8 c;
 
@@ -1820,8 +1888,11 @@ s8 builtin_lrem(s8 a, s8 b)
 
 	return c;
 }
+#endif /* !(SUPPORT_DIVISION && SUPPORT_LONG && SUPPORT_LONG_DIV) */
 
-s8 builtin_lshl(s8 a, s4 b) 
+
+#if !(SUPPORT_LONG && SUPPORT_LONG_SHIFT)
+s8 builtin_lshl(s8 a, s4 b)
 {
 	s8 c;
 
@@ -1834,7 +1905,7 @@ s8 builtin_lshl(s8 a, s4 b)
 	return c;
 }
 
-s8 builtin_lshr(s8 a, s4 b) 
+s8 builtin_lshr(s8 a, s4 b)
 {
 	s8 c;
 
@@ -1847,7 +1918,7 @@ s8 builtin_lshr(s8 a, s4 b)
 	return c;
 }
 
-s8 builtin_lushr(s8 a, s4 b) 
+s8 builtin_lushr(s8 a, s4 b)
 {
 	s8 c;
 
@@ -1859,8 +1930,11 @@ s8 builtin_lushr(s8 a, s4 b)
 
 	return c;
 }
+#endif /* !(SUPPORT_LONG && SUPPORT_LONG_SHIFT) */
 
-s8 builtin_land(s8 a, s8 b) 
+
+#if !(SUPPORT_LONG && SUPPORT_LONG_LOGICAL)
+s8 builtin_land(s8 a, s8 b)
 {
 	s8 c;
 
@@ -1873,7 +1947,7 @@ s8 builtin_land(s8 a, s8 b)
 	return c;
 }
 
-s8 builtin_lor(s8 a, s8 b) 
+s8 builtin_lor(s8 a, s8 b)
 {
 	s8 c;
 
@@ -1898,36 +1972,28 @@ s8 builtin_lxor(s8 a, s8 b)
 
 	return c;
 }
+#endif /* !(SUPPORT_LONG && SUPPORT_LONG_LOGICAL) */
 
-s8 builtin_lneg(s8 a) 
-{
-	s8 c;
 
-#if U8_AVAILABLE
-	c = -a;
-#else
-	c = builtin_i2l(0);
-#endif
-
-	return c;
-}
-
-s4 builtin_lcmp(s8 a, s8 b) 
+#if !(SUPPORT_LONG && SUPPORT_LONG_CMP)
+s4 builtin_lcmp(s8 a, s8 b)
 { 
 #if U8_AVAILABLE
-	if (a < b) return -1;
-	if (a > b) return 1;
+	if (a < b)
+		return -1;
+
+	if (a > b)
+		return 1;
+
 	return 0;
 #else
 	return 0;
 #endif
 }
+#endif /* !(SUPPORT_LONG && SUPPORT_LONG_CMP) */
 
 
-
-
-
-/*********** Functions for floating point operations *************************/
+/* functions for unsupported floating instructions ****************************/
 
 /* used to convert FLT_xxx defines into float values */
 
@@ -1951,6 +2017,7 @@ static inline float longBitsToDouble(s8 l)
 }
 
 
+#if !SUPPORT_FLOAT
 float builtin_fadd(float a, float b)
 {
 	if (isnanf(a)) return intBitsToFloat(FLT_NAN);
@@ -2064,12 +2131,6 @@ float builtin_fdiv(float a, float b)
 }
 
 
-float builtin_frem(float a, float b)
-{
-	return fmodf(a, b);
-}
-
-
 float builtin_fneg(float a)
 {
 	if (isnanf(a)) return a;
@@ -2078,8 +2139,10 @@ float builtin_fneg(float a)
 		else return copysignf(a, -copysignf(1.0, a));
 	}
 }
+#endif /* !SUPPORT_FLOAT */
 
 
+#if !SUPPORT_FLOAT || defined(ENABLE_INTRP)
 s4 builtin_fcmpl(float a, float b)
 {
 	if (isnanf(a)) return -1;
@@ -2106,11 +2169,18 @@ s4 builtin_fcmpg(float a, float b)
 	if (a == b) return 0;
 	return -1;
 }
+#endif /* !SUPPORT_FLOAT || defined(ENABLE_INTRP) */
 
 
+float builtin_frem(float a, float b)
+{
+	return fmodf(a, b);
+}
 
-/************************* Functions for doubles ****************************/
 
+/* functions for unsupported double instructions ******************************/
+
+#if !SUPPORT_DOUBLE
 double builtin_dadd(double a, double b)
 {
 	if (isnan(a)) return longBitsToDouble(DBL_NAN);
@@ -2219,11 +2289,6 @@ double builtin_ddiv(double a, double b)
 }
 
 
-double builtin_drem(double a, double b)
-{
-	return fmod(a, b);
-}
-
 /* builtin_dneg ****************************************************************
 
    Implemented as described in VM Spec.
@@ -2253,8 +2318,10 @@ double builtin_dneg(double a)
 		}
 	}
 }
+#endif /* !SUPPORT_DOUBLE */
 
 
+#if !SUPPORT_DOUBLE || defined(ENABLE_INTRP)
 s4 builtin_dcmpl(double a, double b)
 {
 	if (isnan(a)) return -1;
@@ -2281,10 +2348,18 @@ s4 builtin_dcmpg(double a, double b)
 	if (a == b) return 0;
 	return -1;
 }
+#endif /* !SUPPORT_DOUBLE || defined(ENABLE_INTRP) */
 
 
-/*********************** Conversion operations ****************************/
+double builtin_drem(double a, double b)
+{
+	return fmod(a, b);
+}
 
+
+/* conversion operations ******************************************************/
+
+#if 0
 s8 builtin_i2l(s4 i)
 {
 #if U8_AVAILABLE
@@ -2297,21 +2372,6 @@ s8 builtin_i2l(s4 i)
 #endif
 }
 
-
-float builtin_i2f(s4 a)
-{
-	float f = (float) a;
-	return f;
-}
-
-
-double builtin_i2d(s4 a)
-{
-	double d = (double) a;
-	return d;
-}
-
-
 s4 builtin_l2i(s8 l)
 {
 #if U8_AVAILABLE
@@ -2320,8 +2380,28 @@ s4 builtin_l2i(s8 l)
 	return l.low;
 #endif
 }
+#endif
 
 
+#if !(SUPPORT_FLOAT && SUPPORT_I2F)
+float builtin_i2f(s4 a)
+{
+	float f = (float) a;
+	return f;
+}
+#endif /* !(SUPPORT_FLOAT && SUPPORT_I2F) */
+
+
+#if !(SUPPORT_DOUBLE && SUPPORT_I2D)
+double builtin_i2d(s4 a)
+{
+	double d = (double) a;
+	return d;
+}
+#endif /* !(SUPPORT_DOUBLE && SUPPORT_I2D) */
+
+
+#if !(SUPPORT_LONG && SUPPORT_FLOAT && SUPPORT_L2F)
 float builtin_l2f(s8 a)
 {
 #if U8_AVAILABLE
@@ -2331,8 +2411,10 @@ float builtin_l2f(s8 a)
 	return 0.0;
 #endif
 }
+#endif /* !(SUPPORT_LONG && SUPPORT_FLOAT && SUPPORT_L2F) */
 
 
+#if !(SUPPORT_LONG && SUPPORT_DOUBLE && SUPPORT_L2D)
 double builtin_l2d(s8 a)
 {
 #if U8_AVAILABLE
@@ -2342,8 +2424,10 @@ double builtin_l2d(s8 a)
 	return 0.0;
 #endif
 }
+#endif /* !(SUPPORT_LONG && SUPPORT_DOUBLE && SUPPORT_L2D) */
 
 
+#if !(SUPPORT_FLOAT && SUPPORT_F2I) || defined(ENABLE_INTRP)
 s4 builtin_f2i(float a) 
 {
 	s4 i;
@@ -2368,8 +2452,10 @@ s4 builtin_f2i(float a)
 		return 2147483647;
 		return (-2147483648); */
 }
+#endif /* !(SUPPORT_FLOAT && SUPPORT_F2I) || defined(ENABLE_INTRP) */
 
 
+#if !(SUPPORT_FLOAT && SUPPORT_LONG && SUPPORT_F2L)
 s8 builtin_f2l(float a)
 {
 	s8 l;
@@ -2394,20 +2480,10 @@ s8 builtin_f2l(float a)
 		return 9223372036854775807L;
 		return (-9223372036854775808L); */
 }
+#endif /* !(SUPPORT_FLOAT && SUPPORT_LONG && SUPPORT_F2L) */
 
 
-double builtin_f2d(float a)
-{
-	if (finitef(a)) return (double) a;
-	else {
-		if (isnanf(a))
-			return longBitsToDouble(DBL_NAN);
-		else
-			return copysign(longBitsToDouble(DBL_POSINF), (double) copysignf(1.0, a) );
-	}
-}
-
-
+#if !(SUPPORT_DOUBLE && SUPPORT_D2I) || defined(ENABLE_INTRP)
 s4 builtin_d2i(double a) 
 { 
 	double d;
@@ -2426,8 +2502,10 @@ s4 builtin_d2i(double a)
 		return 2147483647;
 	return (-2147483647-1);
 }
+#endif /* !(SUPPORT_DOUBLE && SUPPORT_D2I) || defined(ENABLE_INTRP) */
 
 
+#if !(SUPPORT_DOUBLE && SUPPORT_LONG && SUPPORT_D2L)
 s8 builtin_d2l(double a)
 {
 	double d;
@@ -2446,8 +2524,21 @@ s8 builtin_d2l(double a)
 		return 9223372036854775807LL;
 	return (-9223372036854775807LL-1);
 }
+#endif /* !(SUPPORT_DOUBLE && SUPPORT_LONG && SUPPORT_D2L) */
 
 
+#if !(SUPPORT_FLOAT && SUPPORT_DOUBLE)
+double builtin_f2d(float a)
+{
+	if (finitef(a)) return (double) a;
+	else {
+		if (isnanf(a))
+			return longBitsToDouble(DBL_NAN);
+		else
+			return copysign(longBitsToDouble(DBL_POSINF), (double) copysignf(1.0, a) );
+	}
+}
+
 float builtin_d2f(double a)
 {
 	if (finite(a))
@@ -2459,6 +2550,7 @@ float builtin_d2f(double a)
 			return copysignf(intBitsToFloat(FLT_POSINF), (float) copysign(1.0, a));
 	}
 }
+#endif /* !(SUPPORT_FLOAT && SUPPORT_DOUBLE) */
 
 
 /* builtin_clone_array *********************************************************
@@ -2486,7 +2578,7 @@ java_arrayheader *builtin_clone_array(void *env, java_arrayheader *o)
 
 *******************************************************************************/
 
-#if defined(USE_THREADS) && defined(NATIVE_THREADS)
+#if defined(ENABLE_THREADS)
 java_objectheader **builtin_asm_get_exceptionptrptr(void)
 {
 	return builtin_get_exceptionptrptr();
@@ -2505,4 +2597,5 @@ java_objectheader **builtin_asm_get_exceptionptrptr(void)
  * c-basic-offset: 4
  * tab-width: 4
  * End:
+ * vim:noexpandtab:sw=4:ts=4:
  */
