@@ -1,9 +1,9 @@
 /* src/vm/exceptions.c - exception related functions
 
-   Copyright (C) 1996-2005 R. Grafl, A. Krall, C. Kruegel, C. Oates,
-   R. Obermaisser, M. Platter, M. Probst, S. Ring, E. Steiner,
-   C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich, J. Wenninger,
-   Institut f. Computersprachen - TU Wien
+   Copyright (C) 1996-2005, 2006 R. Grafl, A. Krall, C. Kruegel,
+   C. Oates, R. Obermaisser, M. Platter, M. Probst, S. Ring,
+   E. Steiner, C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich,
+   J. Wenninger, Institut f. Computersprachen - TU Wien
 
    This file is part of CACAO.
 
@@ -19,25 +19,28 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-   02111-1307, USA.
+   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
+   02110-1301, USA.
 
-   Contact: cacao@complang.tuwien.ac.at
+   Contact: cacao@cacaojvm.org
 
    Authors: Christian Thalinger
 
    Changes: Edwin Steiner
 
-   $Id: exceptions.c,v 1.36 2005/11/11 11:16:42 twisti Exp $
+   $Id: exceptions.c 4921 2006-05-15 14:24:36Z twisti $
 
 */
 
 
+#include "config.h"
+
+#include <assert.h>
 #include <string.h>
 #include <stdarg.h>
 #include <stdlib.h>
 
-#include "config.h"
+#include "vm/types.h"
 
 #include "mm/memory.h"
 #include "native/native.h"
@@ -51,14 +54,15 @@
 #include "vm/loader.h"
 #include "vm/options.h"
 #include "vm/stringlocal.h"
-#include "vm/tables.h"
+#include "vm/vm.h"
 #include "vm/jit/asmpart.h"
 #include "vm/jit/jit.h"
+#include "vm/jit/methodheader.h"
 
 
 /* for raising exceptions from native methods *********************************/
 
-#if !defined(USE_THREADS) || !defined(NATIVE_THREADS)
+#if !defined(ENABLE_THREADS)
 java_objectheader *_no_threads_exceptionptr = NULL;
 #endif
 
@@ -100,6 +104,13 @@ bool exceptions_init(void)
 		!link_class(class_java_lang_NoClassDefFoundError))
 		return false;
 
+	/* java/lang/LinkageError */
+
+	if (!(class_java_lang_LinkageError =
+		  load_class_bootstrap(utf_java_lang_LinkageError)) ||
+		!link_class(class_java_lang_LinkageError))
+		return false;
+
 	/* java/lang/NoSuchMethodError */
 
 	if (!(class_java_lang_NoSuchMethodError =
@@ -176,8 +187,9 @@ static void throw_exception_exit_intern(bool doexit)
 									   false);
 
 		/* print the stacktrace */
+
 		if (pss) {
-			asm_calljavafunction(pss, xptr, NULL, NULL, NULL);
+			(void) vm_call_method(pss, xptr);
 
 			/* This normally means, we are EXTREMLY out of memory or have a   */
 			/* serious problem while printStackTrace. But may be another      */
@@ -189,7 +201,7 @@ static void throw_exception_exit_intern(bool doexit)
 				t = (java_lang_Throwable *) *exceptionptr;
 
 				fprintf(stderr, "Exception while printStackTrace(): ");
-				utf_fprint_classname(stderr, t->header.vftbl->class->name);
+				utf_fprint_printable_ascii_classname(stderr, t->header.vftbl->class->name);
 
 				if (t->detailMessage) {
 					char *buf;
@@ -203,7 +215,7 @@ static void throw_exception_exit_intern(bool doexit)
 			}
 
 		} else {
-			utf_fprint_classname(stderr, c->name);
+			utf_fprint_printable_ascii_classname(stderr, c->name);
 			fprintf(stderr, ": printStackTrace()V not found!\n");
 		}
 
@@ -302,6 +314,13 @@ void exceptions_throw_outofmemory_exit(void)
 
    Creates an exception object with the given name and initalizes it.
 
+   IN:
+      classname....class name in UTF-8
+
+   RETURN VALUE:
+      an exception pointer (in any case -- either it is the newly created
+	  exception, or an exception thrown while trying to create it).
+
 *******************************************************************************/
 
 java_objectheader *new_exception(const char *classname)
@@ -326,23 +345,26 @@ java_objectheader *new_exception(const char *classname)
    Creates an exception object with the given name and initalizes it
    with the given char message.
 
+   IN:
+      classname....class name in UTF-8
+	  message......message in UTF-8
+
+   RETURN VALUE:
+      an exception pointer (in any case -- either it is the newly created
+	  exception, or an exception thrown while trying to create it).
+
 *******************************************************************************/
 
 java_objectheader *new_exception_message(const char *classname,
 										 const char *message)
 {
-	java_objectheader *o;
-	classinfo         *c;
-   
-	if (!(c = load_class_bootstrap(utf_new_char(classname))))
+	java_lang_String *s;
+
+	s = javastring_new_from_utf_string(message);
+	if (!s)
 		return *exceptionptr;
 
-	o = native_new_and_init_string(c, javastring_new_char(message));
-
-	if (!o)
-		return *exceptionptr;
-
-	return o;
+	return new_exception_javastring(classname, s);
 }
 
 
@@ -351,6 +373,14 @@ java_objectheader *new_exception_message(const char *classname,
    Creates an exception object with the given name and initalizes it
    with the given java/lang/Throwable exception.
 
+   IN:
+      classname....class name in UTF-8
+	  throwable....the given Throwable
+
+   RETURN VALUE:
+      an exception pointer (in any case -- either it is the newly created
+	  exception, or an exception thrown while trying to create it).
+
 *******************************************************************************/
 
 java_objectheader *new_exception_throwable(const char *classname,
@@ -376,22 +406,25 @@ java_objectheader *new_exception_throwable(const char *classname,
    Creates an exception object with the given name and initalizes it
    with the given utf message.
 
+   IN:
+      classname....class name in UTF-8
+	  message......the message as an utf *
+
+   RETURN VALUE:
+      an exception pointer (in any case -- either it is the newly created
+	  exception, or an exception thrown while trying to create it).
+
 *******************************************************************************/
 
 java_objectheader *new_exception_utfmessage(const char *classname, utf *message)
 {
-	java_objectheader *o;
-	classinfo         *c;
-   
-	if (!(c = load_class_bootstrap(utf_new_char(classname))))
+	java_lang_String *s;
+
+	s = javastring_new(message);
+	if (!s)
 		return *exceptionptr;
 
-	o = native_new_and_init_string(c, javastring_new(message));
-
-	if (!o)
-		return *exceptionptr;
-
-	return o;
+	return new_exception_javastring(classname, s);
 }
 
 
@@ -400,6 +433,14 @@ java_objectheader *new_exception_utfmessage(const char *classname, utf *message)
    Creates an exception object with the given name and initalizes it
    with the given java/lang/String message.
 
+   IN:
+      classname....class name in UTF-8
+	  message......the message as a java.lang.String
+
+   RETURN VALUE:
+      an exception pointer (in any case -- either it is the newly created
+	  exception, or an exception thrown while trying to create it).
+
 *******************************************************************************/
 
 java_objectheader *new_exception_javastring(const char *classname,
@@ -425,6 +466,14 @@ java_objectheader *new_exception_javastring(const char *classname,
    Creates an exception object with the given name and initalizes it
    with the given int value.
 
+   IN:
+      classname....class name in UTF-8
+	  i............the integer
+
+   RETURN VALUE:
+      an exception pointer (in any case -- either it is the newly created
+	  exception, or an exception thrown while trying to create it).
+
 *******************************************************************************/
 
 java_objectheader *new_exception_int(const char *classname, s4 i)
@@ -448,6 +497,14 @@ java_objectheader *new_exception_int(const char *classname, s4 i)
 
    generates a java.lang.ClassFormatError for the classloader
 
+   IN:
+      c............the class in which the error was found
+	  message......UTF-8 format string
+
+   RETURN VALUE:
+      an exception pointer (in any case -- either it is the newly created
+	  exception, or an exception thrown while trying to create it).
+
 *******************************************************************************/
 
 java_objectheader *new_classformaterror(classinfo *c, const char *message, ...)
@@ -462,7 +519,7 @@ java_objectheader *new_classformaterror(classinfo *c, const char *message, ...)
 	msglen = 0;
 
 	if (c)
-		msglen += utf_strlen(c->name) + strlen(" (");
+		msglen += utf_bytes(c->name) + strlen(" (");
 
 	va_start(ap, message);
 	msglen += get_variable_message_length(message, ap);
@@ -480,7 +537,7 @@ java_objectheader *new_classformaterror(classinfo *c, const char *message, ...)
 	/* print message into allocated buffer */
 
 	if (c) {
-		utf_sprint_classname(msg, c->name);
+		utf_copy_classname(msg, c->name);
 		strcat(msg, " (");
 	}
 
@@ -499,18 +556,53 @@ java_objectheader *new_classformaterror(classinfo *c, const char *message, ...)
 }
 
 
+/* exceptions_throw_classformaterror *******************************************
+
+   Generate a java.lang.ClassFormatError for the VM system and throw it.
+
+   IN:
+      c............the class in which the error was found
+	  message......UTF-8 format string
+
+   RETURN VALUE:
+      an exception pointer (in any case -- either it is the newly created
+	  exception, or an exception thrown while trying to create it).
+
+*******************************************************************************/
+
+void exceptions_throw_classformaterror(classinfo *c, const char *message, ...)
+{
+	va_list ap;
+
+	va_start(ap, message);
+	*exceptionptr = new_classformaterror(c, message, ap);
+	va_end(ap);
+}
+
+
 /* new_classnotfoundexception **************************************************
 
    Generates a java.lang.ClassNotFoundException for the classloader.
 
+   IN:
+      name.........name of the class not found as a utf *
+
+   RETURN VALUE:
+      an exception pointer (in any case -- either it is the newly created
+	  exception, or an exception thrown while trying to create it).
+
 *******************************************************************************/
 
 java_objectheader *new_classnotfoundexception(utf *name)
 {
 	java_objectheader *o;
+	java_lang_String  *s;
 
-	o = native_new_and_init_string(class_java_lang_ClassNotFoundException,
-								   javastring_new(name));
+	s = javastring_new(name);
+	if (!s)
+		return *exceptionptr;
+
+	o = native_new_and_init_string(class_java_lang_ClassNotFoundException, s);
 
 	if (!o)
 		return *exceptionptr;
@@ -523,14 +615,25 @@ java_objectheader *new_classnotfoundexception(utf *name)
 
    Generates a java.lang.NoClassDefFoundError
 
+   IN:
+      name.........name of the class not found as a utf *
+
+   RETURN VALUE:
+      an exception pointer (in any case -- either it is the newly created
+	  exception, or an exception thrown while trying to create it).
+
 *******************************************************************************/
 
 java_objectheader *new_noclassdeffounderror(utf *name)
 {
 	java_objectheader *o;
+	java_lang_String  *s;
 
-	o = native_new_and_init_string(class_java_lang_NoClassDefFoundError,
-								   javastring_new(name));
+	s = javastring_new(name);
+	if (!s)
+		return *exceptionptr;
+
+	o = native_new_and_init_string(class_java_lang_NoClassDefFoundError, s);
 
 	if (!o)
 		return *exceptionptr;
@@ -584,6 +687,13 @@ void classnotfoundexception_to_noclassdeffounderror(void)
 
    Generates a java.lang.InternalError for the VM.
 
+   IN:
+      message......UTF-8 message format string
+
+   RETURN VALUE:
+      an exception pointer (in any case -- either it is the newly created
+	  exception, or an exception thrown while trying to create it).
+
 *******************************************************************************/
 
 java_objectheader *new_internalerror(const char *message, ...)
@@ -621,10 +731,73 @@ java_objectheader *new_internalerror(const char *message, ...)
 }
 
 
+/* exceptions_new_linkageerror *************************************************
+
+   Generates a java.lang.LinkageError with an error message.
+
+   IN:
+      message......UTF-8 message
+	  c............class related to the error. If this is != NULL
+	               the name of c is appended to the error message.
+
+   RETURN VALUE:
+      an exception pointer (in any case -- either it is the newly created
+	  exception, or an exception thrown while trying to create it).
+
+*******************************************************************************/
+
+java_objectheader *exceptions_new_linkageerror(const char *message,
+											   classinfo *c)
+{
+	java_objectheader *o;
+	char              *msg;
+	s4                 msglen;
+
+	/* calculate exception message length */
+
+	msglen = strlen(message) + 1;
+	if (c) {
+		msglen += utf_bytes(c->name);
+	}
+		
+	/* allocate memory */
+
+	msg = MNEW(char, msglen);
+
+	/* generate message */
+
+	strcpy(msg,message);
+	if (c) {
+		utf_cat_classname(msg, c->name);
+	}
+
+	o = native_new_and_init_string(class_java_lang_LinkageError,
+								   javastring_new_from_utf_string(msg));
+
+	/* free memory */
+
+	MFREE(msg, char, msglen);
+
+	if (!o)
+		return *exceptionptr;
+
+	return o;
+}
+
+
 /* exceptions_new_nosuchmethoderror ********************************************
 
    Generates a java.lang.NoSuchMethodError with an error message.
 
+   IN:
+      c............class in which the method was not found
+	  name.........name of the method
+	  desc.........descriptor of the method
+
+   RETURN VALUE:
+      an exception pointer (in any case -- either it is the newly created
+	  exception, or an exception thrown while trying to create it).
+
 *******************************************************************************/
 
 java_objectheader *exceptions_new_nosuchmethoderror(classinfo *c,
@@ -636,8 +809,8 @@ java_objectheader *exceptions_new_nosuchmethoderror(classinfo *c,
 
 	/* calculate exception message length */
 
-	msglen = utf_strlen(c->name) + strlen(".") + utf_strlen(name) +
-		utf_strlen(desc) + strlen("0");
+	msglen = utf_bytes(c->name) + strlen(".") + utf_bytes(name) +
+		utf_bytes(desc) + strlen("0");
 
 	/* allocate memory */
 
@@ -645,25 +818,53 @@ java_objectheader *exceptions_new_nosuchmethoderror(classinfo *c,
 
 	/* generate message */
 
-	utf_sprint(msg, c->name);
+	utf_copy_classname(msg, c->name);
 	strcat(msg, ".");
-	utf_strcat(msg, name);
-	utf_strcat(msg, desc);
+	utf_cat(msg, name);
+	utf_cat(msg, desc);
 
 	o = native_new_and_init_string(class_java_lang_NoSuchMethodError,
-								   javastring_new_char(msg));
+								   javastring_new_from_utf_string(msg));
 
 	/* free memory */
 
 	MFREE(msg, char, msglen);
 
+	if (!o)
+		return *exceptionptr;
+
 	return o;
 }
 
 
+/* exceptions_throw_nosuchmethoderror ******************************************
+
+   Generates a java.lang.NoSuchMethodError with an error message.
+
+   IN:
+      c............class in which the method was not found
+	  name.........name of the method
+	  desc.........descriptor of the method
+
+*******************************************************************************/
+
+void exceptions_throw_nosuchmethoderror(classinfo *c, utf *name, utf *desc)
+{
+	*exceptionptr = exceptions_new_nosuchmethoderror(c, name, desc);
+}
+
+
 /* new_unsupportedclassversionerror ********************************************
 
-   generates a java.lang.UnsupportedClassVersionError for the classloader
+   Generate a java.lang.UnsupportedClassVersionError for the classloader
+
+   IN:
+      c............class in which the method was not found
+	  message......UTF-8 format string
+
+   RETURN VALUE:
+      an exception pointer (in any case -- either it is the newly created
+	  exception, or an exception thrown while trying to create it).
 
 *******************************************************************************/
 
@@ -676,7 +877,7 @@ java_objectheader *new_unsupportedclassversionerror(classinfo *c, const char *me
 
 	/* calculate exception message length */
 
-	msglen = utf_strlen(c->name) + strlen(" (") + strlen(")") + strlen("0");
+	msglen = utf_bytes(c->name) + strlen(" (") + strlen(")") + strlen("0");
 
 	va_start(ap, message);
 	msglen += get_variable_message_length(message, ap);
@@ -688,7 +889,7 @@ java_objectheader *new_unsupportedclassversionerror(classinfo *c, const char *me
 
 	/* generate message */
 
-	utf_sprint_classname(msg, c->name);
+	utf_copy_classname(msg, c->name);
 	strcat(msg, " (");
 
 	va_start(ap, message);
@@ -712,7 +913,15 @@ java_objectheader *new_unsupportedclassversionerror(classinfo *c, const char *me
 
 /* new_verifyerror *************************************************************
 
-   generates a java.lang.VerifyError for the jit compiler
+   Generates a java.lang.VerifyError for the JIT compiler.
+
+   IN:
+      m............method in which the error was found
+	  message......UTF-8 format string
+
+   RETURN VALUE:
+      an exception pointer (in any case -- either it is the newly created
+	  exception, or an exception thrown while trying to create it).
 
 *******************************************************************************/
 
@@ -730,9 +939,9 @@ java_objectheader *new_verifyerror(methodinfo *m, const char *message, ...)
 	msglen = 0;
 
 	if (m)
-		msglen = strlen("(class: ") + utf_strlen(m->class->name) +
-			strlen(", method: ") + utf_strlen(m->name) +
-			strlen(" signature: ") + utf_strlen(m->descriptor) +
+		msglen = strlen("(class: ") + utf_bytes(m->class->name) +
+			strlen(", method: ") + utf_bytes(m->name) +
+			strlen(" signature: ") + utf_bytes(m->descriptor) +
 			strlen(") ") + strlen("0");
 
 	va_start(ap, message);
@@ -747,11 +956,11 @@ java_objectheader *new_verifyerror(methodinfo *m, const char *message, ...)
 
 	if (m) {
 		strcpy(msg, "(class: ");
-		utf_strcat(msg, m->class->name);
+		utf_cat_classname(msg, m->class->name);
 		strcat(msg, ", method: ");
-		utf_strcat(msg, m->name);
+		utf_cat(msg, m->name);
 		strcat(msg, " signature: ");
-		utf_strcat(msg, m->descriptor);
+		utf_cat(msg, m->descriptor);
 		strcat(msg, ") ");
 	}
 
@@ -771,6 +980,80 @@ java_objectheader *new_verifyerror(methodinfo *m, const char *message, ...)
 }
 
 
+/* exceptions_throw_verifyerror_for_stack **************************************
+
+   throws a java.lang.VerifyError for an invalid stack slot type
+
+   IN:
+      m............method in which the error was found
+	  type.........the expected type
+
+   RETURN VALUE:
+      an exception pointer (in any case -- either it is the newly created
+	  exception, or an exception thrown while trying to create it).
+
+*******************************************************************************/
+
+void exceptions_throw_verifyerror_for_stack(methodinfo *m,int type)
+{
+	java_objectheader *o;
+	char              *msg;
+	s4                 msglen;
+	char              *typename;
+
+	/* calculate exception message length */
+
+	msglen = 0;
+
+	if (m)
+		msglen = strlen("(class: ") + utf_bytes(m->class->name) +
+			strlen(", method: ") + utf_bytes(m->name) +
+			strlen(" signature: ") + utf_bytes(m->descriptor) +
+			strlen(") Expecting to find longest-------typename on stack") 
+			+ strlen("0");
+
+	/* allocate memory */
+
+	msg = MNEW(char, msglen);
+
+	/* generate message */
+
+	if (m) {
+		strcpy(msg, "(class: ");
+		utf_cat_classname(msg, m->class->name);
+		strcat(msg, ", method: ");
+		utf_cat(msg, m->name);
+		strcat(msg, " signature: ");
+		utf_cat(msg, m->descriptor);
+		strcat(msg, ") ");
+	}
+	else {
+		msg[0] = 0;
+	}
+
+	strcat(msg,"Expecting to find ");
+	switch (type) {
+		case TYPE_INT: typename = "integer"; break;
+		case TYPE_LNG: typename = "long"; break;
+		case TYPE_FLT: typename = "float"; break;
+		case TYPE_DBL: typename = "double"; break;
+		case TYPE_ADR: typename = "object/array"; break;
+		default:       typename = "<INVALID>"; assert(0); break;
+	}
+	strcat(msg, typename);
+	strcat(msg, " on stack");
+
+	/* create exception object */
+
+	o = new_exception_message(string_java_lang_VerifyError, msg);
+
+	/* free memory */
+
+	MFREE(msg, char, msglen);
+
+	*exceptionptr = o;
+}
+
 /* new_arithmeticexception *****************************************************
 
    Generates a java.lang.ArithmeticException for the jit compiler.
@@ -791,18 +1074,19 @@ java_objectheader *new_arithmeticexception(void)
 }
 
 
-/* new_arrayindexoutofboundsexception ******************************************
+/* exceptions_new_arrayindexoutofboundsexception *******************************
 
-   Generates a java.lang.ArrayIndexOutOfBoundsException for the JIT
-   compiler.
+   Generates a java.lang.ArrayIndexOutOfBoundsException for the VM
+   system.
 
 *******************************************************************************/
 
 java_objectheader *new_arrayindexoutofboundsexception(s4 index)
 {
 	java_objectheader *e;
-	methodinfo *m;
-	java_lang_String *s;
+	methodinfo        *m;
+	java_objectheader *o;
+	java_lang_String  *s;
 
 	/* convert the index into a String, like Sun does */
 
@@ -812,28 +1096,46 @@ java_objectheader *new_arrayindexoutofboundsexception(s4 index)
 								 class_java_lang_Object,
 								 true);
 
-	if (!m)
+	if (m == NULL)
 		return *exceptionptr;
 
-	s = (java_lang_String *) asm_calljavafunction(m,
-												  (void *) (ptrint) index,
-												  NULL,
-												  NULL,
-												  NULL);
+	o = vm_call_method(m, NULL, index);
 
-	if (!s)
+	s = (java_lang_String *) o;
+
+	if (s == NULL)
 		return *exceptionptr;
 
 	e = new_exception_javastring(string_java_lang_ArrayIndexOutOfBoundsException,
 								 s);
 
-	if (!e)
+	if (e == NULL)
 		return *exceptionptr;
 
 	return e;
 }
 
 
+/* exceptions_throw_arrayindexoutofboundsexception *****************************
+
+   Generates a java.lang.ArrayIndexOutOfBoundsException for the VM
+   system.
+
+*******************************************************************************/
+
+void exceptions_throw_arrayindexoutofboundsexception(void)
+{
+	java_objectheader *e;
+
+	e = new_exception(string_java_lang_ArrayIndexOutOfBoundsException);
+
+	if (!e)
+		return;
+
+	*exceptionptr = e;
+}
+
+
 /* new_arraystoreexception *****************************************************
 
    generates a java.lang.ArrayStoreException for the jit compiler
@@ -873,7 +1175,7 @@ java_objectheader *new_classcastexception(void)
 }
 
 
-/* new_illegalargumentexception ************************************************
+/* exceptions_new_illegalargumentexception *************************************
 
    Generates a java.lang.IllegalArgumentException for the VM system.
 
@@ -883,13 +1185,28 @@ java_objectheader *new_illegalargumentexception(void)
 {
 	java_objectheader *e;
 
-	if (!(e = native_new_and_init(class_java_lang_IllegalArgumentException)))
+	e = native_new_and_init(class_java_lang_IllegalArgumentException);
+
+	if (!e)
 		return *exceptionptr;
 
 	return e;
 }
 
 
+/* exceptions_throw_illegalargumentexception ***********************************
+
+   Generates a java.lang.IllegalArgumentException for the VM system
+   and throw it in the VM system.
+
+*******************************************************************************/
+
+void exceptions_throw_illegalargumentexception(void)
+{
+	*exceptionptr = new_illegalargumentexception();
+}
+
+
 /* new_illegalmonitorstateexception ********************************************
 
    Generates a java.lang.IllegalMonitorStateException for the VM
@@ -901,16 +1218,18 @@ java_objectheader *new_illegalmonitorstateexception(void)
 {
 	java_objectheader *e;
 
-	if (!(e = native_new_and_init(class_java_lang_IllegalMonitorStateException)))
+	e = native_new_and_init(class_java_lang_IllegalMonitorStateException);
+
+	if (!e)
 		return *exceptionptr;
 
 	return e;
 }
 
 
-/* new_negativearraysizeexception **********************************************
+/* exceptions_new_negativearraysizeexception ***********************************
 
-   generates a java.lang.NegativeArraySizeException for the jit compiler
+   Generates a java.lang.NegativeArraySizeException for the VM system.
 
 *******************************************************************************/
 
@@ -927,6 +1246,18 @@ java_objectheader *new_negativearraysizeexception(void)
 }
 
 
+/* exceptions_throw_negativearraysizeexception *********************************
+
+   Generates a java.lang.NegativeArraySizeException for the VM system.
+
+*******************************************************************************/
+
+void exceptions_throw_negativearraysizeexception(void)
+{
+	*exceptionptr = new_negativearraysizeexception();
+}
+
+
 /* new_nullpointerexception ****************************************************
 
    generates a java.lang.NullPointerException for the jit compiler
@@ -946,6 +1277,193 @@ java_objectheader *new_nullpointerexception(void)
 }
 
 
+/* exceptions_throw_nullpointerexception ***************************************
+
+   Generates a java.lang.NullPointerException for the VM system and
+   throw it in the VM system.
+
+*******************************************************************************/
+
+void exceptions_throw_nullpointerexception(void)
+{
+	*exceptionptr = new_nullpointerexception();
+}
+
+
+/* exceptions_new_stringindexoutofboundsexception ******************************
+
+   Generates a java.lang.StringIndexOutOfBoundsException for the VM
+   system.
+
+*******************************************************************************/
+
+java_objectheader *exceptions_new_stringindexoutofboundsexception(void)
+{
+	java_objectheader *e;
+
+	e = new_exception(string_java_lang_StringIndexOutOfBoundsException);
+
+	if (e == NULL)
+		return *exceptionptr;
+
+	return e;
+}
+
+
+/* exceptions_throw_stringindexoutofboundsexception ****************************
+
+   Throws a java.lang.StringIndexOutOfBoundsException for the VM
+   system.
+
+*******************************************************************************/
+
+void exceptions_throw_stringindexoutofboundsexception(void)
+{
+	*exceptionptr = exceptions_new_stringindexoutofboundsexception();
+}
+
+
+/* exceptions_handle_exception *************************************************
+
+   Try to find an exception handler for the given exception and return it.
+   If no handler is found, exit the monitor of the method (if any)
+   and return NULL.
+
+   IN:
+      xptr.........the exception object
+	  xpc..........PC of where the exception was thrown
+	  pv...........Procedure Value of the current method
+	  sp...........current stack pointer
+
+   RETURN VALUE:
+      the address of the first matching exception handler, or
+	  NULL if no handler was found
+
+*******************************************************************************/
+
+u1 *exceptions_handle_exception(java_objectheader *xptr, u1 *xpc, u1 *pv, u1 *sp)
+{
+	methodinfo            *m;
+	s4                     framesize;
+	s4                     issync;
+	exceptionentry        *ex;
+	s4                     exceptiontablelength;
+	s4                     i;
+	classref_or_classinfo  cr;
+	classinfo             *c;
+#if defined(ENABLE_THREADS)
+	java_objectheader     *o;
+#endif
+
+	/* get methodinfo pointer from method header */
+
+	m                    = *((methodinfo **)    (pv + MethodPointer));
+	framesize            = *((s4 *)             (pv + FrameSize));
+	issync               = *((s4 *)             (pv + IsSync));
+	ex                   =   (exceptionentry *) (pv + ExTableStart);
+	exceptiontablelength = *((s4 *)             (pv + ExTableSize));
+
+#if !defined(NDEBUG)
+	/* print exception trace */
+
+	if (opt_verbose || opt_verbosecall || opt_verboseexception)
+		builtin_trace_exception(xptr, m, xpc, 1);
+#endif
+
+	for (i = 0; i < exceptiontablelength; i++) {
+		/* ATTENTION: keep this here, as we need to decrement the
+           pointer before the loop executes! */
+
+		ex--;
+
+		/* If the start and end PC is NULL, this means we have the
+		   special case of asm_vm_call_method.  So, just return the
+		   proper exception handler. */
+
+		if ((ex->startpc == NULL) && (ex->endpc == NULL))
+			return (u1 *) (ptrint) &asm_vm_call_method_exception_handler;
+
+		/* is the xpc is the current catch range */
+
+		if ((ex->startpc <= xpc) && (xpc < ex->endpc)) {
+			cr = ex->catchtype;
+
+			/* NULL catches everything */
+
+			if (cr.any == NULL) {
+#if !defined(NDEBUG)
+				/* Print stacktrace of exception when caught. */
+
+				if (opt_verboseexception) {
+					exceptions_print_exception(xptr);
+					stacktrace_print_trace(xptr);
+				}
+#endif
+
+				return ex->handlerpc;
+			}
+
+			/* resolve or load/link the exception class */
+
+			if (IS_CLASSREF(cr)) {
+				c = resolve_classref_eager(cr.ref);
+
+			} else {
+				c = cr.cls;
+
+				if (!(c->state & CLASS_LOADED))
+					/* use the methods' classloader */
+					if (!load_class_from_classloader(c->name,
+													 m->class->classloader))
+						return NULL;
+
+				if (!(c->state & CLASS_LINKED))
+					if (!link_class(c))
+						return NULL;
+			}
+
+			/* is the thrown exception an instance of the catch class? */
+
+			if (builtin_instanceof(xptr, c)) {
+#if !defined(NDEBUG)
+				/* Print stacktrace of exception when caught. */
+
+				if (opt_verboseexception) {
+					exceptions_print_exception(xptr);
+					stacktrace_print_trace(xptr);
+				}
+#endif
+
+				return ex->handlerpc;
+			}
+		}
+	}
+
+#if defined(ENABLE_THREADS)
+	/* is this method synchronized? */
+
+	if (issync) {
+		/* get synchronization object */
+
+# if defined(__MIPS__) && (SIZEOF_VOID_P == 4)
+		/* XXX change this if we ever want to use 4-byte stackslots */
+		o = *((java_objectheader **) (sp + issync - 8));
+# else
+		o = *((java_objectheader **) (sp + issync - SIZEOF_VOID_P));
+#endif
+
+		assert(o != NULL);
+
+		builtin_monitorexit(o);
+	}
+#endif
+
+	/* none of the exceptions catch this one */
+
+	return NULL;
+}
+
+
 /* exceptions_print_exception **************************************************
 
    Prints an exception, the detail message and the cause, if
@@ -953,6 +1471,7 @@ java_objectheader *new_nullpointerexception(void)
 
 *******************************************************************************/
 
+#if !defined(NDEBUG)
 void exceptions_print_exception(java_objectheader *xptr)
 {
 	java_lang_Throwable   *t;
@@ -960,17 +1479,23 @@ void exceptions_print_exception(java_objectheader *xptr)
 	utf                   *u;
 
 	t = (java_lang_Throwable *) xptr;
+
+	if (t == NULL) {
+		puts("NULL\n");
+		return;
+	}
+
 	cause = t->cause;
 
 	/* print the root exception */
 
-	utf_display_classname(t->header.vftbl->class->name);
+	utf_display_printable_ascii_classname(t->header.vftbl->class->name);
 
 	if (t->detailMessage) {
 		u = javastring_toutf(t->detailMessage, false);
 
 		printf(": ");
-		utf_display(u);
+		utf_display_printable_ascii(u);
 	}
 
 	putc('\n', stdout);
@@ -979,18 +1504,19 @@ void exceptions_print_exception(java_objectheader *xptr)
 
 	if (cause && (cause != t)) {
 		printf("Caused by: ");
-		utf_display_classname(cause->header.vftbl->class->name);
+		utf_display_printable_ascii_classname(cause->header.vftbl->class->name);
 
 		if (cause->detailMessage) {
 			u = javastring_toutf(cause->detailMessage, false);
 
 			printf(": ");
-			utf_display(u);
+			utf_display_printable_ascii(u);
 		}
 
 		putc('\n', stdout);
 	}
 }
+#endif /* !defined(NDEBUG) */
 
 
 /*
@@ -1004,4 +1530,5 @@ void exceptions_print_exception(java_objectheader *xptr)
  * c-basic-offset: 4
  * tab-width: 4
  * End:
+ * vim:noexpandtab:sw=4:ts=4:
  */
