@@ -1,9 +1,9 @@
 /* src/vm/jit/alpha/patcher.c - Alpha code patching functions
 
-   Copyright (C) 1996-2005 R. Grafl, A. Krall, C. Kruegel, C. Oates,
-   R. Obermaisser, M. Platter, M. Probst, S. Ring, E. Steiner,
-   C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich, J. Wenninger,
-   Institut f. Computersprachen - TU Wien
+   Copyright (C) 1996-2005, 2006 R. Grafl, A. Krall, C. Kruegel,
+   C. Oates, R. Obermaisser, M. Platter, M. Probst, S. Ring,
+   E. Steiner, C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich,
+   J. Wenninger, Institut f. Computersprachen - TU Wien
 
    This file is part of CACAO.
 
@@ -19,16 +19,16 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-   02111-1307, USA.
+   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
+   02110-1301, USA.
 
-   Contact: cacao@complang.tuwien.ac.at
+   Contact: cacao@cacaojvm.org
 
    Authors: Christian Thalinger
 
    Changes:
 
-   $Id: patcher.c,v 1.20 2005/11/08 10:58:04 twisti Exp $
+   $Id: patcher.c 4634 2006-03-16 15:16:16Z twisti $
 
 */
 
@@ -39,6 +39,7 @@
 #include "mm/memory.h"
 #include "native/native.h"
 #include "vm/builtin.h"
+#include "vm/class.h"
 #include "vm/field.h"
 #include "vm/initialize.h"
 #include "vm/options.h"
@@ -94,7 +95,7 @@ bool patcher_get_putstatic(u1 *sp)
 
 	/* check if the field's class is initialized */
 
-	if (!fi->class->initialized) {
+	if (!(fi->class->state & CLASS_INITIALIZED)) {
 		if (!initialize_class(fi->class)) {
 			PATCHER_MONITOREXIT;
 
@@ -108,7 +109,7 @@ bool patcher_get_putstatic(u1 *sp)
 
 	/* synchronize instruction cache */
 
-	asm_sync_instruction_cache();
+	md_icacheflush(NULL, 0);
 
 	/* patch the field value's address */
 
@@ -172,7 +173,7 @@ bool patcher_get_putfield(u1 *sp)
 
 	/* synchronize instruction cache */
 
-	asm_sync_instruction_cache();
+	md_icacheflush(NULL, 0);
 
 	PATCHER_MARK_PATCHED_MONITOREXIT;
 
@@ -229,7 +230,7 @@ bool patcher_aconst(u1 *sp)
 
 	/* synchronize instruction cache */
 
-	asm_sync_instruction_cache();
+	md_icacheflush(NULL, 0);
 
 	/* patch the classinfo pointer */
 
@@ -293,7 +294,7 @@ bool patcher_builtin_multianewarray(u1 *sp)
 
 	/* synchronize instruction cache */
 
-	asm_sync_instruction_cache();
+	md_icacheflush(NULL, 0);
 
 	/* patch the classinfo pointer */
 
@@ -356,7 +357,7 @@ bool patcher_builtin_arraycheckcast(u1 *sp)
 
 	/* synchronize instruction cache */
 
-	asm_sync_instruction_cache();
+	md_icacheflush(NULL, 0);
 
 	/* patch the classinfo pointer */
 
@@ -418,7 +419,7 @@ bool patcher_invokestatic_special(u1 *sp)
 
 	/* synchronize instruction cache */
 
-	asm_sync_instruction_cache();
+	md_icacheflush(NULL, 0);
 
 	/* patch stubroutine */
 
@@ -487,7 +488,7 @@ bool patcher_invokevirtual(u1 *sp)
 
 	/* synchronize instruction cache */
 
-	asm_sync_instruction_cache();
+	md_icacheflush(NULL, 0);
 
 	PATCHER_MARK_PATCHED_MONITOREXIT;
 
@@ -558,7 +559,7 @@ bool patcher_invokeinterface(u1 *sp)
 
 	/* synchronize instruction cache */
 
-	asm_sync_instruction_cache();
+	md_icacheflush(NULL, 0);
 
 	PATCHER_MARK_PATCHED_MONITOREXIT;
 
@@ -614,7 +615,7 @@ bool patcher_checkcast_instanceof_flags(u1 *sp)
 
 	/* synchronize instruction cache */
 
-	asm_sync_instruction_cache();
+	md_icacheflush(NULL, 0);
 
 	/* patch class flags */
 
@@ -687,7 +688,7 @@ bool patcher_checkcast_instanceof_interface(u1 *sp)
 
 	/* synchronize instruction cache */
 
-	asm_sync_instruction_cache();
+	md_icacheflush(NULL, 0);
 
 	PATCHER_MARK_PATCHED_MONITOREXIT;
 
@@ -745,7 +746,7 @@ bool patcher_checkcast_instanceof_class(u1 *sp)
 
 	/* synchronize instruction cache */
 
-	asm_sync_instruction_cache();
+	md_icacheflush(NULL, 0);
 
 	/* patch super class' vftbl */
 
@@ -786,7 +787,7 @@ bool patcher_clinit(u1 *sp)
 
 	/* check if the class is initialized */
 
-	if (!c->initialized) {
+	if (!(c->state & CLASS_INITIALIZED)) {
 		if (!initialize_class(c)) {
 			PATCHER_MONITOREXIT;
 
@@ -800,7 +801,7 @@ bool patcher_clinit(u1 *sp)
 
 	/* synchronize instruction cache */
 
-	asm_sync_instruction_cache();
+	md_icacheflush(NULL, 0);
 
 	PATCHER_MARK_PATCHED_MONITOREXIT;
 
@@ -816,6 +817,7 @@ bool patcher_clinit(u1 *sp)
 
 *******************************************************************************/
 
+#ifdef ENABLE_VERIFIER
 bool patcher_athrow_areturn(u1 *sp)
 {
 	u1                *ra;
@@ -852,12 +854,13 @@ bool patcher_athrow_areturn(u1 *sp)
 
 	/* synchronize instruction cache */
 
-	asm_sync_instruction_cache();
+	md_icacheflush(NULL, 0);
 
 	PATCHER_MARK_PATCHED_MONITOREXIT;
 
 	return true;
 }
+#endif /* ENABLE_VERIFIER */
 
 
 /* patcher_resolve_native ******************************************************
@@ -866,7 +869,7 @@ bool patcher_athrow_areturn(u1 *sp)
 
 *******************************************************************************/
 
-#if !defined(ENABLE_STATICVM)
+#if !defined(WITH_STATIC_CLASSPATH)
 bool patcher_resolve_native(u1 *sp)
 {
 	u1                *ra;
@@ -907,7 +910,7 @@ bool patcher_resolve_native(u1 *sp)
 
 	/* synchronize instruction cache */
 
-	asm_sync_instruction_cache();
+	md_icacheflush(NULL, 0);
 
 	/* patch native function pointer */
 
@@ -917,7 +920,7 @@ bool patcher_resolve_native(u1 *sp)
 
 	return true;
 }
-#endif /* !defined(ENABLE_STATICVM) */
+#endif /* !defined(WITH_STATIC_CLASSPATH) */
 
 
 /*
