@@ -1,9 +1,9 @@
 /* src/vm/jit/i386/patcher.c - i386 code patching functions
 
-   Copyright (C) 1996-2005 R. Grafl, A. Krall, C. Kruegel, C. Oates,
-   R. Obermaisser, M. Platter, M. Probst, S. Ring, E. Steiner,
-   C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich, J. Wenninger,
-   Institut f. Computersprachen - TU Wien
+   Copyright (C) 1996-2005, 2006 R. Grafl, A. Krall, C. Kruegel,
+   C. Oates, R. Obermaisser, M. Platter, M. Probst, S. Ring,
+   E. Steiner, C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich,
+   J. Wenninger, Institut f. Computersprachen - TU Wien
 
    This file is part of CACAO.
 
@@ -19,16 +19,16 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-   02111-1307, USA.
+   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
+   02110-1301, USA.
 
-   Contact: cacao@complang.tuwien.ac.at
+   Contact: cacao@cacaojvm.org
 
    Authors: Christian Thalinger
 
    Changes:
 
-   $Id: patcher.c,v 1.20 2005/11/07 21:12:01 twisti Exp $
+   $Id: patcher.c 4611 2006-03-15 11:18:30Z twisti $
 
 */
 
@@ -39,12 +39,14 @@
 #include "mm/memory.h"
 #include "native/native.h"
 #include "vm/builtin.h"
+#include "vm/class.h"
 #include "vm/field.h"
 #include "vm/initialize.h"
 #include "vm/options.h"
 #include "vm/resolve.h"
 #include "vm/references.h"
 #include "vm/jit/patcher.h"
+#include "vm/jit/i386/codegen.h"
 
 
 /* patcher_get_putstatic *******************************************************
@@ -66,15 +68,15 @@ bool patcher_get_putstatic(u1 *sp)
 
 	/* get stuff from the stack */
 
-	ra    = (u1 *)                *((ptrint *) (sp + 4 * 4));
+	ra    = (u1 *)                *((ptrint *) (sp + 5 * 4));
 	o     = (java_objectheader *) *((ptrint *) (sp + 3 * 4));
 	mcode =                       *((u8 *)     (sp + 1 * 4));
 	uf    = (unresolved_field *)  *((ptrint *) (sp + 0 * 4));
 
 	/* calculate and set the new return address */
 
-	ra = ra - 5;
-	*((ptrint *) (sp + 4 * 4)) = (ptrint) ra;
+	ra = ra - PATCHER_CALL_SIZE;
+	*((ptrint *) (sp + 5 * 4)) = (ptrint) ra;
 
 	PATCHER_MONITORENTER;
 
@@ -88,7 +90,7 @@ bool patcher_get_putstatic(u1 *sp)
 
 	/* check if the field's class is initialized */
 
-	if (!fi->class->initialized) {
+	if (!(fi->class->state & CLASS_INITIALIZED)) {
 		if (!initialize_class(fi->class)) {
 			PATCHER_MONITOREXIT;
 
@@ -135,15 +137,15 @@ bool patcher_getfield(u1 *sp)
 
 	/* get stuff from the stack */
 
-	ra    = (u1 *)                *((ptrint *) (sp + 4 * 4));
+	ra    = (u1 *)                *((ptrint *) (sp + 5 * 4));
 	o     = (java_objectheader *) *((ptrint *) (sp + 3 * 4));
 	mcode =                       *((u8 *)     (sp + 1 * 4));
 	uf    = (unresolved_field *)  *((ptrint *) (sp + 0 * 4));
 
 	/* calculate and set the new return address */
 
-	ra = ra - 5;
-	*((ptrint *) (sp + 4 * 4)) = (ptrint) ra;
+	ra = ra - PATCHER_CALL_SIZE;
+	*((ptrint *) (sp + 5 * 4)) = (ptrint) ra;
 
 	PATCHER_MONITORENTER;
 
@@ -199,15 +201,15 @@ bool patcher_putfield(u1 *sp)
 
 	/* get stuff from the stack */
 
-	ra    = (u1 *)                *((ptrint *) (sp + 4 * 4));
+	ra    = (u1 *)                *((ptrint *) (sp + 5 * 4));
 	o     = (java_objectheader *) *((ptrint *) (sp + 3 * 4));
 	mcode =                       *((u8 *)     (sp + 1 * 4));
 	uf    = (unresolved_field *)  *((ptrint *) (sp + 0 * 4));
 
 	/* calculate and set the new return address */
 
-	ra = ra - 5;
-	*((ptrint *) (sp + 4 * 4)) = (ptrint) ra;
+	ra = ra - PATCHER_CALL_SIZE;
+	*((ptrint *) (sp + 5 * 4)) = (ptrint) ra;
 
 	PATCHER_MONITORENTER;
 
@@ -235,16 +237,14 @@ bool patcher_putfield(u1 *sp)
 		*((u4 *) (ra + 2)) = (u4) (fi->offset);
 
 	} else {
-		/* long code is very special:
+		/* The long code is special:
 		 *
-		 * 8b 8c 24 00 00 00 00       mov    0x00000000(%esp),%ecx
-		 * 8b 94 24 00 00 00 00       mov    0x00000000(%esp),%edx
 		 * 89 8d 00 00 00 00          mov    %ecx,0x00000000(%ebp)
 		 * 89 95 00 00 00 00          mov    %edx,0x00000000(%ebp)
 		 */
 
-		*((u4 *) (ra + 7 + 7 + 2)) = (u4) (fi->offset);
-		*((u4 *) (ra + 7 + 7 + 6 + 2)) = (u4) (fi->offset + 4);
+		*((u4 *) (ra + 2))     = (u4) (fi->offset);
+		*((u4 *) (ra + 6 + 2)) = (u4) (fi->offset + 4);
 	}
 
 	PATCHER_MARK_PATCHED_MONITOREXIT;
@@ -272,15 +272,15 @@ bool patcher_putfieldconst(u1 *sp)
 
 	/* get stuff from the stack */
 
-	ra    = (u1 *)                *((ptrint *) (sp + 4 * 4));
+	ra    = (u1 *)                *((ptrint *) (sp + 5 * 4));
 	o     = (java_objectheader *) *((ptrint *) (sp + 3 * 4));
 	mcode =                       *((u8 *)     (sp + 1 * 4));
 	uf    = (unresolved_field *)  *((ptrint *) (sp + 0 * 4));
 
 	/* calculate and set the new return address */
 
-	ra = ra - 5;
-	*((ptrint *) (sp + 4 * 4)) = (ptrint) ra;
+	ra = ra - PATCHER_CALL_SIZE;
+	*((ptrint *) (sp + 5 * 4)) = (ptrint) ra;
 
 	PATCHER_MONITORENTER;
 
@@ -344,15 +344,15 @@ bool patcher_aconst(u1 *sp)
 
 	/* get stuff from the stack */
 
-	ra    = (u1 *)                *((ptrint *) (sp + 4 * 4));
+	ra    = (u1 *)                *((ptrint *) (sp + 5 * 4));
 	o     = (java_objectheader *) *((ptrint *) (sp + 3 * 4));
 	mcode =                       *((u8 *)     (sp + 1 * 4));
 	cr    = (constant_classref *) *((ptrint *) (sp + 0 * 4));
 
 	/* calculate and set the new return address */
 
-	ra = ra - 5;
-	*((ptrint *) (sp + 4 * 4)) = (ptrint) ra;
+	ra = ra - PATCHER_CALL_SIZE;
+	*((ptrint *) (sp + 5 * 4)) = (ptrint) ra;
 
 	PATCHER_MONITORENTER;
 
@@ -409,15 +409,15 @@ bool patcher_builtin_multianewarray(u1 *sp)
 
 	/* get stuff from the stack */
 
-	ra    = (u1 *)                *((ptrint *) (sp + 4 * 4));
+	ra    = (u1 *)                *((ptrint *) (sp + 5 * 4));
 	o     = (java_objectheader *) *((ptrint *) (sp + 3 * 4));
 	mcode =                       *((u8 *)     (sp + 1 * 4));
 	cr    = (constant_classref *) *((ptrint *) (sp + 0 * 4));
 
 	/* calculate and set the new return address */
 
-	ra = ra - 5;
-	*((ptrint *) (sp + 4 * 4)) = (ptrint) ra;
+	ra = ra - PATCHER_CALL_SIZE;
+	*((ptrint *) (sp + 5 * 4)) = (ptrint) ra;
 
 	PATCHER_MONITORENTER;
 
@@ -474,15 +474,15 @@ bool patcher_builtin_arraycheckcast(u1 *sp)
 
 	/* get stuff from the stack */
 
-	ra    = (u1 *)                *((ptrint *) (sp + 4 * 4));
+	ra    = (u1 *)                *((ptrint *) (sp + 5 * 4));
 	o     = (java_objectheader *) *((ptrint *) (sp + 3 * 4));
 	mcode =                       *((u8 *)     (sp + 1 * 4));
 	cr    = (constant_classref *) *((ptrint *) (sp + 0 * 4));
 
 	/* calculate and set the new return address */
 
-	ra = ra - 5;
-	*((ptrint *) (sp + 4 * 4)) = (ptrint) ra;
+	ra = ra - PATCHER_CALL_SIZE;
+	*((ptrint *) (sp + 5 * 4)) = (ptrint) ra;
 
 	PATCHER_MONITORENTER;
 
@@ -538,15 +538,15 @@ bool patcher_invokestatic_special(u1 *sp)
 
 	/* get stuff from the stack */
 
-	ra    = (u1 *)                *((ptrint *) (sp + 4 * 4));
+	ra    = (u1 *)                *((ptrint *) (sp + 5 * 4));
 	o     = (java_objectheader *) *((ptrint *) (sp + 3 * 4));
 	mcode =                       *((u8 *)     (sp + 1 * 4));
 	um    = (unresolved_method *) *((ptrint *) (sp + 0 * 4));
 
 	/* calculate and set the new return address */
 
-	ra = ra - 5;
-	*((ptrint *) (sp + 4 * 4)) = (ptrint) ra;
+	ra = ra - PATCHER_CALL_SIZE;
+	*((ptrint *) (sp + 5 * 4)) = (ptrint) ra;
 
 	PATCHER_MONITORENTER;
 
@@ -599,15 +599,15 @@ bool patcher_invokevirtual(u1 *sp)
 
 	/* get stuff from the stack */
 
-	ra    = (u1 *)                *((ptrint *) (sp + 4 * 4));
+	ra    = (u1 *)                *((ptrint *) (sp + 5 * 4));
 	o     = (java_objectheader *) *((ptrint *) (sp + 3 * 4));
 	mcode =                       *((u8 *)     (sp + 1 * 4));
 	um    = (unresolved_method *) *((ptrint *) (sp + 0 * 4));
 
 	/* calculate and set the new return address */
 
-	ra = ra - 5;
-	*((ptrint *) (sp + 4 * 4)) = (ptrint) ra;
+	ra = ra - PATCHER_CALL_SIZE;
+	*((ptrint *) (sp + 5 * 4)) = (ptrint) ra;
 
 	PATCHER_MONITORENTER;
 
@@ -662,15 +662,15 @@ bool patcher_invokeinterface(u1 *sp)
 
 	/* get stuff from the stack */
 
-	ra    = (u1 *)                *((ptrint *) (sp + 4 * 4));
+	ra    = (u1 *)                *((ptrint *) (sp + 5 * 4));
 	o     = (java_objectheader *) *((ptrint *) (sp + 3 * 4));
 	mcode =                       *((u8 *)     (sp + 1 * 4));
 	um    = (unresolved_method *) *((ptrint *) (sp + 0 * 4));
 
 	/* calculate and set the new return address */
 
-	ra = ra - 5;
-	*((ptrint *) (sp + 4 * 4)) = (ptrint) ra;
+	ra = ra - PATCHER_CALL_SIZE;
+	*((ptrint *) (sp + 5 * 4)) = (ptrint) ra;
 
 	PATCHER_MONITORENTER;
 
@@ -727,15 +727,15 @@ bool patcher_checkcast_instanceof_flags(u1 *sp)
 
 	/* get stuff from the stack */
 
-	ra    = (u1 *)                *((ptrint *) (sp + 4 * 4));
+	ra    = (u1 *)                *((ptrint *) (sp + 5 * 4));
 	o     = (java_objectheader *) *((ptrint *) (sp + 3 * 4));
 	mcode =                       *((u8 *)     (sp + 1 * 4));
 	cr    = (constant_classref *) *((ptrint *) (sp + 0 * 4));
 
 	/* calculate and set the new return address */
 
-	ra = ra - 5;
-	*((ptrint *) (sp + 4 * 4)) = (ptrint) ra;
+	ra = ra - PATCHER_CALL_SIZE;
+	*((ptrint *) (sp + 5 * 4)) = (ptrint) ra;
 
 	PATCHER_MONITORENTER;
 
@@ -790,15 +790,15 @@ bool patcher_checkcast_instanceof_interface(u1 *sp)
 
 	/* get stuff from the stack */
 
-	ra    = (u1 *)                *((ptrint *) (sp + 4 * 4));
+	ra    = (u1 *)                *((ptrint *) (sp + 5 * 4));
 	o     = (java_objectheader *) *((ptrint *) (sp + 3 * 4));
 	mcode =                       *((u8 *)     (sp + 1 * 4));
 	cr    = (constant_classref *) *((ptrint *) (sp + 0 * 4));
 
 	/* calculate and set the new return address */
 
-	ra = ra - 5;
-	*((ptrint *) (sp + 4 * 4)) = (ptrint) ra;
+	ra = ra - PATCHER_CALL_SIZE;
+	*((ptrint *) (sp + 5 * 4)) = (ptrint) ra;
 
 	PATCHER_MONITORENTER;
 
@@ -857,15 +857,15 @@ bool patcher_checkcast_class(u1 *sp)
 
 	/* get stuff from the stack */
 
-	ra    = (u1 *)                *((ptrint *) (sp + 4 * 4));
+	ra    = (u1 *)                *((ptrint *) (sp + 5 * 4));
 	o     = (java_objectheader *) *((ptrint *) (sp + 3 * 4));
 	mcode =                       *((u8 *)     (sp + 1 * 4));
 	cr    = (constant_classref *) *((ptrint *) (sp + 0 * 4));
 
 	/* calculate and set the new return address */
 
-	ra = ra - 5;
-	*((ptrint *) (sp + 4 * 4)) = (ptrint) ra;
+	ra = ra - PATCHER_CALL_SIZE;
+	*((ptrint *) (sp + 5 * 4)) = (ptrint) ra;
 
 	PATCHER_MONITORENTER;
 
@@ -920,15 +920,15 @@ bool patcher_instanceof_class(u1 *sp)
 
 	/* get stuff from the stack */
 
-	ra    = (u1 *)                *((ptrint *) (sp + 4 * 4));
+	ra    = (u1 *)                *((ptrint *) (sp + 5 * 4));
 	o     = (java_objectheader *) *((ptrint *) (sp + 3 * 4));
 	mcode =                       *((u8 *)     (sp + 1 * 4));
 	cr    = (constant_classref *) *((ptrint *) (sp + 0 * 4));
 
 	/* calculate and set the new return address */
 
-	ra = ra - 5;
-	*((ptrint *) (sp + 4 * 4)) = (ptrint) ra;
+	ra = ra - PATCHER_CALL_SIZE;
+	*((ptrint *) (sp + 5 * 4)) = (ptrint) ra;
 
 	PATCHER_MONITORENTER;
 
@@ -979,21 +979,21 @@ bool patcher_clinit(u1 *sp)
 
 	/* get stuff from the stack */
 
-	ra    = (u1 *)                *((ptrint *) (sp + 4 * 4));
+	ra    = (u1 *)                *((ptrint *) (sp + 5 * 4));
 	o     = (java_objectheader *) *((ptrint *) (sp + 3 * 4));
 	mcode =                       *((u8 *)     (sp + 1 * 4));
 	c     = (classinfo *)         *((ptrint *) (sp + 0 * 4));
 
 	/* calculate and set the new return address */
 
-	ra = ra - 5;
-	*((ptrint *) (sp + 4 * 4)) = (ptrint) ra;
+	ra = ra - PATCHER_CALL_SIZE;
+	*((ptrint *) (sp + 5 * 4)) = (ptrint) ra;
 
 	PATCHER_MONITORENTER;
 
 	/* check if the class is initialized */
 
-	if (!c->initialized) {
+	if (!(c->state & CLASS_INITIALIZED)) {
 		if (!initialize_class(c)) {
 			PATCHER_MONITOREXIT;
 
@@ -1020,6 +1020,7 @@ bool patcher_clinit(u1 *sp)
 
 *******************************************************************************/
 
+#ifdef ENABLE_VERIFIER
 bool patcher_athrow_areturn(u1 *sp)
 {
 	u1                *ra;
@@ -1030,15 +1031,15 @@ bool patcher_athrow_areturn(u1 *sp)
 
 	/* get stuff from the stack */
 
-	ra    = (u1 *)                *((ptrint *) (sp + 4 * 4));
+	ra    = (u1 *)                *((ptrint *) (sp + 5 * 4));
 	o     = (java_objectheader *) *((ptrint *) (sp + 3 * 4));
 	mcode =                       *((u8 *)     (sp + 1 * 4));
 	uc    = (unresolved_class *)  *((ptrint *) (sp + 0 * 4));
 
 	/* calculate and set the new return address */
 
-	ra = ra - 5;
-	*((ptrint *) (sp + 4 * 4)) = (ptrint) ra;
+	ra = ra - PATCHER_CALL_SIZE;
+	*((ptrint *) (sp + 5 * 4)) = (ptrint) ra;
 
 	PATCHER_MONITORENTER;
 
@@ -1059,6 +1060,7 @@ bool patcher_athrow_areturn(u1 *sp)
 
 	return true;
 }
+#endif /* ENABLE_VERIFIER */
 
 
 /* patcher_resolve_native ******************************************************
@@ -1072,7 +1074,7 @@ bool patcher_athrow_areturn(u1 *sp)
 
 *******************************************************************************/
 
-#if !defined(ENABLE_STATICVM)
+#if !defined(WITH_STATIC_CLASSPATH)
 bool patcher_resolve_native(u1 *sp)
 {
 	u1                *ra;
@@ -1083,15 +1085,15 @@ bool patcher_resolve_native(u1 *sp)
 
 	/* get stuff from the stack */
 
-	ra    = (u1 *)                *((ptrint *) (sp + 4 * 4));
+	ra    = (u1 *)                *((ptrint *) (sp + 5 * 4));
 	o     = (java_objectheader *) *((ptrint *) (sp + 3 * 4));
 	mcode =                       *((u8 *)     (sp + 1 * 4));
 	m     = (methodinfo *)        *((ptrint *) (sp + 0 * 4));
 
 	/* calculate and set the new return address */
 
-	ra = ra - 5;
-	*((ptrint *) (sp + 4 * 4)) = (ptrint) ra;
+	ra = ra - PATCHER_CALL_SIZE;
+	*((ptrint *) (sp + 5 * 4)) = (ptrint) ra;
 
 	PATCHER_MONITORENTER;
 
@@ -1121,7 +1123,7 @@ bool patcher_resolve_native(u1 *sp)
 
 	return true;
 }
-#endif /* !defined(ENABLE_STATICVM) */
+#endif /* !defined(WITH_STATIC_CLASSPATH) */
 
 
 /*
