@@ -1,9 +1,9 @@
 /* src/vm/jit/jit.c - calls the code generation functions
 
-   Copyright (C) 1996-2005 R. Grafl, A. Krall, C. Kruegel, C. Oates,
-   R. Obermaisser, M. Platter, M. Probst, S. Ring, E. Steiner,
-   C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich, J. Wenninger,
-   Institut f. Computersprachen - TU Wien
+   Copyright (C) 1996-2005, 2006 R. Grafl, A. Krall, C. Kruegel,
+   C. Oates, R. Obermaisser, M. Platter, M. Probst, S. Ring,
+   E. Steiner, C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich,
+   J. Wenninger, Institut f. Computersprachen - TU Wien
 
    This file is part of CACAO.
 
@@ -19,27 +19,28 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-   02111-1307, USA.
+   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
+   02110-1301, USA.
 
-   Contact: cacao@complang.tuwien.ac.at
+   Contact: cacao@cacaojvm.org
 
    Authors: Andreas Krall
             Reinhard Grafl
 
    Changes: Edwin Steiner
             Christian Thalinger
-	    Christian Ullrich
+            Christian Ullrich
 
-   $Id: jit.c,v 1.104 2005/11/22 23:42:43 twisti Exp $
+   $Id: jit.c 4957 2006-05-26 11:48:10Z edwin $
 
 */
 
 
 #include "config.h"
-
 #include "vm/types.h"
 
+#include <assert.h>
+
 #include "mm/memory.h"
 #include "native/native.h"
 #include "toolbox/logging.h"
@@ -51,41 +52,51 @@
 #include "vm/method.h"
 #include "vm/options.h"
 #include "vm/statistics.h"
-#include "vm/tables.h"
 #include "vm/jit/asmpart.h"
-#include "vm/jit/codegen.inc.h"
+#include "vm/jit/codegen-common.h"
 #include "vm/jit/disass.h"
+#include "vm/jit/dseg.h"
 #include "vm/jit/jit.h"
 
-#ifdef LSRA
-# include "vm/jit/lsra.h"
-#endif
 
 #include "vm/jit/parse.h"
 #include "vm/jit/reg.h"
 #include "vm/jit/stack.h"
 
-#if defined(USE_INLINING)
-# include "vm/jit/inline/inline.h"
-# include "vm/jit/inline/parseRT.h"
-# include "vm/jit/inline/parseXTA.h"
+#include "vm/jit/allocator/simplereg.h"
+#if defined(ENABLE_LSRA)
+# include "vm/jit/allocator/lsra.h"
+#endif
+
+#if defined(ENABLE_IFCONV)
+# include "vm/jit/ifconv/ifconv.h"
 #endif
 
 #include "vm/jit/loop/analyze.h"
 #include "vm/jit/loop/graph.h"
 #include "vm/jit/loop/loop.h"
 #include "vm/jit/verify/typecheck.h"
+#include "vm/rt-timing.h"
 
-#if defined(USE_THREADS)
-# if defined(NATIVE_THREADS)
-#  include "threads/native/threads.h"
-# else
-#  include "threads/green/threads.h"
-# endif
+#if defined(ENABLE_THREADS)
+# include "threads/native/threads.h"
 #endif
 
 
+/* debug macros ***************************************************************/
 
+#if !defined(NDEBUG)
+#define DEBUG_JIT_COMPILEVERBOSE(x) \
+    do { \
+        if (compileverbose) { \
+            log_message_method(x, m); \
+        } \
+    } while (0)
+#else
+#define DEBUG_JIT_COMPILEVERBOSE(x)    /* nothing */
+#endif
+
+ 
 /* global switches ************************************************************/
 
 int stackreq[256];
@@ -153,7 +164,7 @@ int jcommandsize[256] = {
 #define ICMD_IFLE_ICONST      20
 	3,
 	                                    /* order of LOAD instructions must be */
-	                                    /* equal to order of TYPE_XXX defines */
+	                                    /* equal to order of TYPE_*   defines */
 #define JAVA_ILOAD            21
 #define ICMD_ILOAD            21        /* op1 = local variable               */
 	2,                      
@@ -254,7 +265,7 @@ int jcommandsize[256] = {
 #define ICMD_SALOAD           53
 	1,
 	                                    /* order of STORE instructions must be*/
-	                                    /* equal to order of TYPE_XXX defines */
+	                                    /* equal to order of TYPE_* defines   */
 #define JAVA_ISTORE           54
 #define ICMD_ISTORE           54        /* op1 = local variable               */
 	2,
@@ -716,10 +727,53 @@ int jcommandsize[256] = {
 #define ICMD_LMULPOW2         215
 	1,
 
+#define ICMD_IF_FCMPEQ        216
+	1,
+#define ICMD_IF_FCMPNE        217
+
+#define ICMD_IF_FCMPL_LT      218
+	1,
+#define ICMD_IF_FCMPL_GE      219
+	1,
+#define ICMD_IF_FCMPL_GT      220
+	1,
+#define ICMD_IF_FCMPL_LE      221
+	1,
+
+#define ICMD_IF_FCMPG_LT      222
+	1,
+#define ICMD_IF_FCMPG_GE      223
+	1,
+#define ICMD_IF_FCMPG_GT      224
+	1,
+#define ICMD_IF_FCMPG_LE      225
+	1,
+
+#define ICMD_IF_DCMPEQ        226
+	1,
+#define ICMD_IF_DCMPNE        227
+	1,
+
+#define ICMD_IF_DCMPL_LT      228
+	1,
+#define ICMD_IF_DCMPL_GE      229
+	1,
+#define ICMD_IF_DCMPL_GT      230
+	1,
+#define ICMD_IF_DCMPL_LE      231
+	1,
+
+#define ICMD_IF_DCMPG_LT      232
+	1,
+#define ICMD_IF_DCMPG_GE      233
+	1,
+#define ICMD_IF_DCMPG_GT      234
+	1,
+#define ICMD_IF_DCMPG_LE      235
+	1,
+
 	/* unused */
-	            1,1,1,1,
-	1,1,1,1,1,1,1,1,1,1,
-	1,1,1,1,1,1,1,1,1,1,
+	        1,1,1,1,1,1,
 	1,1,1,1,1,1,1,1,1,1,
 	1,1,1,1,1
 };
@@ -943,18 +997,41 @@ char *icmd_names[256] = {
 	"IMULPOW2       ", /*             214 */
 	"LMULPOW2       ", /*             215 */
 
-	"UNDEF216", "UNDEF217", "UNDEF218", "UNDEF219", "UNDEF220",
-	"UNDEF221", "UNDEF222", "UNDEF223", "UNDEF224", "UNDEF225",
-	"UNDEF226", "UNDEF227", "UNDEF228", "UNDEF229", "UNDEF230",
-	"UNDEF231", "UNDEF232", "UNDEF233", "UNDEF234", "UNDEF235",
+	"IF_FCMPEQ      ", /*             216 */
+	"IF_FCMPNE      ", /*             217 */
+
+	"IF_FCMPL_LT    ", /*             218 */
+	"IF_FCMPL_GE    ", /*             219 */
+	"IF_FCMPL_GT    ", /*             220 */
+	"IF_FCMPL_LE    ", /*             221 */
+
+	"IF_FCMPG_LT    ", /*             222 */
+	"IF_FCMPG_GE    ", /*             223 */
+	"IF_FCMPG_GT    ", /*             224 */
+	"IF_FCMPG_LE    ", /*             225 */
+
+	"IF_DCMPEQ      ", /*             226 */
+	"IF_DCMPNE      ", /*             227 */
+
+	"IF_DCMPL_LT    ", /*             228 */
+	"IF_DCMPL_GE    ", /*             229 */
+	"IF_DCMPL_GT    ", /*             230 */
+	"IF_DCMPL_LE    ", /*             231 */
+	
+	"IF_DCMPG_LT    ", /*             232 */
+	"IF_DCMPG_GE    ", /*             233 */
+	"IF_DCMPG_GT    ", /*             234 */
+	"IF_DCMPG_LE    ", /*             235 */
+	
 	"UNDEF236", "UNDEF237", "UNDEF238", "UNDEF239", "UNDEF240",
 	"UNDEF241", "UNDEF242", "UNDEF243", "UNDEF244", "UNDEF245",
 	"UNDEF246", "UNDEF247", "UNDEF248", "UNDEF249", "UNDEF250",
 
 	"INLINE_START   ", /*             251 */
 	"INLINE_END     ", /*             252 */
+	"INLINE_GOTO    ", /*             253 */
 
-	"UNDEF253", "UNDEF254",
+	"UNDEF254",
 
 	"BUILTIN        "  /*             255 */
 };
@@ -1189,10 +1266,6 @@ void jit_init(void)
 {
 	s4 i;
 
-#if defined(__ALPHA__)
-	has_ext_instr_set = ! has_no_x_instr_set();
-#endif
-
 	for (i = 0; i < 256; i++)
 		stackreq[i] = 1;
 
@@ -1265,11 +1338,7 @@ void jit_init(void)
 	stackreq[JAVA_IFNONNULL]    = 0;
 	stackreq[JAVA_GOTO_W]       = 0;
 	stackreq[JAVA_BREAKPOINT]   = 0;
-
-	/* we need one dummy stack slot for IINC in order to */
-	/* avoid that the modified local variable is */
-	/* kept on the stack (see stack.c, ICMD_IINC) */
-	stackreq[JAVA_IINC]         = 1;
+	stackreq[JAVA_IINC]         = 0;
 	
 	stackreq[JAVA_SWAP] = 2;
 	stackreq[JAVA_DUP2] = 2;
@@ -1310,28 +1379,42 @@ static u1 *do_nothing_function(void)
 
 /* jit_compile *****************************************************************
 
-   jit_compile, new version of compiler, translates one method to machine code
+   Translates one method to machine code.
 
 *******************************************************************************/
 
-static u1 *jit_compile_intern(methodinfo *m, codegendata *cd, registerdata *rd,
-							  loopdata *ld, t_inlining_globals *id);
+static u1 *jit_compile_intern(jitdata *jd);
 
 u1 *jit_compile(methodinfo *m)
 {
-	u1                 *r;
-	codegendata        *cd;
-	registerdata       *rd;
-	loopdata           *ld;
-	t_inlining_globals *id;
-	s4                  dumpsize;
+	u1      *r;
+	jitdata *jd;
+	s4       dumpsize;
 
-#if defined(STATISTICS)
-	if (opt_stat)
-		count_jit_calls++;
+	STATISTICS(count_jit_calls++);
+
+	/* Initialize the static function's class. */
+
+	/* ATTENTION: This MUST be done before the method lock is aquired,
+	   otherwise we could run into a deadlock with <clinit>'s that
+	   call static methods of it's own class. */
+
+  	if ((m->flags & ACC_STATIC) && !(m->class->state & CLASS_INITIALIZED)) {
+#if !defined(NDEBUG)
+		if (initverbose)
+			log_message_class("Initialize class ", m->class);
 #endif
 
-#if defined(USE_THREADS)
+		if (!initialize_class(m->class))
+			return NULL;
+
+		/* check if the method has been compiled during initialization */
+
+		if ((m->code != NULL) && (m->code->entrypoint != NULL))
+			return m->code->entrypoint;
+	}
+
+#if defined(ENABLE_THREADS)
 	/* enter a monitor on the method */
 
 	builtin_monitorenter((java_objectheader *) m);
@@ -1339,23 +1422,21 @@ u1 *jit_compile(methodinfo *m)
 
 	/* if method has been already compiled return immediately */
 
-	if (m->entrypoint) {
-#if defined(USE_THREADS)
+	if (m->code) {
+#if defined(ENABLE_THREADS)
 		builtin_monitorexit((java_objectheader *) m);
 #endif
 
-		return m->entrypoint;
+		assert(m->code->entrypoint);
+		return m->code->entrypoint;
 	}
 
-#if defined(STATISTICS)
-	if (opt_stat)
-		count_methods++;
-#endif
+	STATISTICS(count_methods++);
 
-#if defined(STATISTICS)
+#if defined(ENABLE_STATISTICS)
 	/* measure time */
 
-	if (getcompilingtime)
+	if (opt_getcompilingtime)
 		compilingtime_start();
 #endif
 
@@ -1363,86 +1444,80 @@ u1 *jit_compile(methodinfo *m)
 
 	dumpsize = dump_size();
 
-	/* allocate memory */
+	/* allocate jitdata structure and fill it */
 
-	cd = DNEW(codegendata);
-	rd = DNEW(registerdata);
-	ld = DNEW(loopdata);
-	id = DNEW(t_inlining_globals);
+	jd = DNEW(jitdata);
 
-#if defined(USE_INLINING)
-	/* RTA static analysis must be called before inlining */
-	if (opt_rt)
-		RT_jit_parse(m); /* will be called just once */
-	                     /* return value ignored for now */
-	/* XTA static analysis must be called before inlining */
-	if (opt_xta)
-		XTA_jit_parse(m); /* will be called just once */
-	                      /* return value ignored for now */
-
-	/* must be called before reg_setup, because it can change maxlocals */
-	/* init reqd to initialize for parse even in no inlining */
-	inlining_setup(m, id);
+	jd->m     = m;
+	jd->cd    = DNEW(codegendata);
+	jd->rd    = DNEW(registerdata);
+#if defined(ENABLE_LOOP)
+	jd->ld    = DNEW(loopdata);
 #endif
+	jd->flags = 0;
+
+	/* Allocate codeinfo memory from the heap as we need to keep them. */
+
+	jd->code  = code_codeinfo_new(m); /* XXX check allocation */
+
+	/* set the flags for the current JIT run */
+
+	if (opt_ifconv)
+		jd->flags |= JITDATA_FLAG_IFCONV;
 
 #if defined(ENABLE_JIT)
 # if defined(ENABLE_INTRP)
 	if (!opt_intrp)
 # endif
 		/* initialize the register allocator */
-		reg_setup(m, rd, id);
+
+		reg_setup(jd);
 #endif
 
 	/* setup the codegendata memory */
-	codegen_setup(m, cd, id);
+
+	codegen_setup(jd);
 
 	/* now call internal compile function */
 
-	r = jit_compile_intern(m, cd, rd, ld, id);
-
-	/* free some memory */
-
-#if defined(ENABLE_JIT)
-# if defined(ENABLE_INTRP)
-	if (!opt_intrp)
-# endif
-		codegen_free(m, cd);
-#endif
+	r = jit_compile_intern(jd);
 
 	/* clear pointers to dump memory area */
 
-	m->basicblocks = NULL;
+	m->basicblocks     = NULL;
 	m->basicblockindex = NULL;
-	m->instructions = NULL;
-	m->stack = NULL;
-	/* NO !!! m->exceptiontable = NULL; */
+	m->instructions    = NULL;
+	m->stack           = NULL;
 
 	/* release dump area */
 
 	dump_release(dumpsize);
 
-#if defined(STATISTICS)
+#if defined(ENABLE_STATISTICS)
 	/* measure time */
 
-	if (getcompilingtime)
+	if (opt_getcompilingtime)
 		compilingtime_stop();
 #endif
 
-	/* define in options.h; Used in main.c, jit.c & inline.c */
-#ifdef INAFTERMAIN
-	if ((utf_new_char("main") == m->name) && (useinliningm))
-		useinlining = false;
-#endif
 
-#if defined(USE_THREADS)
+#if defined(ENABLE_THREADS)
 	/* leave the monitor */
 
-	builtin_monitorexit((java_objectheader *) m );
+	builtin_monitorexit((java_objectheader *) m);
 #endif
 
 	if (r) {
-		if (compileverbose)
-			log_message_method("Running: ", m);
+		DEBUG_JIT_COMPILEVERBOSE("Running: ");
+
+	} else {
+		/* We had an exception! Finish stuff here if necessary. */
+
+		/* Release memory for basic block profiling information. */
+
+		if (opt_prof)
+			if (m->bbfrequency)
+				MFREE(m->bbfrequency, u4, m->basicblockcount);
 	}
 
 	/* return pointer to the methods entry point */
@@ -1450,206 +1525,315 @@ u1 *jit_compile(methodinfo *m)
 	return r;
 }
 
-
 /* jit_compile_intern **********************************************************
 
    Static internal function which does the actual compilation.
 
 *******************************************************************************/
 
-static u1 *jit_compile_intern(methodinfo *m, codegendata *cd, registerdata *rd,
-							  loopdata *ld, t_inlining_globals *id)
+static u1 *jit_compile_intern(jitdata *jd)
 {
+	methodinfo  *m;
+	codegendata *cd;
+	codeinfo    *code;
+
+#if defined(ENABLE_RT_TIMING)
+	struct timespec time_start,time_checks,time_parse,time_stack,
+					time_typecheck,time_loop,time_ifconv,time_alloc,
+					time_rplpoints,time_codegen;
+#endif
+	
+	RT_TIMING_GET_TIME(time_start);
+
+	/* get required compiler data */
+
+	m    = jd->m;
+	code = jd->code;
+	cd   = jd->cd;
+	
 	/* print log message for compiled method */
 
-	if (compileverbose)
-		log_message_method("Compiling: ", m);
-
-	/* initialize the static function's class */
-
-  	if ((m->flags & ACC_STATIC) && !m->class->initialized) {
-		if (initverbose)
-			log_message_class("Initialize class ", m->class);
-
-		if (!initialize_class(m->class))
-			return NULL;
-	}
+	DEBUG_JIT_COMPILEVERBOSE("Compiling: ");
 
 	/* handle native methods and create a native stub */
 
 	if (m->flags & ACC_NATIVE) {
 		functionptr f;
 
-#if defined(ENABLE_STATICVM)
+#if defined(WITH_STATIC_CLASSPATH)
 		f = native_findfunction(m->class->name,	m->name, m->descriptor,
 								(m->flags & ACC_STATIC));
-		if (!f)
+		if (f == NULL)
 			return NULL;
 #else
 
 		f = NULL;
 #endif
 
-		m->entrypoint = codegen_createnativestub(f, m);
+		code = codegen_createnativestub(f, m);
 
-		return m->entrypoint;
+		assert(!m->code); /* native methods are never recompiled */
+		m->code = code;
+		
+		return code->entrypoint;
 	}
 
 	/* if there is no javacode, print error message and return empty method   */
 
 	if (!m->jcode) {
-		if (compileverbose)
-			log_message_method("No code given for: ", m);
+		DEBUG_JIT_COMPILEVERBOSE("No code given for: ");
 
-		m->entrypoint = (u1 *) (ptrint) do_nothing_function;
+		code->entrypoint = (u1 *) (ptrint) do_nothing_function;
+		m->code = code;
 
-		return m->entrypoint;           /* return empty method                */
+		return code->entrypoint;        /* return empty method                */
 	}
 
 	/* initialisation of variables and subsystems */
 
 	m->isleafmethod = true;
 
-#if defined(STATISTICS)
+#if defined(ENABLE_STATISTICS)
 	if (opt_stat) {
-		count_tryblocks += m->exceptiontablelength;
+		count_tryblocks    += m->exceptiontablelength;
 		count_javacodesize += m->jcodelength + 18;
-		count_javaexcsize += m->exceptiontablelength * SIZEOF_VOID_P;
+		count_javaexcsize  += m->exceptiontablelength * SIZEOF_VOID_P;
 	}
 #endif
 
-	/* initialise parameter type descriptor */
-
-	method_descriptor2types(m);
+	RT_TIMING_GET_TIME(time_checks);
 
 	/* call the compiler passes ***********************************************/
 
-	if (compileverbose)
-		log_message_method("Parsing: ", m);
+	DEBUG_JIT_COMPILEVERBOSE("Parsing: ");
 
 	/* call parse pass */
 
-	if (!parse(m, cd, id)) {
-		if (compileverbose)
-			log_message_method("Exception while parsing: ", m);
+	if (!parse(jd)) {
+		DEBUG_JIT_COMPILEVERBOSE("Exception while parsing: ");
 
 		return NULL;
 	}
+	RT_TIMING_GET_TIME(time_parse);
 
-	if (compileverbose) {
-		log_message_method("Parsing done: ", m);
-		log_message_method("Analysing: ", m);
-	}
+	DEBUG_JIT_COMPILEVERBOSE("Parsing done: ");
+	DEBUG_JIT_COMPILEVERBOSE("Analysing: ");
 
 	/* call stack analysis pass */
 
-	if (!analyse_stack(m, cd, rd)) {
-		if (compileverbose)
-			log_message_method("Exception while analysing: ", m);
+	if (!stack_analyse(jd)) {
+		DEBUG_JIT_COMPILEVERBOSE("Exception while analysing: ");
 
 		return NULL;
 	}
+	RT_TIMING_GET_TIME(time_stack);
 
-	if (compileverbose)
-		log_message_method("Analysing done: ", m);
+	DEBUG_JIT_COMPILEVERBOSE("Analysing done: ");
 
-#ifdef CACAO_TYPECHECK
+#ifdef ENABLE_VERIFIER
 	if (opt_verify) {
-		if (compileverbose)
-			log_message_method("Typechecking: ", m);
+		DEBUG_JIT_COMPILEVERBOSE("Typechecking: ");
 
 		/* call typecheck pass */
-		if (!typecheck(m, cd, rd)) {
-			if (compileverbose)
-				log_message_method("Exception while typechecking: ", m);
+		if (!typecheck(jd)) {
+			DEBUG_JIT_COMPILEVERBOSE("Exception while typechecking: ");
 
 			return NULL;
 		}
 
-		if (compileverbose)
-			log_message_method("Typechecking done: ", m);
+		DEBUG_JIT_COMPILEVERBOSE("Typechecking done: ");
 	}
 #endif
+	RT_TIMING_GET_TIME(time_typecheck);
 
+#if defined(ENABLE_LOOP)
 	if (opt_loops) {
-		depthFirst(m, ld);
-		analyseGraph(m, ld);
-		optimize_loops(m, cd, ld);
+		depthFirst(jd);
+		analyseGraph(jd);
+		optimize_loops(jd);
 	}
+#endif
+	RT_TIMING_GET_TIME(time_loop);
+
+#if defined(ENABLE_IFCONV)
+	if (jd->flags & JITDATA_FLAG_IFCONV)
+		if (!ifconv_static(jd))
+			return NULL;
+#endif
+	RT_TIMING_GET_TIME(time_ifconv);
 
 #if defined(ENABLE_JIT)
 # if defined(ENABLE_INTRP)
 	if (!opt_intrp) {
 # endif
-		if (compileverbose)
-			log_message_method("Allocating registers: ", m);
+		DEBUG_JIT_COMPILEVERBOSE("Allocating registers: ");
 
 		/* allocate registers */
-#ifdef LSRA
+# if defined(ENABLE_LSRA)
 		if (opt_lsra) {
-			lsra(m, cd, rd, id);
-#ifdef STATISTICS
-			if (opt_stat) count_methods_allocated_by_lsra++;
-#endif
+			if (!lsra(jd))
+				return NULL;
+
+			STATISTICS(count_methods_allocated_by_lsra++);
+
 		} else
-#endif /* LSRA */
+# endif /* defined(ENABLE_LSRA) */
 		{
-#ifdef STATISTICS
-			if (opt_stat)
-				count_locals_conflicts += (cd->maxlocals-1)*(cd->maxlocals);
-#endif		
-			regalloc(m, cd, rd);
+			STATISTICS(count_locals_conflicts += (cd->maxlocals - 1) * (cd->maxlocals));
+
+			regalloc(jd);
 		}
 
-#ifdef STATISTICS
-		if (opt_stat)
-			reg_make_statistics(m, cd, rd);
-#endif
+		STATISTICS(reg_make_statistics(jd));
 
-		if (compileverbose)
-			log_message_method("Allocating registers done: ", m);
+		DEBUG_JIT_COMPILEVERBOSE("Allocating registers done: ");
 # if defined(ENABLE_INTRP)
 	}
 # endif
 #endif /* defined(ENABLE_JIT) */
+	RT_TIMING_GET_TIME(time_alloc);
 
-	if (compileverbose)
-		log_message_method("Generating code: ", m);
+	/* Allocate memory for basic block profiling information. This
+	   _must_ be done after loop optimization and register allocation,
+	   since they can change the basic block count. */
+
+	if (opt_prof)
+		m->bbfrequency = MNEW(u4, m->basicblockcount);
+
+	DEBUG_JIT_COMPILEVERBOSE("Generating code: ");
+
+	/* create the replacement points */
+
+	if (!replace_create_replacement_points(jd))
+		return NULL;
+	RT_TIMING_GET_TIME(time_rplpoints);
 
 	/* now generate the machine code */
 
-	if (!codegen(m, cd, rd)) {
-		if (compileverbose)
-			log_message_method("Exception while generating code: ", m);
+#if defined(ENABLE_JIT)
+# if defined(ENABLE_INTRP)
+	if (opt_intrp) {
+		if (!intrp_codegen(jd)) {
+			DEBUG_JIT_COMPILEVERBOSE("Exception while generating code: ");
+
+			return NULL;
+		}
+	} else
+# endif
+		{
+			if (!codegen(jd)) {
+				DEBUG_JIT_COMPILEVERBOSE("Exception while generating code: ");
+
+				return NULL;
+			}
+		}
+#else
+	if (!intrp_codegen(jd)) {
+		DEBUG_JIT_COMPILEVERBOSE("Exception while generating code: ");
 
 		return NULL;
 	}
+#endif
+	RT_TIMING_GET_TIME(time_codegen);
 
-	if (compileverbose)
-		log_message_method("Generating code done: ", m);
+	DEBUG_JIT_COMPILEVERBOSE("Generating code done: ");
 
+#if !defined(NDEBUG)
 	/* intermediate and assembly code listings */
 		
 	if (opt_showintermediate) {
-		show_icmd_method(m, cd, rd);
+		stack_show_method(jd);
 
 	} else if (opt_showdisassemble) {
-		disassemble((u1 *) (ptrint) m->entrypoint,
-					(u1 *) (ptrint) m->entrypoint + (m->mcodelength - cd->dseglen));
+# if defined(ENABLE_DISASSEMBLER)
+		DISASSEMBLE(code->entrypoint,
+					code->entrypoint + (code->mcodelength - cd->dseglen));
+# endif
 	}
 
 	if (opt_showddatasegment)
-		dseg_display(m, cd);
+		dseg_display(jd);
+#endif
 
-	if (compileverbose)
-		log_message_method("Compiling done: ", m);
+	DEBUG_JIT_COMPILEVERBOSE("Compiling done: ");
+
+	/* switch to the newly generated code */
+
+	assert(code);
+	assert(code->entrypoint);
+
+	/* add the current compile version to the methodinfo */
+
+	code->prev = m->code;
+	m->code = code;
+
+	RT_TIMING_TIME_DIFF(time_start,time_checks,RT_TIMING_JIT_CHECKS);
+	RT_TIMING_TIME_DIFF(time_checks,time_parse,RT_TIMING_JIT_PARSE);
+	RT_TIMING_TIME_DIFF(time_parse,time_stack,RT_TIMING_JIT_STACK);
+	RT_TIMING_TIME_DIFF(time_stack,time_typecheck,RT_TIMING_JIT_TYPECHECK);
+	RT_TIMING_TIME_DIFF(time_typecheck,time_loop,RT_TIMING_JIT_LOOP);
+	RT_TIMING_TIME_DIFF(time_loop,time_alloc,RT_TIMING_JIT_ALLOC);
+	RT_TIMING_TIME_DIFF(time_alloc,time_rplpoints,RT_TIMING_JIT_RPLPOINTS);
+	RT_TIMING_TIME_DIFF(time_rplpoints,time_codegen,RT_TIMING_JIT_CODEGEN);
+	RT_TIMING_TIME_DIFF(time_start,time_codegen,RT_TIMING_JIT_TOTAL);
 
 	/* return pointer to the methods entry point */
 
-	return m->entrypoint;
+	return code->entrypoint;
 } 
 
 
+/* jit_asm_compile *************************************************************
+
+   This method is called from asm_vm_call_method and does things like:
+   create stackframe info for exceptions, compile the method, patch
+   the entrypoint of the method into the calculated address in the JIT
+   code, and flushes the instruction cache.
+
+*******************************************************************************/
+
+u1 *jit_asm_compile(methodinfo *m, u1 *mptr, u1 *sp, u1 *ra)
+{
+	stackframeinfo  sfi;
+	u1             *entrypoint;
+	u1             *pa;
+	ptrint         *p;
+
+	/* create the stackframeinfo (XPC is equal to RA) */
+
+	stacktrace_create_extern_stackframeinfo(&sfi, NULL, sp, ra, ra);
+
+	/* actually compile the method */
+
+	entrypoint = jit_compile(m);
+
+	/* remove the stackframeinfo */
+
+	stacktrace_remove_stackframeinfo(&sfi);
+
+	/* there was a problem during compilation */
+
+	if (entrypoint == NULL)
+		return NULL;
+
+	/* get the method patch address */
+
+	pa = md_get_method_patch_address(ra, &sfi, mptr);
+
+	/* patch the method entry point */
+
+	p = (ptrint *) pa;
+
+	*p = (ptrint) entrypoint;
+
+	/* flush the instruction cache */
+
+	md_icacheflush(pa, SIZEOF_VOID_P);
+
+	return entrypoint;
+}
+
+
 /*
  * These are local overrides for various environment variables in Emacs.
  * Please do not remove this and leave it at the end of the file, where
@@ -1661,4 +1845,5 @@ static u1 *jit_compile_intern(methodinfo *m, codegendata *cd, registerdata *rd,
  * c-basic-offset: 4
  * tab-width: 4
  * End:
+ * vim:noexpandtab:sw=4:ts=4:
  */
