@@ -1,9 +1,9 @@
 /* src/vm/jit/jit.h - code generation header
 
-   Copyright (C) 1996-2005 R. Grafl, A. Krall, C. Kruegel, C. Oates,
-   R. Obermaisser, M. Platter, M. Probst, S. Ring, E. Steiner,
-   C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich, J. Wenninger,
-   Institut f. Computersprachen - TU Wien
+   Copyright (C) 1996-2005, 2006 R. Grafl, A. Krall, C. Kruegel,
+   C. Oates, R. Obermaisser, M. Platter, M. Probst, S. Ring,
+   E. Steiner, C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich,
+   J. Wenninger, Institut f. Computersprachen - TU Wien
 
    This file is part of CACAO.
 
@@ -19,17 +19,18 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-   02111-1307, USA.
+   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
+   02110-1301, USA.
 
-   Contact: cacao@complang.tuwien.ac.at
+   Contact: cacao@cacaojvm.org
 
    Authors: Andreas Krall
             Reinhard Grafl
 
    Changes: Christian Thalinger
+   			Edwin Steiner
 
-   $Id: jit.h,v 1.46 2005/11/22 23:42:43 twisti Exp $
+   $Id: jit.h 4959 2006-05-26 12:09:29Z edwin $
 
 */
 
@@ -37,24 +38,67 @@
 #ifndef _JIT_H
 #define _JIT_H
 
-/* resolve typedef cycles *****************************************************/
+/* forward typedefs ***********************************************************/
 
+typedef struct jitdata jitdata;
 typedef struct stackelement stackelement;
 typedef stackelement *stackptr;
 typedef struct basicblock basicblock;
+typedef struct branchref branchref;
 typedef struct instruction instruction;
 typedef struct subroutineinfo subroutineinfo;
+typedef struct insinfo_inline insinfo_inline;
 
 
+#include "config.h"
+#include "vm/types.h"
+
 #include "toolbox/chain.h"
 #include "vm/global.h"
-#include "vm/references.h"
 #include "vm/method.h"
-#include "vm/jit/codegen.inc.h"
+#include "vm/references.h"
+#include "vm/statistics.h"
+#include "vm/jit/codegen-common.h"
+#include "vm/jit/reg.h"
+#include "vm/jit/stacktrace.h"
+
+#if defined(ENABLE_INLINING)
+# include "vm/jit/inline/inline.h"
+#endif
+
+#if defined(ENABLE_LOOP)
+# include "vm/jit/loop/loop.h"
+#endif
+
 #include "vm/jit/verify/typeinfo.h"
 
 
-/**************************** resolve typedef-cycles **************************/
+/* common jit/codegen macros **************************************************/
+
+#if defined(ENABLE_STATISTICS)
+# define COUNT(x)        (x)++
+# define COUNT_SPILLS    count_spills++
+#else
+# define COUNT(x)        /* nothing */
+# define COUNT_SPILLS    /* nothing */
+#endif
+
+
+/* jitdata ********************************************************************/
+
+#define JITDATA_FLAG_IFCONV    0x00000001
+
+struct jitdata {
+	methodinfo   *m;                    /* methodinfo of the method compiled  */
+	codeinfo     *code;
+	codegendata  *cd;
+	registerdata *rd;
+#if defined(ENABLE_LOOP)
+	loopdata     *ld;
+#endif
+	u4            flags;                /* contains JIT compiler flags        */
+};
+
 
 /************************** stack element structure ***************************/
 
@@ -72,7 +116,10 @@ typedef struct subroutineinfo subroutineinfo;
 #define IS_INT_LNG_TYPE(a)      (!((a) & TYPE_FLT))
 #define IS_FLT_DBL_TYPE(a)      ((a) & TYPE_FLT)
 #define IS_2_WORD_TYPE(a)       ((a) & TYPE_LNG)
-#define IS_ADR_TYPE(a)          ((a) & TYPE_ADR)
+
+#define IS_INT_TYPE(a)          ((a) == TYPE_INT)
+#define IS_LNG_TYPE(a)          ((a) == TYPE_LNG)
+#define IS_ADR_TYPE(a)          ((a) == TYPE_ADR)
 
 
 /* flags */
@@ -80,7 +127,11 @@ typedef struct subroutineinfo subroutineinfo;
 #define SAVEDVAR   1            /* variable has to survive method invocations */
 #define INMEMORY   2            /* variable stored in memory                  */
 #define SAVEDTMP   4            /* temporary variable using a saved register  */
-#define TMPARG     8            /* temporary variable using a argument register  */
+#define TMPARG     8            /* temporary variable using a arg register    */
+#define STCOPY    16            /* there is another stackslot alive "below"   */
+                                /* using the same register/memory location    */
+#define STKEEP    32            /* to prevent reg_mark_copy to free this      */
+                                /* stackslot */
 
 /* variable kinds */
 
@@ -94,7 +145,7 @@ typedef struct subroutineinfo subroutineinfo;
 struct stackelement {
 	stackptr prev;              /* pointer to next element towards bottom     */
 	s4       type;              /* slot type of stack element                 */
-#ifdef CACAO_TYPECHECK
+#ifdef ENABLE_VERIFIER
 	typeinfo typeinfo;          /* info on reference types                    */
 #endif
 	s4       flags;             /* flags (SAVED, INMEMORY)                    */
@@ -115,12 +166,61 @@ struct instruction {
 	                            /* and as address for list of targets for     */
 	                            /* statements                                 */
 	u2          line;           /* line number in source file                 */
-	methodinfo *method;         /* needed for inlining. can't be done on      */
-	                            /* basic block level, since an inlined        */
-	                            /* function doesn't necessarily start         */
-	                            /* a new block                                */
 };
 
+#define INSTRUCTION_IS_RESOLVED(iptr) \
+	(!((ptrint)(iptr)->target & 0x01)) /* XXX target used temporarily as flag */
+
+#define INSTRUCTION_IS_UNRESOLVED(iptr) \
+	((ptrint)(iptr)->target & 0x01) /* XXX target used temporarily as flag */
+
+#define INSTRUCTION_GET_FIELDREF(iptr,fref) \
+	do { \
+		if (INSTRUCTION_IS_UNRESOLVED(iptr)) \
+			fref = ((unresolved_field *) (iptr)->val.a)->fieldref; \
+		else \
+			fref = ((constant_FMIref *)(iptr)->val.a); \
+	} while (0)
+
+#define INSTRUCTION_GET_METHODREF(iptr,mref) \
+	do { \
+		if (INSTRUCTION_IS_UNRESOLVED(iptr)) \
+			mref = ((unresolved_method *) (iptr)->val.a)->methodref; \
+		else \
+			mref = ((constant_FMIref *)(iptr)->val.a); \
+	} while (0)
+
+#define INSTRUCTION_GET_FIELDDESC(iptr,fd) \
+	do { \
+		if (INSTRUCTION_IS_UNRESOLVED(iptr)) \
+			fd = ((unresolved_field *)(iptr)->val.a)->fieldref->parseddesc.fd; \
+		else \
+			fd = ((constant_FMIref *)(iptr)->val.a)->parseddesc.fd; \
+	} while (0)
+
+#define INSTRUCTION_GET_METHODDESC(iptr,md) \
+	do { \
+		if (INSTRUCTION_IS_UNRESOLVED(iptr)) \
+			md = ((unresolved_method *) (iptr)->val.a)->methodref->parseddesc.md; \
+		else \
+			md = ((constant_FMIref *)(iptr)->val.a)->parseddesc.md; \
+	} while (0)
+
+#define INSTRUCTION_UNRESOLVED_METHOD(iptr) \
+	((unresolved_method *) (iptr)->val.a)
+
+#define INSTRUCTION_UNRESOLVED_FIELD(iptr) \
+	((unresolved_field *) (iptr)->val.a)
+
+#define INSTRUCTION_RESOLVED_FMIREF(iptr) \
+    ((constant_FMIref *)(iptr)->val.a)
+
+#define INSTRUCTION_RESOLVED_FIELDINFO(iptr) \
+    (INSTRUCTION_RESOLVED_FMIREF(iptr)->p.field)
+
+#define INSTRUCTION_RESOLVED_METHODINFO(iptr) \
+    (INSTRUCTION_RESOLVED_FMIREF(iptr)->p.method)
+
 #define INSTRUCTION_PUTCONST_TYPE(iptr) \
 	((iptr)[0].op1)
 
@@ -136,40 +236,101 @@ struct instruction {
 #define INSTRUCTION_PUTCONST_FIELDREF(iptr) \
 	((unresolved_field *)((iptr)[1].target))
 
-/**************************** basic block structure ***************************/
+/* for ICMD_ACONST */
+
+#define ICMD_ACONST_IS_CLASS(iptr) \
+	((ptrint)(iptr)->target & 0x02) /* XXX target used temporarily as flag */
+
+#define ICMD_ACONST_CLASSREF_OR_CLASSINFO(iptr) \
+(CLASSREF_OR_CLASSINFO((iptr)->val.a))
+
+#define ICMD_ACONST_RESOLVED_CLASSINFO(iptr) \
+	((classinfo *) (iptr)->val.a)
+
+#define ICMD_ACONST_UNRESOLVED_CLASSREF(iptr) \
+	((constant_classref *) (iptr)->val.a)
+
+
+/* additional info structs for special instructions ***************************/
+
+/* for ICMD_INLINE_START and ICMD_INLINE_END */
+
+struct insinfo_inline {
+	methodinfo *method;         /* the inlined method starting/ending here    */
+	methodinfo *outer;          /* the outer method suspended/resumed here    */
+	s4          startmpc;       /* machine code offset of start of inlining   */          
+	s4          synclocal;      /* local index used for synchronization       */
+	bool        synchronize;    /* true if synchronization is needed          */
+};
+
+/* basicblock *****************************************************************/
  
-/*                    flags                                                   */
+/* flags */
+
+#define BBDELETED            -2
+#define BBUNDEF              -1
+#define BBREACHED            0
+#define BBFINISHED           1
 
-#define BBDELETED     -2
-#define BBUNDEF       -1
-#define BBREACHED     0
-#define BBFINISHED    1
 #define BBTYPECHECK_UNDEF    2
 #define BBTYPECHECK_REACHED  3
 
-#define BBTYPE_STD 0            /* standard basic block type                  */
-#define BBTYPE_EXH 1            /* exception handler basic block type         */
-#define BBTYPE_SBR 2            /* subroutine basic block type                */
+#define BBTYPE_STD           0  /* standard basic block type                  */
+#define BBTYPE_EXH           1  /* exception handler basic block type         */
+#define BBTYPE_SBR           2  /* subroutine basic block type                */
 
+#define BBFLAG_REPLACEMENT   0x01  /* put a replacement point at the start    */
+
+/* XXX basicblock wastes quite a lot of memory by having four flag fields     */
+/* (flags, bitflags, type and lflags). Probably the last three could be       */
+/* combined without loss of efficiency. The first one could be combined with  */
+/* the others by using bitfields.                                             */
 
 struct basicblock {
-	int          flags;         /* used during stack analysis, init with -1   */
-	int          type;          /* basic block type (std, xhandler, subroutine*/
+	s4           debug_nr;      /* basic block number                         */
+	s4           flags;         /* used during stack analysis, init with -1   */
+	s4           bitflags;      /* OR of BBFLAG_... constants, init with 0    */
+	s4           type;          /* basic block type (std, xhandler, subroutine*/
 	instruction *iinstr;        /* pointer to intermediate code instructions  */
-	int          icount;        /* number of intermediate code instructions   */
-	int          mpc;           /* machine code pc at start of block          */
+	s4           icount;        /* number of intermediate code instructions   */
+	s4           mpc;           /* machine code pc at start of block          */
 	stackptr     instack;       /* stack at begin of basic block              */
 	stackptr     outstack;      /* stack at end of basic block                */
-	int          indepth;       /* stack depth at begin of basic block        */
-	int          outdepth;      /* stack depth end of basic block             */
-	int          pre_count;     /* count of predecessor basic blocks          */
-	struct branchref *branchrefs; /* list of branches to be patched           */
+	s4           indepth;       /* stack depth at begin of basic block        */
+	s4           outdepth;      /* stack depth end of basic block             */
+	s4           pre_count;     /* count of predecessor basic blocks          */
+	branchref   *branchrefs;    /* list of branches to be patched             */
 
 	basicblock  *next;          /* used to build a BB list (instead of array) */
-	int          lflags;        /* used during loop copying, init with 0	  */
+	s4           lflags;        /* used during loop copying, init with 0	  */
 	basicblock  *copied_to;     /* points to the copy of this basic block	  */
                                 /* when loop nodes are copied                 */
-	int debug_nr;
+	stackptr     stack;         /* start of stack array for this block        */
+	                            /* (see doc/stack.txt)                        */
+	methodinfo  *method;        /* method this block belongs to               */
+};
+
+/* macro for initializing newly allocated basicblock:s                        */
+
+#define BASICBLOCK_INIT(bptr,m)                            \
+		do {                                               \
+			bptr->mpc = -1;                                \
+			bptr->flags = -1;                              \
+			bptr->bitflags = 0;                            \
+			bptr->lflags = 0;                              \
+			bptr->type = BBTYPE_STD;                       \
+			bptr->branchrefs = NULL;                       \
+			bptr->pre_count = 0;                           \
+			bptr->method = (m);                            \
+			bptr->debug_nr = (m)->c_debug_nr++;            \
+		} while (0)
+			
+
+/* branchref *****************************************************************/
+
+struct branchref {
+	s4         branchpos;       /* patching position in code segment          */
+	branchref *next;            /* next element in branchref list             */
 };
 
 
@@ -178,6 +339,7 @@ struct basicblock {
 #define ACONST_LOAD     0  /* ACONST_NULL or LDC instruction                  */
 #define ACONST_BUILTIN  1  /* constant argument for a builtin function call   */
 
+
 /********** JavaVM operation codes (sorted) and instruction lengths ***********/
 
 extern char *icmd_names[256];
@@ -244,7 +406,7 @@ extern int jcommandsize[256];
 #define ICMD_IFLE_ICONST      20
 
 	                                    /* order of LOAD instructions must be */
-	                                    /* equal to order of TYPE_XXX defines */
+	                                    /* equal to order of TYPE_* defines   */
 #define JAVA_ILOAD            21
 #define ICMD_ILOAD            21        /* op1 = local variable               */
 
@@ -345,7 +507,7 @@ extern int jcommandsize[256];
 #define ICMD_SALOAD           53
 
 	                                    /* order of STORE instructions must be*/
-	                                    /* equal to order of TYPE_XXX defines */
+	                                    /* equal to order of TYPE_* defines   */
 #define JAVA_ISTORE           54
 #define ICMD_ISTORE           54        /* op1 = local variable               */
 
@@ -783,36 +945,56 @@ extern int jcommandsize[256];
 /* UNDEF203 */
 
 #define ICMD_IASTORECONST     204
-
 #define ICMD_LASTORECONST     205
-
 #define ICMD_FASTORECONST     206
-
 #define ICMD_DASTORECONST     207
-
 #define ICMD_AASTORECONST     208
-
 #define ICMD_BASTORECONST     209
-
 #define ICMD_CASTORECONST     210
-
 #define ICMD_SASTORECONST     211
 
 #define ICMD_PUTSTATICCONST   212
-
 #define ICMD_PUTFIELDCONST    213
 
 #define ICMD_IMULPOW2         214
-
 #define ICMD_LMULPOW2         215
 
+#define ICMD_IF_FCMPEQ        216
+#define ICMD_IF_FCMPNE        217
+
+#define ICMD_IF_FCMPL_LT      218
+#define ICMD_IF_FCMPL_GE      219
+#define ICMD_IF_FCMPL_GT      220
+#define ICMD_IF_FCMPL_LE      221
+
+#define ICMD_IF_FCMPG_LT      222
+#define ICMD_IF_FCMPG_GE      223
+#define ICMD_IF_FCMPG_GT      224
+#define ICMD_IF_FCMPG_LE      225
+
+#define ICMD_IF_DCMPEQ        226
+#define ICMD_IF_DCMPNE        227
+
+#define ICMD_IF_DCMPL_LT      228
+#define ICMD_IF_DCMPL_GE      229
+#define ICMD_IF_DCMPL_GT      230
+#define ICMD_IF_DCMPL_LE      231
+
+#define ICMD_IF_DCMPG_LT      232
+#define ICMD_IF_DCMPG_GE      233
+#define ICMD_IF_DCMPG_GT      234
+#define ICMD_IF_DCMPG_LE      235
+
 #define ICMD_INLINE_START     251       /* instruction before inlined method  */
 #define ICMD_INLINE_END       252       /* instruction after inlined method   */
+#define ICMD_INLINE_GOTO      253       /* jump to caller of inlined method   */
 
 #define ICMD_BUILTIN          255       /* internal opcode                    */
 
-#define ICMD_READONLY_ARG     1024      /* used for inlining, opcodes 1024-1028 are used */
-#define ICMD_CLEAR_ARGREN     1029      /* indicates the start of a new inlined method argument renaming must be reset */
+/* define some ICMD masks *****************************************************/
+
+#define ICMD_OPCODE_MASK      0x00ff    /* mask to get the opcode             */
+#define ICMD_CONDITION_MASK   0xff00    /* mask to get the condition          */
 
 
 /******************* description of JavaVM instructions ***********************/
@@ -850,9 +1032,24 @@ void jit_close(void);
 /* compile a method with jit compiler */
 u1 *jit_compile(methodinfo *m);
 
-/* machine dependent initialization */
+/* patch the method entrypoint */
+u1 *jit_asm_compile(methodinfo *m, u1 *mptr, u1 *sp, u1 *ra);
+
+/* machine dependent functions */
+#if defined(ENABLE_JIT)
 void md_init(void);
 
+u1  *md_get_method_patch_address(u1 *ra, stackframeinfo *sfi, u1 *mptr);
+
+void md_cacheflush(u1 *addr, s4 nbytes);
+void md_icacheflush(u1 *addr, s4 nbytes);
+void md_dcacheflush(u1 *addr, s4 nbytes);
+#endif
+
+#if defined(ENABLE_INTRP)
+void intrp_md_init(void);
+#endif
+
 #endif /* _JIT_H */
 
 
@@ -867,4 +1064,5 @@ void md_init(void);
  * c-basic-offset: 4
  * tab-width: 4
  * End:
+ * vim:noexpandtab:sw=4:ts=4:
  */
