@@ -1,9 +1,9 @@
 /* src/vm/jit/mips/codegen.h - code generation macros and definitions for MIPS
 
-   Copyright (C) 1996-2005 R. Grafl, A. Krall, C. Kruegel, C. Oates,
-   R. Obermaisser, M. Platter, M. Probst, S. Ring, E. Steiner,
-   C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich, J. Wenninger,
-   Institut f. Computersprachen - TU Wien
+   Copyright (C) 1996-2005, 2006 R. Grafl, A. Krall, C. Kruegel,
+   C. Oates, R. Obermaisser, M. Platter, M. Probst, S. Ring,
+   E. Steiner, C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich,
+   J. Wenninger, Institut f. Computersprachen - TU Wien
 
    This file is part of CACAO.
 
@@ -19,14 +19,14 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-   02111-1307, USA.
+   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
+   02110-1301, USA.
 
-   Contact: cacao@complang.tuwien.ac.at
+   Contact: cacao@cacaojvm.org
 
    Authors: Andreas Krall
 
-   $Id: codegen.h,v 1.26 2005/10/21 12:03:33 twisti Exp $
+   $Id: codegen.h 4905 2006-05-11 13:43:55Z twisti $
 
 */
 
@@ -37,6 +37,8 @@
 #include "config.h"
 #include "vm/types.h"
 
+#include "vm/jit/jit.h"
+
 
 /* some defines ***************************************************************/
 
@@ -45,19 +47,10 @@
 
 /* additional functions and macros to generate code ***************************/
 
-#ifdef STATISTICS
-#define COUNT_SPILLS count_spills++
-#else
-#define COUNT_SPILLS
-#endif
-
-
-/* gen_nullptr_check(objreg) */
-
 #define gen_nullptr_check(objreg) \
     if (checknull) { \
-        M_BEQZ((objreg), 0); \
-        codegen_addxnullrefs(cd, mcodeptr); \
+        M_BEQZ(objreg, 0); \
+        codegen_add_nullpointerexception_ref(cd); \
         M_NOP; \
     }
 
@@ -66,193 +59,50 @@
         M_ILD(REG_ITMP3, s1, OFFSET(java_arrayheader, size)); \
         M_CMPULT(s2, REG_ITMP3, REG_ITMP3); \
         M_BEQZ(REG_ITMP3, 0); \
-        codegen_addxboundrefs(cd, mcodeptr, s2); \
+        codegen_add_arrayindexoutofboundsexception_ref(cd, s2); \
         M_NOP; \
     }
 
 #define gen_div_check(r) \
-    M_BEQZ((r), 0); \
-    codegen_addxdivrefs(cd, mcodeptr); \
-    M_NOP;
+    do { \
+        M_BEQZ((r), 0); \
+        codegen_add_arithmeticexception_ref(cd); \
+        M_NOP; \
+    } while (0)
 
 
 /* MCODECHECK(icnt) */
 
 #define MCODECHECK(icnt) \
-	if ((mcodeptr + (icnt)) > cd->mcodeend) \
-        mcodeptr = codegen_increase(cd, (u1 *) mcodeptr)
+    do { \
+        if ((cd->mcodeptr + (icnt) * 4) > cd->mcodeend) \
+            codegen_increase(cd); \
+    } while (0)
 
 
 #define ALIGNCODENOP \
-    if ((int) ((long) mcodeptr & 7)) { \
+    if ((s4) ((ptrint) cd->mcodeptr & 7)) { \
         M_NOP; \
     }
 
 
-/* M_INTMOVE:
-     generates an integer-move from register a to b.
-     if a and b are the same int-register, no code will be generated.
-*/ 
-
 #define M_INTMOVE(a,b) if (a != b) { M_MOV(a, b); }
 
-
-/* M_FLTMOVE:
-    generates a floating-point-move from register a to b.
-    if a and b are the same float-register, no code will be generated
-*/ 
-
-#define M_FLTMOVE(a,b) if (a != b) { M_DMOV(a, b); }
-
-#define M_TFLTMOVE(t,a,b) \
+#define M_FLTMOVE(a,b) \
     do { \
         if ((a) != (b)) \
-            if ((t) == TYPE_DBL) { \
-                M_DMOV(a,b); \
-            } else { \
-                M_FMOV(a,b); \
-            } \
+            M_FMOV(a, b); \
     } while (0)
 
-#define M_TFLD(t,a,b,disp) \
-    if ((t) == TYPE_DBL) { \
-	  M_DLD(a,b,disp); \
-    } else { \
-	  M_FLD(a,b,disp); \
-    }
+#define M_DBLMOVE(a,b) \
+    do { \
+        if ((a) != (b)) \
+            M_DMOV(a, b); \
+    } while (0)
 
-#define M_TFST(t,a,b,disp) \
-    if ((t)==TYPE_DBL) \
-	  {M_DST(a,b,disp);} \
-    else \
-	  {M_FST(a,b,disp);}
-
-#define M_CCFLTMOVE(t1,t2,a,b) \
-	if ((t1)==(t2)) \
-	  {M_TFLTMOVE(t1,a,b);} \
-	else \
-	  if ((t1)==TYPE_DBL) \
-		{M_CVTDF(a,b);} \
-	  else \
-		{M_CVTFD(a,b);}
-
-#define M_CCFLD(t1,t2,a,b,disp) \
-    if ((t1)==(t2)) \
-	  {M_DLD(a,b,disp);} \
-	else { \
-	  M_DLD(REG_FTMP1,b,disp); \
-	  if ((t1)==TYPE_DBL) \
-	    {M_CVTDF(REG_FTMP1,a);} \
-	  else \
-	    {M_CVTFD(REG_FTMP1,a);} \
-	}
-	  
-#define M_CCFST(t1,t2,a,b,disp) \
-    if ((t1)==(t2)) \
-	  {M_DST(a,b,disp);} \
-	else { \
-	  if ((t1)==TYPE_DBL) \
-	    {M_CVTDF(a,REG_FTMP1);} \
-	  else \
-	    {M_CVTFD(a,REG_FTMP1);} \
-	  M_DST(REG_FTMP1,b,disp); \
-	}
-	  
-
-/* var_to_reg_xxx:
-    this function generates code to fetch data from a pseudo-register
-    into a real register. 
-    If the pseudo-register has actually been assigned to a real 
-    register, no code will be emitted, since following operations
-    can use this register directly.
-    
-    v: pseudoregister to be fetched from
-    tempregnum: temporary register to be used if v is actually spilled to ram
-
-    return: the register number, where the operand can be found after 
-            fetching (this wil be either tempregnum or the register
-            number allready given to v)
-*/
-
-#define var_to_reg_int(regnr,v,tempnr) { \
-	if ((v)->flags & INMEMORY) { \
-		COUNT_SPILLS; \
-        M_LLD(tempnr, REG_SP, 8 * (v)->regoff); \
-        regnr = tempnr; \
-	} else regnr = (v)->regoff; \
-}
-
-
-#define var_to_reg_flt(regnr,v,tempnr) { \
-	if ((v)->flags & INMEMORY) { \
-		COUNT_SPILLS; \
-        M_DLD(tempnr, REG_SP, 8 * (v)->regoff); \
-        regnr = tempnr; \
-	} else regnr = (v)->regoff; \
-}
-
-
-/* store_reg_to_var_xxx:
-    This function generates the code to store the result of an operation
-    back into a spilled pseudo-variable.
-    If the pseudo-variable has not been spilled in the first place, this 
-    function will generate nothing.
-    
-    v ............ Pseudovariable
-    tempregnum ... Number of the temporary registers as returned by
-                   reg_of_var.
-*/	
-
-#define store_reg_to_var_int(sptr, tempregnum) {       \
-	if ((sptr)->flags & INMEMORY) {                    \
-		COUNT_SPILLS;                                  \
-		M_LST(tempregnum, REG_SP, 8 * (sptr)->regoff); \
-		}                                              \
-	}
-
-#define store_reg_to_var_flt(sptr, tempregnum) {       \
-	if ((sptr)->flags & INMEMORY) {                    \
-		COUNT_SPILLS;                                  \
-		M_DST(tempregnum, REG_SP, 8 * (sptr)->regoff); \
-		}                                              \
-	}
-
-
-#define M_COPY(from,to) \
-	d = reg_of_var(rd, to, REG_IFTMP); \
-	if ((from->regoff != to->regoff) || \
-	    ((from->flags ^ to->flags) & INMEMORY)) { \
-		if (IS_FLT_DBL_TYPE(from->type)) { \
-			var_to_reg_flt(s1, from, d); \
-			M_TFLTMOVE(from->type, s1, d); \
-			store_reg_to_var_flt(to, d); \
-		} else { \
-			var_to_reg_int(s1, from, d); \
-			M_INTMOVE(s1, d); \
-			store_reg_to_var_int(to, d); \
-		} \
-	}
-
-
-#define ICONST(r,c) \
-    if ((c) >= -32768 && (c) <= 32767) { \
-        M_IADD_IMM(REG_ZERO, (c), (r)); \
-    } else if ((c) >= 0 && (c) <= 0xffff) { \
-        M_OR_IMM(REG_ZERO, (c), (r)); \
-    } else { \
-        disp = dseg_adds4(cd, (c)); \
-        M_ILD((r), REG_PV, disp); \
-    }
-
-#define LCONST(r,c) \
-    if ((c) >= -32768 && (c) <= 32767) { \
-        M_LADD_IMM(REG_ZERO, (c), (r)); \
-    } else if ((c) >= 0 && (c) <= 0xffff) { \
-        M_OR_IMM(REG_ZERO, (c), (r)); \
-    } else { \
-        disp = dseg_adds8(cd, (c)); \
-        M_LLD((r), REG_PV, disp); \
-    }
+#define M_COPY(s,d)                     emit_copy(jd, iptr, (s), (d))
+#define ICONST(r,c)                     emit_iconst(cd, (r), (c))
+#define LCONST(r,c)                     emit_lconst(cd, (r), (c))
 
 
 /* macros to create code ******************************************************/
@@ -269,13 +119,22 @@
 
 
 #define M_ITYPE(op,rs,rt,imm) \
-    *(mcodeptr++) = (((op) << 26) | ((rs) << 21) | ((rt) << 16) | ((imm) & 0xffff))
+    do { \
+        *((u4 *) cd->mcodeptr) = (((op) << 26) | ((rs) << 21) | ((rt) << 16) | ((imm) & 0xffff)); \
+        cd->mcodeptr += 4; \
+    } while (0)
 
 #define M_JTYPE(op,imm) \
-    *(mcodeptr++) = (((op) << 26) | ((off) & 0x3ffffff))
+    do { \
+        *((u4 *) cd->mcodeptr) = (((op) << 26) | ((off) & 0x3ffffff)); \
+        cd->mcodeptr += 4; \
+    } while (0)
 
 #define M_RTYPE(op,rs,rt,rd,sa,fu) \
-    *(mcodeptr++) = (((op) << 26) | ((rs) << 21) | ((rt) << 16) | ((rd) << 11) | ((sa) << 6) | (fu))
+    do { \
+        *((u4 *) cd->mcodeptr) = (((op) << 26) | ((rs) << 21) | ((rt) << 16) | ((rd) << 11) | ((sa) << 6) | (fu)); \
+        cd->mcodeptr += 4; \
+    } while (0)
 
 #define M_FP2(fu, fmt, fs, fd)       M_RTYPE(0x11, fmt,  0, fs, fd, fu)
 #define M_FP3(fu, fmt, fs, ft, fd)   M_RTYPE(0x11, fmt, ft, fs, fd, fu)
