@@ -1,9 +1,9 @@
 /* src/vm/jit/mips/patcher.c - MIPS code patching functions
 
-   Copyright (C) 1996-2005 R. Grafl, A. Krall, C. Kruegel, C. Oates,
-   R. Obermaisser, M. Platter, M. Probst, S. Ring, E. Steiner,
-   C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich, J. Wenninger,
-   Institut f. Computersprachen - TU Wien
+   Copyright (C) 1996-2005, 2006 R. Grafl, A. Krall, C. Kruegel,
+   C. Oates, R. Obermaisser, M. Platter, M. Probst, S. Ring,
+   E. Steiner, C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich,
+   J. Wenninger, Institut f. Computersprachen - TU Wien
 
    This file is part of CACAO.
 
@@ -19,28 +19,27 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-   02111-1307, USA.
+   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
+   02110-1301, USA.
 
-   Contact: cacao@complang.tuwien.ac.at
+   Contact: cacao@cacaojvm.org
 
    Authors: Christian Thalinger
 
    Changes:
 
-   $Id: patcher.c,v 1.16 2005/11/11 11:59:05 twisti Exp $
+   $Id: patcher.c 4640 2006-03-16 17:24:18Z twisti $
 
 */
 
 
-#include <sys/cachectl.h>
-
 #include "config.h"
 #include "vm/types.h"
 
 #include "mm/memory.h"
 #include "native/native.h"
 #include "vm/builtin.h"
+#include "vm/class.h"
 #include "vm/field.h"
 #include "vm/initialize.h"
 #include "vm/options.h"
@@ -100,7 +99,7 @@ bool patcher_get_putstatic(u1 *sp)
 
 	/* check if the field's class is initialized */
 
-	if (!fi->class->initialized) {
+	if (!(fi->class->state & CLASS_INITIALIZED)) {
 		if (!initialize_class(fi->class)) {
 			PATCHER_MONITOREXIT;
 
@@ -120,7 +119,7 @@ bool patcher_get_putstatic(u1 *sp)
 
 	/* synchronize instruction cache */
 
-	cacheflush(ra, 2 * 4, ICACHE);
+	md_icacheflush(ra, 2 * 4);
 
 	/* patch the field value's address */
 
@@ -128,7 +127,7 @@ bool patcher_get_putstatic(u1 *sp)
 
 	/* synchronize data cache */
 
-	cacheflush(pv + disp, SIZEOF_VOID_P, DCACHE);
+	md_dcacheflush(pv + disp, SIZEOF_VOID_P);
 
 	PATCHER_MARK_PATCHED_MONITOREXIT;
 
@@ -214,12 +213,12 @@ bool patcher_get_putfield(u1 *sp)
 	if (opt_showdisassemble) {
 #if SIZEOF_VOID_P == 4
 		if (fi->type == TYPE_LNG) {
-			cacheflush(ra - 2 * 4, 4 * 4, ICACHE);
+			md_icacheflush(ra - 2 * 4, 4 * 4);
 		} else
 #endif
-			cacheflush(ra - 2 * 4, 3 * 4, ICACHE);
+			md_icacheflush(ra - 2 * 4, 3 * 4);
 	} else {
-		cacheflush(ra, 2 * 4, ICACHE);
+		md_icacheflush(ra, 2 * 4);
 	}
 
 	PATCHER_MARK_PATCHED_MONITOREXIT;
@@ -287,7 +286,7 @@ bool patcher_aconst(u1 *sp)
 
 	/* synchronize instruction cache */
 
-	cacheflush(ra, 2 * 4, ICACHE);
+	md_icacheflush(ra, 2 * 4);
 
 	/* patch the classinfo pointer */
 
@@ -295,7 +294,7 @@ bool patcher_aconst(u1 *sp)
 
 	/* synchronize data cache */
 
-	cacheflush(pv + disp, SIZEOF_VOID_P * 1, DCACHE);
+	md_dcacheflush(pv + disp, SIZEOF_VOID_P);
 
 	PATCHER_MARK_PATCHED_MONITOREXIT;
 
@@ -366,7 +365,7 @@ bool patcher_builtin_multianewarray(u1 *sp)
 
 	/* synchronize instruction cache */
 
-	cacheflush(ra, 2 * 4, ICACHE);
+	md_icacheflush(ra, 2 * 4);
 
 	/* patch the classinfo pointer */
 
@@ -374,7 +373,7 @@ bool patcher_builtin_multianewarray(u1 *sp)
 
 	/* synchronize data cache */
 
-	cacheflush(pv + disp, SIZEOF_VOID_P, DCACHE);
+	md_dcacheflush(pv + disp, SIZEOF_VOID_P);
 
 	PATCHER_MARK_PATCHED_MONITOREXIT;
 
@@ -444,7 +443,7 @@ bool patcher_builtin_arraycheckcast(u1 *sp)
 
 	/* synchronize instruction cache */
 
-	cacheflush(ra, 2 * 4, ICACHE);
+	md_icacheflush(ra, 2 * 4);
 
 	/* patch the classinfo pointer */
 
@@ -452,7 +451,7 @@ bool patcher_builtin_arraycheckcast(u1 *sp)
 
 	/* synchronize data cache */
 
-	cacheflush(pv + disp, SIZEOF_VOID_P * 1, DCACHE);
+	md_dcacheflush(pv + disp, SIZEOF_VOID_P);
 
 	PATCHER_MARK_PATCHED_MONITOREXIT;
 
@@ -521,7 +520,7 @@ bool patcher_invokestatic_special(u1 *sp)
 
 	/* synchronize instruction cache */
 
-	cacheflush(ra, 2 * 4, ICACHE);
+	md_icacheflush(ra, 2 * 4);
 
 	/* patch stubroutine */
 
@@ -529,7 +528,7 @@ bool patcher_invokestatic_special(u1 *sp)
 
 	/* synchronize data cache */
 
-	cacheflush(pv + disp, SIZEOF_VOID_P, DCACHE);
+	md_dcacheflush(pv + disp, SIZEOF_VOID_P);
 
 	PATCHER_MARK_PATCHED_MONITOREXIT;
 
@@ -606,9 +605,9 @@ bool patcher_invokevirtual(u1 *sp)
 	/* synchronize instruction cache */
 
 	if (opt_showdisassemble)
-		cacheflush(ra - 2 * 4, 4 * 4, ICACHE);
+		md_icacheflush(ra - 2 * 4, 4 * 4);
 	else
-		cacheflush(ra, 2 * 4, ICACHE);
+		md_icacheflush(ra, 2 * 4);
 
 	PATCHER_MARK_PATCHED_MONITOREXIT;
 
@@ -691,9 +690,9 @@ bool patcher_invokeinterface(u1 *sp)
 	/* synchronize instruction cache */
 
 	if (opt_showdisassemble)
-		cacheflush(ra - 2 * 4, 5 * 4, ICACHE);
+		md_icacheflush(ra - 2 * 4, 5 * 4);
 	else
-		cacheflush(ra, 3 * 4, ICACHE);
+		md_icacheflush(ra, 3 * 4);
 
 	PATCHER_MARK_PATCHED_MONITOREXIT;
 
@@ -763,7 +762,7 @@ bool patcher_checkcast_instanceof_flags(u1 *sp)
 
 	/* synchronize instruction cache */
 
-	cacheflush(ra, 2 * 4, ICACHE);
+	md_icacheflush(ra, 2 * 4);
 
 	/* patch class flags */
 
@@ -771,7 +770,7 @@ bool patcher_checkcast_instanceof_flags(u1 *sp)
 
 	/* synchronize data cache */
 
-	cacheflush(pv + disp, sizeof(s4), DCACHE);
+	md_dcacheflush(pv + disp, sizeof(s4));
 
 	PATCHER_MARK_PATCHED_MONITOREXIT;
 
@@ -852,9 +851,9 @@ bool patcher_checkcast_instanceof_interface(u1 *sp)
 	/* synchronize instruction cache */
 
 	if (opt_showdisassemble)
-		cacheflush(ra - 2 * 4, 8 * 4, ICACHE);
+		md_icacheflush(ra - 2 * 4, 8 * 4);
 	else
-		cacheflush(ra, 6 * 4, ICACHE);
+		md_icacheflush(ra, 6 * 4);
 
 	PATCHER_MARK_PATCHED_MONITOREXIT;
 
@@ -922,7 +921,7 @@ bool patcher_checkcast_instanceof_class(u1 *sp)
 
 	/* synchronize instruction cache */
 
-	cacheflush(ra, 2 * 4, ICACHE);
+	md_icacheflush(ra, 2 * 4);
 
 	/* patch super class' vftbl */
 
@@ -930,7 +929,7 @@ bool patcher_checkcast_instanceof_class(u1 *sp)
 
 	/* synchronize data cache */
 
-	cacheflush(pv + disp, SIZEOF_VOID_P, DCACHE);
+	md_dcacheflush(pv + disp, SIZEOF_VOID_P);
 
 	PATCHER_MARK_PATCHED_MONITOREXIT;
 
@@ -971,7 +970,7 @@ bool patcher_clinit(u1 *sp)
 
 	/* check if the class is initialized */
 
-	if (!c->initialized) {
+	if (!(c->state & CLASS_INITIALIZED)) {
 		if (!initialize_class(c)) {
 			PATCHER_MONITOREXIT;
 
@@ -991,7 +990,7 @@ bool patcher_clinit(u1 *sp)
 
 	/* synchronize instruction cache */
 
-	cacheflush(ra, 2 * 4, ICACHE);
+	md_icacheflush(ra, 2 * 4);
 
 	PATCHER_MARK_PATCHED_MONITOREXIT;
 
@@ -1007,6 +1006,7 @@ bool patcher_clinit(u1 *sp)
 
 *******************************************************************************/
 
+#ifdef ENABLE_VERIFIER
 bool patcher_athrow_areturn(u1 *sp)
 {
 	u1                *ra;
@@ -1053,12 +1053,13 @@ bool patcher_athrow_areturn(u1 *sp)
 
 	/* synchronize instruction cache */
 
-	cacheflush(ra, 2 * 4, ICACHE);
+	md_icacheflush(ra, 2 * 4);
 
 	PATCHER_MARK_PATCHED_MONITOREXIT;
 
 	return true;
 }
+#endif /* ENABLE_VERIFIER */
 
 
 /* patcher_resolve_native ******************************************************
@@ -1067,7 +1068,7 @@ bool patcher_athrow_areturn(u1 *sp)
 
 *******************************************************************************/
 
-#if !defined(ENABLE_STATICVM)
+#if !defined(WITH_STATIC_CLASSPATH)
 bool patcher_resolve_native(u1 *sp)
 {
 	u1                *ra;
@@ -1123,7 +1124,7 @@ bool patcher_resolve_native(u1 *sp)
 
 	/* synchronize instruction cache */
 
-	cacheflush(ra, 2 * 4, ICACHE);
+	md_icacheflush(ra, 2 * 4);
 
 	/* patch native function pointer */
 
@@ -1131,13 +1132,13 @@ bool patcher_resolve_native(u1 *sp)
 
 	/* synchronize data cache */
 
-	cacheflush(pv + disp, SIZEOF_VOID_P, DCACHE);
+	md_dcacheflush(pv + disp, SIZEOF_VOID_P);
 
 	PATCHER_MARK_PATCHED_MONITOREXIT;
 
 	return true;
 }
-#endif /* !defined(ENABLE_STATICVM) */
+#endif /* !defined(WITH_STATIC_CLASSPATH) */
 
 
 /*
