@@ -1,10 +1,10 @@
 /* src/vm/jit/powerpc/codegen.h - code generation macros and definitions for
-                              32-bit PowerPC
+                                  32-bit PowerPC
 
-   Copyright (C) 1996-2005 R. Grafl, A. Krall, C. Kruegel, C. Oates,
-   R. Obermaisser, M. Platter, M. Probst, S. Ring, E. Steiner,
-   C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich, J. Wenninger,
-   Institut f. Computersprachen - TU Wien
+   Copyright (C) 1996-2005, 2006 R. Grafl, A. Krall, C. Kruegel,
+   C. Oates, R. Obermaisser, M. Platter, M. Probst, S. Ring,
+   E. Steiner, C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich,
+   J. Wenninger, Institut f. Computersprachen - TU Wien
 
    This file is part of CACAO.
 
@@ -20,10 +20,10 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-   02111-1307, USA.
+   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
+   02110-1301, USA.
 
-   Contact: cacao@complang.tuwien.ac.at
+   Contact: cacao@cacaojvm.org
 
    Authors: Andreas Krall
             Stefan Ring
@@ -31,7 +31,7 @@
    Changes: Christian Thalinger
             Christian Ullrich
 
-   $Id: codegen.h,v 1.23 2005/11/17 19:01:00 twisti Exp $
+   $Id: codegen.h 4898 2006-05-10 15:51:46Z twisti $
 
 */
 
@@ -50,21 +50,13 @@
 
 /* additional functions and macros to generate code ***************************/
 
-
-#if defined(STATISTICS)
-#define COUNT_SPILLS count_spills++
-#else
-#define COUNT_SPILLS
-#endif
-
-
 /* gen_nullptr_check(objreg) */
 
 #define gen_nullptr_check(objreg) \
     if (checknull) { \
         M_TST((objreg)); \
         M_BEQ(0); \
-        codegen_addxnullrefs(cd, mcodeptr); \
+        codegen_add_nullpointerexception_ref(cd); \
     }
 
 #define gen_bound_check \
@@ -72,15 +64,17 @@
         M_ILD(REG_ITMP3, s1, OFFSET(java_arrayheader, size));\
         M_CMPU(s2, REG_ITMP3);\
         M_BGE(0);\
-        codegen_addxboundrefs(cd, mcodeptr, s2); \
+        codegen_add_arrayindexoutofboundsexception_ref(cd, s2); \
     }
 
 
 /* MCODECHECK(icnt) */
 
 #define MCODECHECK(icnt) \
-	if ((mcodeptr + (icnt)) > cd->mcodeend) \
-        mcodeptr = codegen_increase(cd, (u1 *) mcodeptr)
+    do { \
+        if ((cd->mcodeptr + (icnt) * 4) > cd->mcodeend) \
+            codegen_increase(cd); \
+    } while (0)
 
 
 /* M_INTMOVE:
@@ -88,18 +82,24 @@
      if a and b are the same int-register, no code will be generated.
 */ 
 
-#define M_INTMOVE(a,b) if ((a) != (b)) { M_MOV(a, b); }
+#define M_INTMOVE(a,b) \
+    do { \
+        if ((a) != (b)) { \
+            M_MOV(a, b); \
+        } \
+    } while (0)
 
-#define M_TINTMOVE(t,a,b) \
-    if ((t) == TYPE_LNG) { \
-        if ((a) <= (b)) \
-            M_INTMOVE(GET_LOW_REG((a)), GET_LOW_REG((b))); \
-        M_INTMOVE(GET_HIGH_REG((a)), GET_HIGH_REG((b))); \
-        if ((a) > (b)) \
-            M_INTMOVE(GET_LOW_REG((a)), GET_LOW_REG((b))); \
-    } else { \
-        M_INTMOVE((a), (b)); \
-    }
+#define M_LNGMOVE(a,b) \
+    do { \
+        if (GET_HIGH_REG(a) == GET_LOW_REG(b)) { \
+            assert((GET_LOW_REG(a) != GET_HIGH_REG(b))); \
+            M_INTMOVE(GET_HIGH_REG(a), GET_HIGH_REG(b)); \
+            M_INTMOVE(GET_LOW_REG(a), GET_LOW_REG(b)); \
+        } else { \
+            M_INTMOVE(GET_LOW_REG(a), GET_LOW_REG(b)); \
+            M_INTMOVE(GET_HIGH_REG(a), GET_HIGH_REG(b)); \
+        } \
+    } while (0)
 
 
 /* M_FLTMOVE:
@@ -107,149 +107,24 @@
     if a and b are the same float-register, no code will be generated
 */ 
 
-#define M_FLTMOVE(a,b) if ((a) != (b)) { M_FMOV(a, b); }
+#define M_FLTMOVE(a,b) \
+    do { \
+        if ((a) != (b)) { \
+            M_FMOV(a, b); \
+        } \
+    } while (0)
 
 
-/* var_to_reg_xxx:
-    this function generates code to fetch data from a pseudo-register
-    into a real register. 
-    If the pseudo-register has actually been assigned to a real 
-    register, no code will be emitted, since following operations
-    can use this register directly.
-    
-    v: pseudoregister to be fetched from
-    tempregnum: temporary register to be used if v is actually spilled to ram
-
-    return: the register number, where the operand can be found after 
-            fetching (this wil be either tempregnum or the register
-            number allready given to v)
-*/
-
-#define var_to_reg_int(regnr,v,tempnr) \
-	do { \
-		if ((v)->flags & INMEMORY) { \
-			COUNT_SPILLS; \
-			if (IS_2_WORD_TYPE((v)->type)) { \
-				M_ILD(GET_HIGH_REG((tempnr)), REG_SP, (v)->regoff * 4); \
-				M_ILD(GET_LOW_REG((tempnr)), REG_SP, (v)->regoff * 4 + 4); \
-			} else \
-				M_ILD((tempnr), REG_SP, (v)->regoff * 4); \
-			regnr = tempnr; \
-		} else { \
-			regnr = (v)->regoff; \
-		} \
-	} while(0)
-
-
-/* fetch only the low part of v, regnr hast to be a single register */
-
-#define var_to_reg_lng_low(regnr,v,tempnr) \
-	do { \
-		if ((v)->flags & INMEMORY) { \
-			COUNT_SPILLS; \
-			M_ILD((tempnr), REG_SP, (v)->regoff * 4 + 4); \
-			regnr = tempnr; \
-		} else { \
-			regnr = GET_LOW_REG((v)->regoff); \
-		} \
-	} while(0)
-
-
-/* fetch only the high part of v, regnr hast to be a single register */
-
-#define var_to_reg_lng_high(regnr,v,tempnr) \
-	do { \
-		if ((v)->flags & INMEMORY) { \
-			COUNT_SPILLS; \
-			M_ILD((tempnr), REG_SP, (v)->regoff * 4); \
-			regnr = tempnr; \
-		} else { \
-			regnr = GET_HIGH_REG((v)->regoff); \
-		} \
-	} while(0)
-
-
-
-#define var_to_reg_flt(regnr,v,tempnr) \
-	do { \
-		if ((v)->flags & INMEMORY) { \
-			COUNT_SPILLS; \
-			if ((v)->type == TYPE_DBL) \
-				M_DLD(tempnr, REG_SP, (v)->regoff * 4); \
-			else \
-				M_FLD(tempnr, REG_SP, (v)->regoff * 4); \
-			regnr = tempnr; \
-		} else { \
-			regnr = (v)->regoff; \
-		} \
-	} while (0)
-
-
-/* store_reg_to_var_xxx:
-    This function generates the code to store the result of an operation
-    back into a spilled pseudo-variable.
-    If the pseudo-variable has not been spilled in the first place, this 
-    function will generate nothing.
-    
-    v ............ Pseudovariable
-    tempregnum ... Number of the temporary registers as returned by
-                   reg_of_var.
-*/	
-#define store_reg_to_var_int0(sptr, tempregnum, a, b) {       \
-	if ((sptr)->flags & INMEMORY) {                    \
-		COUNT_SPILLS;                                  \
-		if (a) M_IST(GET_HIGH_REG((tempregnum)), REG_SP, 4 * (sptr)->regoff); \
-		if ((b) && IS_2_WORD_TYPE((sptr)->type)) \
-			M_IST(GET_LOW_REG((tempregnum)), REG_SP, 4 * (sptr)->regoff + 4); \
-		}                                              \
-	}
-
-#define store_reg_to_var_int(sptr, tempregnum) \
-	store_reg_to_var_int0(sptr, tempregnum, 1, 1)
-
-#define store_reg_to_var_flt(sptr, tempregnum) {       \
-	if ((sptr)->flags & INMEMORY) {                    \
-		COUNT_SPILLS;                                  \
-		if ((sptr)->type==TYPE_DBL) \
-			M_DST(tempregnum, REG_SP, 4 * (sptr)->regoff); \
-		else \
-			M_FST(tempregnum, REG_SP, 4 * (sptr)->regoff); \
-		}                                              \
-	}
-
-
-#define ICONST(reg,c) \
-    if (((c) >= 0 && (c) <= 32767) || ((c) >= -32768 && (c) < 0)) {\
-        M_LDA((reg), REG_ZERO, (c)); \
-    } else { \
-        a = dseg_adds4(cd, c); \
-        M_ILD((reg), REG_PV, a); \
-    }
+#define M_COPY(s,d)                     emit_copy(jd, iptr, (s), (d))
+#define ICONST(d,c)                     emit_iconst(cd, (d), (c))
 
 #define LCONST(reg,c) \
     ICONST(GET_HIGH_REG((reg)), (s4) ((s8) (c) >> 32));	\
     ICONST(GET_LOW_REG((reg)), (s4) ((s8) (c)));
 
 
-#define M_COPY(from,to) \
-			d = reg_of_var(rd, to, REG_IFTMP); \
-			if ((from->regoff != to->regoff) || \
-			    ((from->flags ^ to->flags) & INMEMORY)) { \
-				if (IS_FLT_DBL_TYPE(from->type)) { \
-					var_to_reg_flt(s1, from, d); \
-					M_FLTMOVE(s1,d); \
-					store_reg_to_var_flt(to, d); \
-					}\
-				else { \
-					var_to_reg_int(s1, from, d); \
-					M_TINTMOVE(from->type,s1,d); \
-					store_reg_to_var_int(to, d); \
-					}\
-				}
-
-
 #define ALIGNCODENOP \
-    if ((s4) ((ptrint) mcodeptr & 7)) { \
+    if ((s4) ((ptrint) cd->mcodeptr & 7)) { \
         M_NOP; \
     }
 
@@ -257,22 +132,37 @@
 /* macros to create code ******************************************************/
 
 #define M_OP3(opcode,y,oe,rc,d,a,b) \
-	*(mcodeptr++) = (((opcode) << 26) | ((d) << 21) | ((a) << 16) | ((b) << 11) | ((oe) << 10) | ((y) << 1) | (rc))
+    do { \
+        *((u4 *) cd->mcodeptr) = (((opcode) << 26) | ((d) << 21) | ((a) << 16) | ((b) << 11) | ((oe) << 10) | ((y) << 1) | (rc)); \
+        cd->mcodeptr += 4; \
+    } while (0)
 
 #define M_OP4(x,y,rc,d,a,b,c) \
-	*(mcodeptr++) = (((x) << 26) | ((d) << 21) | ((a) << 16) | ((b) << 11) | ((c) << 6) | ((y) << 1) | (rc))
+    do { \
+        *((u4 *) cd->mcodeptr) = (((x) << 26) | ((d) << 21) | ((a) << 16) | ((b) << 11) | ((c) << 6) | ((y) << 1) | (rc)); \
+        cd->mcodeptr += 4; \
+    } while (0)
 
 #define M_OP2_IMM(x,d,a,i) \
-	*(mcodeptr++) = (((x) << 26) | ((d) << 21) | ((a) << 16) | ((i) & 0xffff))
+    do { \
+        *((u4 *) cd->mcodeptr) = (((x) << 26) | ((d) << 21) | ((a) << 16) | ((i) & 0xffff)); \
+        cd->mcodeptr += 4; \
+    } while (0)
 
 #define M_BRMASK     0x0000fffc                     /* (((1 << 16) - 1) & ~3) */
 #define M_BRAMASK    0x03fffffc                     /* (((1 << 26) - 1) & ~3) */
 
 #define M_BRA(x,i,a,l) \
-	*(mcodeptr++) = (((x) << 26) | ((((i) * 4) + 4) & M_BRAMASK) | ((a) << 1) | (l))
+    do { \
+        *((u4 *) cd->mcodeptr) = (((x) << 26) | ((((i) * 4) + 4) & M_BRAMASK) | ((a) << 1) | (l)); \
+        cd->mcodeptr += 4; \
+    } while (0)
 
 #define M_BRAC(x,bo,bi,i,a,l) \
-	*(mcodeptr++) = (((x) << 26) | ((bo) << 21) | ((bi) << 16) | (((i) * 4 + 4) & M_BRMASK) | ((a) << 1) | (l))
+    do { \
+        *((u4 *) cd->mcodeptr) = (((x) << 26) | ((bo) << 21) | ((bi) << 16) | (((i) * 4 + 4) & M_BRMASK) | ((a) << 1) | (l)); \
+        cd->mcodeptr += 4; \
+    } while (0)
 
 
 /* instruction macros *********************************************************/
@@ -398,6 +288,24 @@
         } \
     } while (0)
 
+#define M_LLD_INTERN(a,b,disp) \
+    do { \
+        M_ILD_INTERN(GET_HIGH_REG(a), b, disp); \
+        M_ILD_INTERN(GET_LOW_REG(a), b, disp + 4); \
+    } while (0)
+
+#define M_LLD(a,b,disp) \
+    do { \
+        s4 lo = (short) (disp); \
+        s4 hi = (short) (((disp) - lo) >> 16); \
+        if (hi == 0) { \
+            M_LLD_INTERN(a,b,lo); \
+        } else { \
+            M_ADDIS(b,hi,a); \
+            M_LLD_INTERN(a,GET_LOW_REG(a),lo); \
+        } \
+    } while (0)
+
 #define M_ALD_INTERN(a,b,disp)          M_ILD_INTERN(a,b,disp)
 #define M_ALD(a,b,disp)                 M_ILD(a,b,disp)
 
@@ -406,9 +314,9 @@
 
 #define M_IST_INTERN(a,b,disp)          M_OP2_IMM(36,a,b,disp)
 
-/* Stores with displacement overflow should only happen with PUTFIELD or on   */
-/* the stack. The PUTFIELD instruction does not use REG_ITMP3 and a           */
-/* reg_of_var call should not use REG_ITMP3!!!                                */
+/* Stores with displacement overflow should only happen with PUTFIELD
+   or on the stack. The PUTFIELD instruction does not use REG_ITMP3
+   and a reg_of_var call should not use REG_ITMP3!!! */
 
 #define M_IST(a,b,disp) \
     do { \
@@ -422,6 +330,24 @@
         } \
     } while (0)
 
+#define M_LST_INTERN(a,b,disp) \
+    do { \
+        M_IST_INTERN(GET_HIGH_REG(a), b, disp); \
+        M_IST_INTERN(GET_LOW_REG(a), b, disp + 4); \
+    } while (0)
+
+#define M_LST(a,b,disp) \
+    do { \
+        s4 lo = (short) (disp); \
+        s4 hi = (short) (((disp) - lo) >> 16); \
+        if (hi == 0) { \
+            M_LST_INTERN(a,b,lo); \
+        } else { \
+            M_ADDIS(b,hi,REG_ITMP3); \
+            M_LST_INTERN(a,REG_ITMP3, lo); \
+        } \
+    } while (0)
+
 #define M_AST_INTERN(a,b,disp)          M_IST_INTERN(a,b,disp)
 #define M_AST(a,b,disp)                 M_IST(a,b,disp)
 
@@ -540,12 +466,6 @@
 #define gen_resolvebranch(ip,so,to) \
 	*((s4*)(ip)-1)=(*((s4*)(ip)-1) & ~M_BRMASK) | (((s4)((to)-(so))+4)&((((*((s4*)(ip)-1)>>26)&63)==18)?M_BRAMASK:M_BRMASK))
 
-
-/* function prototypes */
-
-void preregpass(methodinfo *m, registerdata *rd);
-void docacheflush(u1 *p, long bytelen);
-
 #endif /* _CODEGEN_H */
 
 
