@@ -1,9 +1,9 @@
 /* src/vm/jit/powerpc/patcher.c - PowerPC code patching functions
 
-   Copyright (C) 1996-2005 R. Grafl, A. Krall, C. Kruegel, C. Oates,
-   R. Obermaisser, M. Platter, M. Probst, S. Ring, E. Steiner,
-   C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich, J. Wenninger,
-   Institut f. Computersprachen - TU Wien
+   Copyright (C) 1996-2005, 2006 R. Grafl, A. Krall, C. Kruegel,
+   C. Oates, R. Obermaisser, M. Platter, M. Probst, S. Ring,
+   E. Steiner, C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich,
+   J. Wenninger, Institut f. Computersprachen - TU Wien
 
    This file is part of CACAO.
 
@@ -19,16 +19,16 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-   02111-1307, USA.
+   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
+   02110-1301, USA.
 
-   Contact: cacao@complang.tuwien.ac.at
+   Contact: cacao@cacaojvm.org
 
    Authors: Christian Thalinger
 
    Changes:
 
-   $Id: patcher.c,v 1.14 2005/11/07 22:31:03 twisti Exp $
+   $Id: patcher.c 4714 2006-03-31 07:14:10Z twisti $
 
 */
 
@@ -39,6 +39,7 @@
 #include "mm/memory.h"
 #include "native/native.h"
 #include "vm/builtin.h"
+#include "vm/class.h"
 #include "vm/field.h"
 #include "vm/initialize.h"
 #include "vm/options.h"
@@ -89,7 +90,7 @@ bool patcher_get_putstatic(u1 *sp)
 
 	/* check if the field's class is initialized */
 
-	if (!fi->class->initialized) {
+	if (!(fi->class->state & CLASS_INITIALIZED)) {
 		if (!initialize_class(fi->class)) {
 			PATCHER_MONITOREXIT;
 
@@ -103,12 +104,16 @@ bool patcher_get_putstatic(u1 *sp)
 
 	/* synchronize instruction cache */
 
-	asm_cacheflush(ra, 4);
+	md_icacheflush(ra, 4);
 
 	/* patch the field value's address */
 
 	*((ptrint *) (pv + disp)) = (ptrint) &(fi->value);
 
+	/* synchronize data cache */
+
+	md_dcacheflush(pv + disp, SIZEOF_VOID_P);
+
 	PATCHER_MARK_PATCHED_MONITOREXIT;
 
 	return true;
@@ -161,18 +166,34 @@ bool patcher_get_putfield(u1 *sp)
 	/* patch the field's offset */
 
 	if (fi->type == TYPE_LNG) {
-		/* if the field has type long, we have to patch two instructions */
+		s2 disp;
 
-		*((u4 *) ra) |= (s2) ((fi->offset + 4) & 0x0000ffff);
-		*((u4 *) (ra + 4)) |= (s2) (fi->offset & 0x0000ffff);
+		/* If the field has type long, we have to patch two
+		   instructions.  But we have to check which instruction is
+		   first.  We do that with the offset of the first
+		   instruction. */
 
+		disp = *((u4 *) (ra + 0));
+
+#if WORDS_BIGENDIAN == 1
+		if (disp == 4) {
+			*((u4 *) (ra + 0)) |= (s2) ((fi->offset + 4) & 0x0000ffff);
+			*((u4 *) (ra + 4)) |= (s2) ((fi->offset + 0) & 0x0000ffff);
+
+		} else {
+			*((u4 *) (ra + 0)) |= (s2) ((fi->offset + 0) & 0x0000ffff);
+			*((u4 *) (ra + 4)) |= (s2) ((fi->offset + 4) & 0x0000ffff);
+		}
+#else
+#error Fix me for LE
+#endif
 	} else {
 		*((u4 *) ra) |= (s2) (fi->offset & 0x0000ffff);
 	}
 
 	/* synchronize instruction cache */
 
-	asm_cacheflush(ra, 8);
+	md_icacheflush(ra, 8);
 
 	PATCHER_MARK_PATCHED_MONITOREXIT;
 
@@ -227,12 +248,16 @@ bool patcher_aconst(u1 *sp)
 
 	/* synchronize instruction cache */
 
-	asm_cacheflush(ra, 4);
+	md_icacheflush(ra, 4);
 
 	/* patch the classinfo pointer */
 
 	*((ptrint *) (pv + disp)) = (ptrint) c;
 
+	/* synchronize data cache */
+
+	md_dcacheflush(pv + disp, SIZEOF_VOID_P);
+
 	PATCHER_MARK_PATCHED_MONITOREXIT;
 
 	return true;
@@ -287,12 +312,16 @@ bool patcher_builtin_multianewarray(u1 *sp)
 
 	/* synchronize instruction cache */
 
-	asm_cacheflush(ra, 4);
+	md_icacheflush(ra, 4);
 
 	/* patch the classinfo pointer */
 
 	*((ptrint *) (pv + disp)) = (ptrint) c;
 
+	/* synchronize data cache */
+
+	md_dcacheflush(pv + disp, SIZEOF_VOID_P);
+
 	PATCHER_MARK_PATCHED_MONITOREXIT;
 
 	return true;
@@ -346,12 +375,16 @@ bool patcher_builtin_arraycheckcast(u1 *sp)
 
 	/* synchronize instruction cache */
 
-	asm_cacheflush(ra, 4);
+	md_icacheflush(ra, 4);
 
 	/* patch the classinfo pointer */
 
 	*((ptrint *) (pv + disp)) = (ptrint) c;
 
+	/* synchronize data cache */
+
+	md_dcacheflush(pv + disp, SIZEOF_VOID_P);
+
 	PATCHER_MARK_PATCHED_MONITOREXIT;
 
 	return true;
@@ -404,12 +437,16 @@ bool patcher_invokestatic_special(u1 *sp)
 
 	/* synchronize instruction cache */
 
-	asm_cacheflush(ra, 4);
+	md_icacheflush(ra, 4);
 
 	/* patch stubroutine */
 
 	*((ptrint *) (pv + disp)) = (ptrint) m->stubroutine;
 
+	/* synchronize data cache */
+
+	md_dcacheflush(pv + disp, SIZEOF_VOID_P);
+
 	PATCHER_MARK_PATCHED_MONITOREXIT;
 
 	return true;
@@ -435,6 +472,7 @@ bool patcher_invokevirtual(u1 *sp)
 	u4                 mcode;
 	unresolved_method *um;
 	methodinfo        *m;
+	s4                 disp;
 
 	/* get stuff from the stack */
 
@@ -464,12 +502,13 @@ bool patcher_invokevirtual(u1 *sp)
 
 	/* patch vftbl index */
 
-	*((s4 *) (ra + 4)) |= (s4) ((OFFSET(vftbl_t, table[0]) +
-								 sizeof(methodptr) * m->vftblindex) & 0x0000ffff);
+	disp = (OFFSET(vftbl_t, table[0]) + sizeof(methodptr) * m->vftblindex);
+
+	*((s4 *) (ra + 4)) |= (disp & 0x0000ffff);
 
 	/* synchronize instruction cache */
 
-	asm_cacheflush(ra, 2 * 4);
+	md_icacheflush(ra, 2 * 4);
 
 	PATCHER_MARK_PATCHED_MONITOREXIT;
 
@@ -497,6 +536,7 @@ bool patcher_invokeinterface(u1 *sp)
 	u4                 mcode;
 	unresolved_method *um;
 	methodinfo        *m;
+	s4                 disp;
 
 	/* get stuff from the stack */
 
@@ -526,17 +566,19 @@ bool patcher_invokeinterface(u1 *sp)
 
 	/* patch interfacetable index */
 
-	*((s4 *) (ra + 1 * 4)) |= (s4) ((OFFSET(vftbl_t, interfacetable[0]) -
-								 sizeof(methodptr*) * m->class->index) & 0x0000ffff);
+	disp = (OFFSET(vftbl_t, interfacetable[0]) - sizeof(methodptr*) * m->class->index);
+
+	*((s4 *) (ra + 1 * 4)) |= (disp & 0x0000ffff);
 
 	/* patch method offset */
 
-	*((s4 *) (ra + 2 * 4)) |=
-		(s4) ((sizeof(methodptr) * (m - m->class->methods)) & 0x0000ffff);
+	disp = (sizeof(methodptr) * (m - m->class->methods));
+
+	*((s4 *) (ra + 2 * 4)) |= (disp & 0x0000ffff);
 
 	/* synchronize instruction cache */
 
-	asm_cacheflush(ra, 3 * 4);
+	md_icacheflush(ra, 3 * 4);
 
 	PATCHER_MARK_PATCHED_MONITOREXIT;
 
@@ -588,12 +630,16 @@ bool patcher_checkcast_instanceof_flags(u1 *sp)
 
 	/* synchronize instruction cache */
 
-	asm_cacheflush(ra, 4);
+	md_icacheflush(ra, 4);
 
 	/* patch class flags */
 
 	*((s4 *) (pv + disp)) = (s4) c->flags;
 
+	/* synchronize data cache */
+
+	md_dcacheflush(pv + disp, SIZEOF_VOID_P);
+
 	PATCHER_MARK_PATCHED_MONITOREXIT;
 
 	return true;
@@ -620,6 +666,7 @@ bool patcher_checkcast_instanceof_interface(u1 *sp)
 	u4                 mcode;
 	constant_classref *cr;
 	classinfo         *c;
+	s4                 disp;
 
 	/* get stuff from the stack */
 
@@ -649,14 +696,17 @@ bool patcher_checkcast_instanceof_interface(u1 *sp)
 
 	/* patch super class index */
 
-	*((s4 *) (ra + 2 * 4)) |= (s4) (-(c->index) & 0x0000ffff);
+	disp = -(c->index);
 
-	*((s4 *) (ra + 4 * 4)) |= (s4) ((OFFSET(vftbl_t, interfacetable[0]) -
-									 c->index * sizeof(methodptr*)) & 0x0000ffff);
+	*((s4 *) (ra + 2 * 4)) |= (disp & 0x0000ffff);
+
+	disp = OFFSET(vftbl_t, interfacetable[0]) - c->index * sizeof(methodptr*);
+
+	*((s4 *) (ra + 4 * 4)) |= (disp & 0x0000ffff);
 
 	/* synchronize instruction cache */
 
-	asm_cacheflush(ra, 5 * 4);
+	md_icacheflush(ra, 5 * 4);
 
 	PATCHER_MARK_PATCHED_MONITOREXIT;
 
@@ -710,12 +760,16 @@ bool patcher_checkcast_class(u1 *sp)
 
 	/* synchronize instruction cache */
 
-	asm_cacheflush(ra, 4);
+	md_icacheflush(ra, 4);
 
 	/* patch super class' vftbl */
 
 	*((ptrint *) (pv + disp)) = (ptrint) c->vftbl;
 
+	/* synchronize data cache */
+
+	md_dcacheflush(pv + disp, SIZEOF_VOID_P);
+
 	PATCHER_MARK_PATCHED_MONITOREXIT;
 
 	return true;
@@ -767,12 +821,16 @@ bool patcher_instanceof_class(u1 *sp)
 
 	/* synchronize instruction cache */
 
-	asm_cacheflush(ra, 4);
+	md_icacheflush(ra, 4);
 
 	/* patch super class' vftbl */
 
 	*((ptrint *) (pv + disp)) = (ptrint) c->vftbl;
 
+	/* synchronize data cache */
+
+	md_dcacheflush(pv + disp, SIZEOF_VOID_P);
+
 	PATCHER_MARK_PATCHED_MONITOREXIT;
 
 	return true;
@@ -803,7 +861,7 @@ bool patcher_clinit(u1 *sp)
 
 	/* check if the class is initialized */
 
-	if (!c->initialized) {
+	if (!(c->state & CLASS_INITIALIZED)) {
 		if (!initialize_class(c)) {
 			PATCHER_MONITOREXIT;
 
@@ -817,7 +875,7 @@ bool patcher_clinit(u1 *sp)
 
 	/* synchronize instruction cache */
 
-	asm_cacheflush(ra, 4);
+	md_icacheflush(ra, 4);
 
 	PATCHER_MARK_PATCHED_MONITOREXIT;
 
@@ -833,6 +891,7 @@ bool patcher_clinit(u1 *sp)
 
 *******************************************************************************/
 
+#ifdef ENABLE_VERIFIER
 bool patcher_athrow_areturn(u1 *sp)
 {
 	u1                *ra;
@@ -864,12 +923,13 @@ bool patcher_athrow_areturn(u1 *sp)
 
 	/* synchronize instruction cache */
 
-	asm_cacheflush(ra, 4);
+	md_icacheflush(ra, 4);
 
 	PATCHER_MARK_PATCHED_MONITOREXIT;
 
 	return true;
 }
+#endif /* ENABLE_VERIFIER */
 
 
 /* patcher_resolve_native ******************************************************
@@ -878,7 +938,7 @@ bool patcher_athrow_areturn(u1 *sp)
 
 *******************************************************************************/
 
-#if !defined(ENABLE_STATICVM)
+#if !defined(WITH_STATIC_CLASSPATH)
 bool patcher_resolve_native(u1 *sp)
 {
 	u1                *ra;
@@ -919,17 +979,21 @@ bool patcher_resolve_native(u1 *sp)
 
 	/* synchronize instruction cache */
 
-	asm_cacheflush(ra, 4);
+	md_icacheflush(ra, 4);
 
 	/* patch native function pointer */
 
 	*((ptrint *) (pv + disp)) = (ptrint) f;
 
+	/* synchronize data cache */
+
+	md_dcacheflush(pv + disp, SIZEOF_VOID_P);
+
 	PATCHER_MARK_PATCHED_MONITOREXIT;
 
 	return true;
 }
-#endif /* !defined(ENABLE_STATICVM) */
+#endif /* !defined(WITH_STATIC_CLASSPATH) */
 
 
 /*
