@@ -1,9 +1,9 @@
 /* src/vm/jit/verify/typecheck.c - typechecking (part of bytecode verification)
 
-   Copyright (C) 1996-2005 R. Grafl, A. Krall, C. Kruegel, C. Oates,
-   R. Obermaisser, M. Platter, M. Probst, S. Ring, E. Steiner,
-   C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich, J. Wenninger,
-   Institut f. Computersprachen - TU Wien
+   Copyright (C) 1996-2005, 2006 R. Grafl, A. Krall, C. Kruegel,
+   C. Oates, R. Obermaisser, M. Platter, M. Probst, S. Ring,
+   E. Steiner, C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich,
+   J. Wenninger, Institut f. Computersprachen - TU Wien
 
    This file is part of CACAO.
 
@@ -19,16 +19,16 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-   02111-1307, USA.
+   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
+   02110-1301, USA.
 
-   Contact: cacao@complang.tuwien.ac.at
+   Contact: cacao@cacaojvm.org
 
    Authors: Edwin Steiner
 
    Changes: Christian Thalinger
 
-   $Id: typecheck.c,v 1.97 2005/11/09 19:33:45 edwin Exp $
+   $Id: typecheck.c 4959 2006-05-26 12:09:29Z edwin $
 
 */
 
@@ -165,11 +165,11 @@ of the typeinfo of the stack slot.
 #include <assert.h>
 #include <string.h>
 
+#include "config.h"
 #include "vm/types.h"
+#include "vm/global.h"
 
-#include "vm/global.h" /* must be here because of CACAO_TYPECHECK */
-
-#ifdef CACAO_TYPECHECK
+#ifdef ENABLE_VERIFIER
 
 #include "mm/memory.h"
 #include "toolbox/logging.h"
@@ -178,7 +178,6 @@ of the typeinfo of the stack slot.
 #include "vm/jit/patcher.h"
 #include "vm/loader.h"
 #include "vm/options.h"
-#include "vm/tables.h"
 #include "vm/jit/jit.h"
 #include "vm/jit/stack.h"
 #include "vm/access.h"
@@ -196,8 +195,9 @@ of the typeinfo of the stack slot.
 #endif
 
 #ifdef TYPECHECK_VERBOSE_OPT
-bool typecheckverbose = false;
-#define DOLOG(action)  do { if (typecheckverbose) {action;} } while(0)
+bool opt_typecheckverbose = false;
+FILE *typecheck_logfile;
+#define DOLOG(action)  do { if (opt_typecheckverbose) {action;} } while(0)
 #else
 #define DOLOG(action)
 #endif
@@ -210,19 +210,19 @@ bool typecheckverbose = false;
 #define LOG3(str,a,b,c)    DOLOG(dolog(str,a,b,c))
 #define LOGIF(cond,str)    DOLOG(do {if (cond) log_text(str);} while(0))
 #ifdef  TYPEINFO_DEBUG
-#define LOGINFO(info)      DOLOG(do {typeinfo_print_short(get_logfile(),(info));log_plain("\n");} while(0))
+#define LOGINFO(info)      DOLOG(do {typeinfo_print_short(typecheck_logfile,(info));log_finish();} while(0))
 #else
 #define LOGINFO(info)
 #define typevectorset_print(x,y,z)
 #endif
-#define LOGFLUSH           DOLOG(fflush(get_logfile()))
-#define LOGNL              DOLOG(log_plain("\n"))
-#define LOGSTR(str)        DOLOG(log_plain(str))
-#define LOGSTR1(str,a)     DOLOG(dolog_plain(str,a))
-#define LOGSTR2(str,a,b)   DOLOG(dolog_plain(str,a,b))
-#define LOGSTR3(str,a,b,c) DOLOG(dolog_plain(str,a,b,c))
-#define LOGSTRu(utf)       DOLOG(log_plain_utf(utf))
-#define LOGNAME(c)         DOLOG(do {log_plain_utf(IS_CLASSREF(c) ? c.ref->name : c.cls->name);} while(0))
+#define LOGFLUSH           DOLOG(fflush(typecheck_logfile))
+#define LOGNL              DOLOG(log_finish())
+#define LOGSTR(str)        DOLOG(log_print(str))
+#define LOGSTR1(str,a)     DOLOG(log_print(str,a))
+#define LOGSTR2(str,a,b)   DOLOG(log_print(str,a,b))
+#define LOGSTR3(str,a,b,c) DOLOG(log_print(str,a,b,c))
+#define LOGNAME(c)         DOLOG(log_message_utf("class: ",(IS_CLASSREF(c) ? c.ref->name : c.cls->name)))
+#define LOGMETHOD(str,m)   DOLOG(log_message_method(str,m))
 #else
 #define LOG(str)
 #define LOG1(str,a)
@@ -236,18 +236,16 @@ bool typecheckverbose = false;
 #define LOGSTR1(str,a)
 #define LOGSTR2(str,a,b)
 #define LOGSTR3(str,a,b,c)
-#define LOGSTRu(utf)
 #define LOGNAME(c)
+#define LOGMETHOD(str,m)
 #endif
 
 #ifdef TYPECHECK_VERBOSE_IMPORTANT
 #define LOGimp(str)     DOLOG(log_text(str))
-#define LOGimpSTR(str)  DOLOG(log_plain(str))
-#define LOGimpSTRu(utf) DOLOG(log_plain_utf(utf))
+#define LOGimpSTR(str)  DOLOG(log_print(str))
 #else
 #define LOGimp(str)
 #define LOGimpSTR(str)
-#define LOGimpSTRu(utf)
 #endif
 
 #if defined(TYPECHECK_VERBOSE) || defined(TYPECHECK_VERBOSE_IMPORTANT)
@@ -515,6 +513,7 @@ typestack_copy(verifier_state *state,stackptr dst,stackptr y,typevector *selecte
 	int k;
 	
 	for (;dst; dst=dst->prev, y=y->prev) {
+		/* XXX only check the following two in debug mode? */
 		if (!y) {
 			*exceptionptr = new_verifyerror(state->m,"Stack depth mismatch");
 			return false;
@@ -703,8 +702,30 @@ typestack_add(stackptr dst,stackptr y,int ky)
 	}
 }
 
-/* XXX document */
-/* 'a' and 'b' are assumed to have passed typestack_canmerge! */
+/* typestack_separable_with ****************************************************
+ 
+   This function answers the question: If variant 'kb' of typestack 'b' is
+   added to typestack 'a', will the result be separable?
+
+   A typestack is called 'separable' if it has at least one slot of type
+   returnAddress that contains at least two different return addresses.
+   (ie. a RET using the value in this slot could go to more than one target)
+   
+   IN:
+	   a................the first typestack
+	   b................the second typestack
+	   kb...............the k-index of the variant that should be selected
+	                    from typestack 'b'
+
+   OUT:
+       true.............the result would be separable
+	   false............the result would not be separable
+
+   PRE-CONDITION:
+       'a' and 'b' are assumed to have passed typestack_canmerge!
+
+*******************************************************************************/
+
 static bool
 typestack_separable_with(stackptr a,stackptr b,int kb)
 {
@@ -727,8 +748,33 @@ typestack_separable_with(stackptr a,stackptr b,int kb)
 	return false;
 }
 
-/* XXX document */
-/* 'a' and 'b' are assumed to have passed typestack_canmerge! */
+/* typestack_separable_from ****************************************************
+ 
+   This function answers the question: Is variant 'ka' of typestack 'a'
+   separable from variant 'kb' of typestack 'b'?
+
+   Two variants of typestacks are called 'separable' from each other, if there
+   is at least one slot for which the variants contain different return addresses.
+   (ie. a RET using the value in this slot would go to one target in the first
+   variant and to another target in the second variant)
+   
+   IN:
+	   a................the first typestack
+	   ka...............the k-index of the variant that should be selected
+	                    from typestack 'a'
+	   b................the second typestack
+	   kb...............the k-index of the variant that should be selected
+	                    from typestack 'b'
+
+   OUT:
+       true.............the variants are separable
+	   false............the variants are not separable
+
+   PRE-CONDITION:
+       'a' and 'b' are assumed to have passed typestack_canmerge!
+
+*******************************************************************************/
+
 static bool
 typestack_separable_from(stackptr a,int ka,stackptr b,int kb)
 {
@@ -792,10 +838,10 @@ typestate_merge(verifier_state *state,
 	typecheck_result r;
 	
 	LOG("merge:");
-	LOGSTR("dstack: "); DOLOG(typestack_print(get_logfile(),deststack)); LOGNL;
-	LOGSTR("ystack: "); DOLOG(typestack_print(get_logfile(),ystack)); LOGNL;
-	LOGSTR("dloc  : "); DOLOG(typevectorset_print(get_logfile(),destloc,state->numlocals)); LOGNL;
-	LOGSTR("yloc  : "); DOLOG(typevectorset_print(get_logfile(),yloc,state->numlocals)); LOGNL;
+	LOGSTR("dstack: "); DOLOG(typestack_print(typecheck_logfile,deststack)); LOGNL;
+	LOGSTR("ystack: "); DOLOG(typestack_print(typecheck_logfile,ystack)); LOGNL;
+	LOGSTR("dloc  : "); DOLOG(typevectorset_print(typecheck_logfile,destloc,state->numlocals)); LOGNL;
+	LOGSTR("yloc  : "); DOLOG(typevectorset_print(typecheck_logfile,yloc,state->numlocals)); LOGNL;
 	LOGFLUSH;
 
 	/* The stack is always merged. If there are returnAddresses on
@@ -869,8 +915,8 @@ typestate_merge(verifier_state *state,
 	}
 	
 	LOG("result:");
-	LOGSTR("dstack: "); DOLOG(typestack_print(get_logfile(),deststack)); LOGNL;
-	LOGSTR("dloc  : "); DOLOG(typevectorset_print(get_logfile(),destloc,state->numlocals)); LOGNL;
+	LOGSTR("dstack: "); DOLOG(typestack_print(typecheck_logfile,deststack)); LOGNL;
+	LOGSTR("dloc  : "); DOLOG(typevectorset_print(typecheck_logfile,destloc,state->numlocals)); LOGNL;
 	LOGFLUSH;
 	
 	return changed;
@@ -918,8 +964,6 @@ typestate_reach(verifier_state *state,
 		stackptr sp;
 		int i;
 
-		/* XXX FIXME FOR INLINING */
-
 		if (!useinlining) {
 			TYPECHECK_COUNT(stat_backwards);
         		LOG("BACKWARDS!");
@@ -1115,7 +1159,10 @@ verify_invocation(verifier_state *state)
 {
 	unresolved_method *um;      /* struct describing the called method */
 	constant_FMIref *mref;           /* reference to the called method */
+	methodinfo *mi;                        /* resolved method (if any) */
 	methoddesc *md;                 /* descriptor of the called method */
+	utf *mname;                                         /* method name */
+	utf *mclassname;                     /* name of the method's class */
 	bool specialmethod;            /* true if a <...> method is called */
 	int opcode;                                   /* invocation opcode */
 	bool callinginit;                      /* true if <init> is called */
@@ -1126,11 +1173,31 @@ verify_invocation(verifier_state *state)
 	stackelement *dst;               /* result stack of the invocation */
 	int i;                                                  /* counter */
     u1 rtype;                          /* return type of called method */
+	resolve_result_t result;
+
+	if (INSTRUCTION_IS_UNRESOLVED(state->iptr)) {
+		/* unresolved method */
+		um = INSTRUCTION_UNRESOLVED_METHOD(state->iptr);
+		mref = um->methodref;
+	}
+	else {
+		um = NULL;
+		mref = INSTRUCTION_RESOLVED_FMIREF(state->iptr);
+	}
 
-	um = (unresolved_method *) state->iptr[0].target;
-	mref = um->methodref;
 	md = mref->parseddesc.md;
-	specialmethod = (mref->name->text[0] == '<');
+	mname = mref->name;
+
+	if (IS_FMIREF_RESOLVED(mref)) {
+		mi = mref->p.method;
+		mclassname = mi->class->name;
+	}
+	else {
+		mi = NULL;
+		mclassname = mref->p.classref->name;
+	}
+
+	specialmethod = (mname->text[0] == '<');
 	opcode = state->iptr[0].opc;
 	dst = state->iptr->dst;
 
@@ -1140,7 +1207,7 @@ verify_invocation(verifier_state *state)
 
 	/* check whether we are calling <init> */
 	
-	callinginit = (opcode == ICMD_INVOKESPECIAL && mref->name == utf_init);
+	callinginit = (opcode == ICMD_INVOKESPECIAL && mname == utf_init);
 	if (specialmethod && !callinginit)
 		TYPECHECK_VERIFYERROR_bool("Invalid invocation of special method");
 
@@ -1172,7 +1239,7 @@ verify_invocation(verifier_state *state)
 				LOGINFO(&(stack->typeinfo));
 				ins = (instruction*)TYPEINFO_NEWOBJECT_INSTRUCTION(stack->typeinfo);
 				if (ins)
-					initclass = CLASSREF_OR_CLASSINFO(ins[-1].target);
+					initclass = ICMD_ACONST_CLASSREF_OR_CLASSINFO(ins-1);
 				else
 					initclass.cls = state->m->class;
 				LOGSTR("class: "); LOGNAME(initclass); LOGNL;
@@ -1235,8 +1302,13 @@ verify_invocation(verifier_state *state)
 			/* must be <init> of current class or direct superclass                   */
 			/* the current class is linked, so must be its superclass. thus we can be */
 			/* sure that resolving will be trivial.                                   */
-			if (!resolve_classref(state->m,mref->classref,resolveLazy,false,true,&cls))
-				return false; /* exception */
+			if (mi) {
+				cls = mi->class;
+			}
+			else {
+				if (!resolve_classref(state->m,mref->p.classref,resolveLazy,false,true,&cls))
+					return false; /* exception */
+			}
 
 			/* if lazy resolving did not succeed, it's not one of the allowed classes */
 			/* otherwise we check it directly                                         */
@@ -1250,22 +1322,39 @@ verify_invocation(verifier_state *state)
 		}
 		else {
 			/* { we are initializing an instance created with NEW } */
-			if ((IS_CLASSREF(initclass) ? initclass.ref->name : initclass.cls->name) != mref->classref->name) {
+			if ((IS_CLASSREF(initclass) ? initclass.ref->name : initclass.cls->name) != mclassname) {
 				TYPECHECK_VERIFYERROR_bool("wrong <init> called for uninitialized reference");
 			}
 		}
 	}
 
-	/* record subtype constraints for parameters */
-	
-	if (!constrain_unresolved_method(um,state->m->class,state->m,state->iptr,state->curstack))
-		return false; /* XXX maybe wrap exception */
-
 	/* try to resolve the method lazily */
-	
-	if (!resolve_method(um,resolveLazy,(methodinfo **) &(state->iptr[0].val.a)))
+
+	result = resolve_method_lazy(state->iptr,state->curstack,state->m);
+	if (result == resolveFailed)
 		return false;
 
+	if (result != resolveSucceeded) {
+		if (!um) {
+			um = create_unresolved_method(state->m->class,
+					state->m, state->iptr);
+
+			if (!um)
+				return false;
+		}
+
+		/* record subtype constraints for parameters */
+
+		if (!constrain_unresolved_method(um,state->m->class,state->m,state->iptr,state->curstack))
+			return false; /* XXX maybe wrap exception */
+
+		/* store the unresolved_method pointer */
+
+		/* XXX this will be changed */
+		state->iptr->val.a = um;
+		state->iptr->target = (void*) 0x01; /* XXX used temporarily as flag */
+	}
+
 	return true;
 }
 
@@ -1361,7 +1450,7 @@ verify_builtin(verifier_state *state)
 	if (ISBUILTIN(BUILTIN_new)) {
 		if (state->iptr[-1].opc != ICMD_ACONST)
 			TYPECHECK_VERIFYERROR_bool("illegal instruction: builtin_new without class");
-		cls.any = state->iptr[-1].target;
+		cls = ICMD_ACONST_CLASSREF_OR_CLASSINFO(state->iptr - 1);
 		TYPEINFO_INIT_NEWOBJECT(dst->typeinfo,state->iptr);
 	}
 	else if (ISBUILTIN(BUILTIN_newarray_boolean)) {
@@ -1399,10 +1488,10 @@ verify_builtin(verifier_state *state)
 	else if (ISBUILTIN(BUILTIN_newarray))
 	{
 		TYPECHECK_INT(state->curstack->prev);
-		if (state->iptr[-1].opc != ICMD_ACONST || !state->iptr[-1].target)
+		if (state->iptr[-1].opc != ICMD_ACONST)
 			TYPECHECK_VERIFYERROR_bool("illegal instruction: builtin_newarray without class");
 		/* XXX check that it is an array class(ref) */
-		typeinfo_init_class(&(dst->typeinfo),CLASSREF_OR_CLASSINFO(state->iptr[-1].target));
+		typeinfo_init_class(&(dst->typeinfo),ICMD_ACONST_CLASSREF_OR_CLASSINFO(state->iptr - 1));
 	}
 	else if (ISBUILTIN(BUILTIN_arrayinstanceof))
 	{
@@ -1452,7 +1541,7 @@ verify_multianewarray(verifier_state *state)
 	}
 
 	/* check array descriptor */
-	if (state->iptr[0].target == NULL) {
+	if (state->iptr[0].val.a != NULL) {
 		/* the array class reference has already been resolved */
 		arrayclass = (classinfo *) state->iptr[0].val.a;
 		if (!arrayclass)
@@ -1471,7 +1560,7 @@ verify_multianewarray(verifier_state *state)
 		
 		/* the array class reference is still unresolved */
 		/* check that the reference indicates an array class of correct dimension */
-		cr = (constant_classref *) state->iptr[0].val.a;
+		cr = (constant_classref *) state->iptr[0].target; /* XXX new instruction format */
 		i = 0;
 		p = cr->name->text;
 		while (p[i] == '[')
@@ -1483,7 +1572,7 @@ verify_multianewarray(verifier_state *state)
 			TYPECHECK_VERIFYERROR_bool("MULTIANEWARRAY dimension to high");
 
 		/* set the array type of the result */
-		if (!typeinfo_init_class(&(state->iptr->dst->typeinfo),CLASSREF_OR_CLASSINFO(state->iptr[0].val.a)))
+		if (!typeinfo_init_class(&(state->iptr->dst->typeinfo),CLASSREF_OR_CLASSINFO(state->iptr[0].target))) /* XXX new instruction format */
 			return false;
 	}
 
@@ -1515,12 +1604,12 @@ verify_basic_block(verifier_state *state)
     basicblock **tptr;    /* pointer into target list of switch instr. */
     classinfo *cls;                                       /* temporary */
     bool maythrow;               /* true if this instruction may throw */
-    classinfo *myclass;
 	unresolved_field *uf;                        /* for field accesses */
-	fieldinfo **fieldinfop;                      /* for field accesses */
+	constant_FMIref *fieldref;                   /* for field accesses */
 	s4 i;
 	s4 b_index;
 	typecheck_result r;
+	resolve_result_t result;
 
 	LOGSTR1("\n---- BLOCK %04d ------------------------------------------------\n",state->bptr->debug_nr);
 	LOGFLUSH;
@@ -1552,26 +1641,23 @@ verify_basic_block(verifier_state *state)
 	typevectorset_copy_inplace(MGET_TYPEVECTOR(state->localbuf,b_index,state->numlocals),
 			state->localset,state->numlocals);
 
-	/* XXX FIXME FOR INLINING */
-	if (!useinlining) {
-		if (state->handlers[0])
-			for (i=0; i<state->numlocals; ++i)
-				if (state->localset->td[i].type == TYPE_ADR
-						&& TYPEINFO_IS_NEWOBJECT(state->localset->td[i].info)) {
-					/* XXX we do not check this for the uninitialized 'this' instance in */
-					/* <init> methods. Otherwise there are problems with try blocks in   */
-					/* <init>. The spec seems to indicate that we should perform the test*/
-					/* in all cases, but this fails with real code.                      */
-					/* Example: org/eclipse/ui/internal/PerspectiveBarNewContributionItem*/
-					/* of eclipse 3.0.2                                                  */
-					if (TYPEINFO_NEWOBJECT_INSTRUCTION(state->localset->td[i].info) != NULL) {
-						/*show_icmd_method(state->m, state->cd, state->rd);*/
-						printf("Uninitialized variale: %d, block: %d\n", i, state->bptr->debug_nr);
-						TYPECHECK_VERIFYERROR_bool("Uninitialized object in local variable inside try block");
-					}
+	if (state->handlers[0])
+		for (i=0; i<state->numlocals; ++i)
+			if (state->localset->td[i].type == TYPE_ADR
+					&& TYPEINFO_IS_NEWOBJECT(state->localset->td[i].info)) {
+				/* XXX we do not check this for the uninitialized 'this' instance in */
+				/* <init> methods. Otherwise there are problems with try blocks in   */
+				/* <init>. The spec seems to indicate that we should perform the test*/
+				/* in all cases, but this fails with real code.                      */
+				/* Example: org/eclipse/ui/internal/PerspectiveBarNewContributionItem*/
+				/* of eclipse 3.0.2                                                  */
+				if (TYPEINFO_NEWOBJECT_INSTRUCTION(state->localset->td[i].info) != NULL) {
+					/*show_icmd_method(state->m, state->cd, state->rd);*/
+					printf("Uninitialized variable: %d, block: %d\n", i, state->bptr->debug_nr);
+					TYPECHECK_VERIFYERROR_bool("Uninitialized object in local variable inside try block");
 				}
-	}
-	DOLOG(typestate_print(get_logfile(),state->curstack,state->localset,state->numlocals));
+			}
+	DOLOG(typestate_print(typecheck_logfile,state->curstack,state->localset,state->numlocals));
 	LOGNL; LOGFLUSH;
 
 	/* loop over the instructions */
@@ -1580,13 +1666,12 @@ verify_basic_block(verifier_state *state)
 	while (--len >= 0)  {
 		TYPECHECK_COUNT(stat_ins);
 
-		DOLOG(typestate_print(get_logfile(),state->curstack,state->localset,state->numlocals));
+		DOLOG(typestate_print(typecheck_logfile,state->curstack,state->localset,state->numlocals));
 		LOGNL; LOGFLUSH;
 
 		DOLOG(show_icmd(state->iptr,false)); LOGNL; LOGFLUSH;
 
 		opcode = state->iptr->opc;
-		myclass = state->iptr->method->class;
 		dst = state->iptr->dst;
 		maythrow = false;
 
@@ -1737,8 +1822,14 @@ verify_basic_block(verifier_state *state)
 			case ICMD_PUTSTATICCONST:
 				TYPECHECK_COUNT(stat_ins_field);
 
-				uf = INSTRUCTION_PUTCONST_FIELDREF(state->iptr);
-				fieldinfop = INSTRUCTION_PUTCONST_FIELDINFO_PTR(state->iptr);
+				if (INSTRUCTION_IS_UNRESOLVED(state->iptr + 1)) {
+					uf = INSTRUCTION_UNRESOLVED_FIELD(state->iptr + 1);
+					fieldref = uf->fieldref;
+				}
+				else {
+					uf = NULL;
+					fieldref = INSTRUCTION_RESOLVED_FMIREF(state->iptr + 1);
+				}
 
 				goto fieldaccess_tail;
 
@@ -1746,8 +1837,14 @@ verify_basic_block(verifier_state *state)
 			case ICMD_PUTSTATIC:
 				TYPECHECK_COUNT(stat_ins_field);
 
-				uf = (unresolved_field *) state->iptr[0].target;
-				fieldinfop = (fieldinfo **) &(state->iptr[0].val.a);
+				if (INSTRUCTION_IS_UNRESOLVED(state->iptr)) {
+					uf = INSTRUCTION_UNRESOLVED_FIELD(state->iptr);
+					fieldref = uf->fieldref;
+				}
+				else {
+					uf = NULL;
+					fieldref = INSTRUCTION_RESOLVED_FMIREF(state->iptr);
+				}
 				
 				goto fieldaccess_tail;
 
@@ -1755,31 +1852,50 @@ verify_basic_block(verifier_state *state)
 			case ICMD_GETSTATIC:
 				TYPECHECK_COUNT(stat_ins_field);
 
-				uf = (unresolved_field *) state->iptr[0].target;
-				fieldinfop = (fieldinfo **) &(state->iptr[0].val.a);
+				if (INSTRUCTION_IS_UNRESOLVED(state->iptr)) {
+					uf = INSTRUCTION_UNRESOLVED_FIELD(state->iptr);
+					fieldref = uf->fieldref;
+				}
+				else {
+					uf = NULL;
+					fieldref = INSTRUCTION_RESOLVED_FMIREF(state->iptr);
+				}
 
 				/* the result is pushed on the stack */
 				if (dst->type == TYPE_ADR) {
-					if (!typeinfo_init_from_typedesc(uf->fieldref->parseddesc.fd,NULL,&(dst->typeinfo)))
+					if (!typeinfo_init_from_typedesc(fieldref->parseddesc.fd,NULL,&(dst->typeinfo)))
 						return false;
 				}
 
 fieldaccess_tail:
-				/* record the subtype constraints for this field access */
-				if (!constrain_unresolved_field(uf,state->m->class,state->m,state->iptr,state->curstack))
-					return false; /* XXX maybe wrap exception? */
-
-				/* try to resolve the field reference */
-				if (!resolve_field(uf,resolveLazy,fieldinfop))
+				/* try to resolve the field reference lazily */
+				result = resolve_field_lazy(state->iptr, state->curstack, state->m);
+				if (result == resolveFailed)
 					return false;
 
-				/* we need a patcher, so this is not a leafmethod */
-#if defined(__MIPS__)
-				if (!*fieldinfop || !(*fieldinfop)->class->initialized)
-					state->cd->method->isleafmethod = false;
-#endif
-				TYPECHECK_COUNTIF(!*fieldinfop,stat_ins_field_unresolved);
-				TYPECHECK_COUNTIF(*fieldinfop && !(*fieldinfop)->class->initialized,stat_ins_field_uninitialized);
+				if (result != resolveSucceeded) {
+					if (!uf) {
+						uf = create_unresolved_field(state->m->class, state->m, state->iptr);
+						if (!uf)
+							return false;
+
+						if (opcode == ICMD_PUTSTATICCONST || opcode == ICMD_PUTFIELDCONST) {
+							state->iptr[1].val.a = uf;
+							state->iptr[1].target = (void*) 0x01; /* XXX target used temporarily as flag */
+						}
+						else {
+							state->iptr[0].val.a = uf;
+							state->iptr[0].target = (void*) 0x01; /* XXX target used temporarily as flag */
+						}
+					}
+
+					/* record the subtype constraints for this field access */
+					if (!constrain_unresolved_field(uf,state->m->class,state->m,state->iptr,state->curstack))
+						return false; /* XXX maybe wrap exception? */
+
+					TYPECHECK_COUNTIF(INSTRUCTION_IS_UNRESOLVED(state->iptr),stat_ins_field_unresolved);
+					TYPECHECK_COUNTIF(INSTRUCTION_IS_RESOLVED(state->iptr) && !INSTRUCTION_RESOLVED_FIELDINFO(state->iptr)->class->initialized,stat_ins_field_uninitialized);
+				}
 					
 				maythrow = true;
 				break;
@@ -1916,9 +2032,10 @@ fieldaccess_tail:
 				/* ADDRESS CONSTANTS                    */
 
 			case ICMD_ACONST:
-				if (state->iptr->target) {
+				if (ICMD_ACONST_IS_CLASS(state->iptr)) {
 					/* a java.lang.Class reference */
-					TYPEINFO_INIT_JAVA_LANG_CLASS(dst->typeinfo,(constant_classref *)state->iptr->target);
+					TYPEINFO_INIT_JAVA_LANG_CLASS(dst->typeinfo,
+												  ICMD_ACONST_CLASSREF_OR_CLASSINFO(state->iptr));
 				}
 				else {
 					if (state->iptr->val.a == NULL)
@@ -1943,7 +2060,7 @@ fieldaccess_tail:
 				if (cls)
 					typeinfo_init_classinfo(&(dst->typeinfo),cls);
 				else
-					if (!typeinfo_init_class(&(dst->typeinfo),CLASSREF_OR_CLASSINFO(state->iptr[0].target)))
+					if (!typeinfo_init_class(&(dst->typeinfo),CLASSREF_OR_CLASSINFO(state->iptr[0].target))) /* XXX new instruction format */
 						return false;
 				maythrow = true;
 				break;
@@ -1958,6 +2075,9 @@ fieldaccess_tail:
 				/****************************************/
 				/* BRANCH INSTRUCTIONS                  */
 
+			case ICMD_INLINE_GOTO:
+				COPYTYPE(state->curstack,dst);
+				/* FALLTHROUGH! */
 			case ICMD_GOTO:
 				superblockend = true;
 				/* FALLTHROUGH! */
@@ -1977,18 +2097,46 @@ fieldaccess_tail:
 			case ICMD_IF_ICMPLE:
 			case ICMD_IF_ACMPEQ:
 			case ICMD_IF_ACMPNE:
+
 			case ICMD_IF_LEQ:
 			case ICMD_IF_LNE:
 			case ICMD_IF_LLT:
 			case ICMD_IF_LGE:
 			case ICMD_IF_LGT:
 			case ICMD_IF_LLE:
+
 			case ICMD_IF_LCMPEQ:
 			case ICMD_IF_LCMPNE:
 			case ICMD_IF_LCMPLT:
 			case ICMD_IF_LCMPGE:
 			case ICMD_IF_LCMPGT:
 			case ICMD_IF_LCMPLE:
+
+			case ICMD_IF_FCMPEQ:
+			case ICMD_IF_FCMPNE:
+
+			case ICMD_IF_FCMPL_LT:
+			case ICMD_IF_FCMPL_GE:
+			case ICMD_IF_FCMPL_GT:
+			case ICMD_IF_FCMPL_LE:
+
+			case ICMD_IF_FCMPG_LT:
+			case ICMD_IF_FCMPG_GE:
+			case ICMD_IF_FCMPG_GT:
+			case ICMD_IF_FCMPG_LE:
+
+			case ICMD_IF_DCMPEQ:
+			case ICMD_IF_DCMPNE:
+
+			case ICMD_IF_DCMPL_LT:
+			case ICMD_IF_DCMPL_GE:
+			case ICMD_IF_DCMPL_GT:
+			case ICMD_IF_DCMPL_LE:
+
+			case ICMD_IF_DCMPG_LT:
+			case ICMD_IF_DCMPG_GE:
+			case ICMD_IF_DCMPG_GT:
+			case ICMD_IF_DCMPG_LE:
 				TYPECHECK_COUNT(stat_ins_branch);
 				tbptr = (basicblock *) state->iptr->target;
 
@@ -2201,13 +2349,6 @@ return_tail:
 				LOG("Should have been converted to builtin function call.");
 				TYPECHECK_ASSERT(false);
 				break;
-
-			case ICMD_READONLY_ARG:
-			case ICMD_CLEAR_ARGREN:
-				LOG2("ICMD %d at %d\n", state->iptr->opc, (int)(state->iptr-state->bptr->iinstr));
-				LOG("Should have been replaced in stack.c.");
-				TYPECHECK_ASSERT(false);
-				break;
 #endif
 
 				/****************************************/
@@ -2217,7 +2358,9 @@ return_tail:
 				 * Instructions below...
 				 *     *) don't operate on local variables,
 				 *     *) don't operate on references,
-				 *     *) don't operate on returnAddresses.
+				 *     *) don't operate on returnAddresses,
+				 *     *) don't affect control flow (except
+				 *        by throwing exceptions).
 				 *
 				 * (These instructions are typechecked in
 				 *  analyse_stack.)
@@ -2272,10 +2415,6 @@ return_tail:
 			case ICMD_LSHL:
 			case ICMD_LSHR:
 			case ICMD_LUSHR:
-#if 0
-			case ICMD_IREM0X10001:
-			case ICMD_LREM0X10001:
-#endif
 			case ICMD_IMULPOW2:
 			case ICMD_LMULPOW2:
 			case ICMD_IDIVPOW2:
@@ -2337,6 +2476,8 @@ return_tail:
 			case ICMD_FNEG:
 			case ICMD_DNEG:
 
+			case ICMD_INLINE_START:
+			case ICMD_INLINE_END:
 
 				/*XXX What shall we do with the following ?*/
 			case ICMD_AASTORECONST:
@@ -2380,7 +2521,7 @@ return_tail:
 
 	LOG("instructions done");
 	LOGSTR("RESULT=> ");
-	DOLOG(typestate_print(get_logfile(),state->curstack,state->localset,state->numlocals));
+	DOLOG(typestate_print(typecheck_logfile,state->curstack,state->localset,state->numlocals));
 	LOGNL; LOGFLUSH;
 
 	/* propagate stack and variables to the following block */
@@ -2578,19 +2719,19 @@ typecheck_reset_flags(verifier_state *state)
 /*    rdata............registerdata for the method                          */
 /*                                                                          */
 /* RETURN VALUE:                                                            */
-/*     m................successful verification                             */
-/*     NULL.............an exception has been thrown                        */
+/*     true.............successful verification                             */
+/*     false............an exception has been thrown                        */
 /*                                                                          */
-/* XXX TODO:                                                                */
-/*     Bytecode verification has not been tested with inlining and          */
-/*     probably does not work correctly with inlining.                      */
 /****************************************************************************/
 
 #define MAXPARAMS 255
 
-methodinfo *typecheck(methodinfo *meth, codegendata *cdata, registerdata *rdata)
+bool typecheck(jitdata *jd)
 {
-	verifier_state state;             /* current state of the verifier */
+	methodinfo     *meth;
+	codegendata    *cd;
+	registerdata   *rd;
+	verifier_state  state;             /* current state of the verifier */
     int i;                                        /* temporary counter */
 
 	/* collect statistics */
@@ -2604,19 +2745,19 @@ methodinfo *typecheck(methodinfo *meth, codegendata *cdata, registerdata *rdata)
 	state.stat_maythrow = false;
 #endif
 
+	/* get required compiler data */
+
+	meth = jd->m;
+	cd   = jd->cd;
+	rd   = jd->rd;
+
 	/* some logging on entry */
 
+	DOLOG(typecheck_logfile = stdout);
     LOGSTR("\n==============================================================================\n");
     /*DOLOG( show_icmd_method(cdata->method,cdata,rdata));*/
     LOGSTR("\n==============================================================================\n");
-    LOGimpSTR("Entering typecheck: ");
-    LOGimpSTRu(cdata->method->name);
-    LOGimpSTR("    ");
-    LOGimpSTRu(cdata->method->descriptor);
-    LOGimpSTR("    (class ");
-    LOGimpSTRu(cdata->method->class->name);
-    LOGimpSTR(")\n");
-	LOGFLUSH;
+    LOGMETHOD("Entering typecheck: ",cdata->method);
 
 	/* initialize the verifier state */
 
@@ -2624,8 +2765,8 @@ methodinfo *typecheck(methodinfo *meth, codegendata *cdata, registerdata *rdata)
 	state.savedstack = NULL;
 	state.jsrencountered = false;
 	state.m = meth;
-	state.cd = cdata;
-	state.rd = rdata;
+	state.cd = cd;
+	state.rd = rd;
 
 	/* check if this method is an instance initializer method */
 
@@ -2659,7 +2800,7 @@ methodinfo *typecheck(methodinfo *meth, codegendata *cdata, registerdata *rdata)
 	/* initialized local variables of first block */
 
 	if (!verify_init_locals(&state))
-		return NULL;
+		return false;
 
     /* initialize the input stack of exception handlers */
 	
@@ -2687,7 +2828,7 @@ methodinfo *typecheck(methodinfo *meth, codegendata *cdata, registerdata *rdata)
 		    /* verify reached block */	
             if (state.bptr->flags == BBTYPECHECK_REACHED) {
                 if (!verify_basic_block(&state))
-					return NULL;
+					return false;
             }
             state.bptr++;
         } /* while blocks */
@@ -2708,12 +2849,13 @@ methodinfo *typecheck(methodinfo *meth, codegendata *cdata, registerdata *rdata)
 
 	typecheck_reset_flags(&state);
 
-	/* just return methodinfo* to indicate everything was ok */
+	/* everything's ok */
+
     LOGimp("exiting typecheck");
-	return state.m;
+	return true;
 }
 
-#endif /* CACAO_TYPECHECK */
+#endif /* ENABLE_VERIFIER */
 
 /*
  * These are local overrides for various environment variables in Emacs.
