@@ -1,9 +1,9 @@
 /* src/vm/jit/verify/typeinfo.c - type system used by the type checker
 
-   Copyright (C) 1996-2005 R. Grafl, A. Krall, C. Kruegel, C. Oates,
-   R. Obermaisser, M. Platter, M. Probst, S. Ring, E. Steiner,
-   C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich, J. Wenninger,
-   Institut f. Computersprachen - TU Wien
+   Copyright (C) 1996-2005, 2006 R. Grafl, A. Krall, C. Kruegel,
+   C. Oates, R. Obermaisser, M. Platter, M. Probst, S. Ring,
+   E. Steiner, C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich,
+   J. Wenninger, Institut f. Computersprachen - TU Wien
 
    This file is part of CACAO.
 
@@ -19,14 +19,14 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-   02111-1307, USA.
+   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
+   02110-1301, USA.
 
-   Contact: cacao@complang.tuwien.ac.at
+   Contact: cacao@cacaojvm.org
 
    Authors: Edwin Steiner
 
-   $Id: typeinfo.c,v 3.41 2005/11/08 19:01:17 edwin Exp $
+   $Id: typeinfo.c 4879 2006-05-05 17:34:49Z edwin $
 
 */
 
@@ -37,7 +37,6 @@
 #include "toolbox/logging.h"
 #include "vm/class.h"
 #include "vm/loader.h"
-#include "vm/tables.h"
 #include "vm/jit/jit.h"
 #include "vm/jit/verify/typeinfo.h"
 #include "vm/descriptor.h"
@@ -51,7 +50,7 @@
 /* check if a linked class implements the interface with the given index */
 #define CLASSINFO_IMPLEMENTS_INTERFACE(cls,index)                   \
     ( ((index) < (cls)->vftbl->interfacetablelength)            \
-      && (VFTBLINTERFACETABLE((cls)->vftbl,(index)) != NULL) )
+      && ( (cls)->vftbl->interfacetable[-(index)] != NULL ) )
 
 /******************************************************************************/
 /* DEBUG HELPERS                                                              */
@@ -783,7 +782,7 @@ interface_extends_interface(classinfo *cls,classinfo *interf)
 	TYPEINFO_ASSERT(interf);
 	TYPEINFO_ASSERT((interf->flags & ACC_INTERFACE) != 0);
 	TYPEINFO_ASSERT((cls->flags & ACC_INTERFACE) != 0);
-	TYPEINFO_ASSERT(cls->linked);
+	TYPEINFO_ASSERT(cls->state & CLASS_LINKED);
 
     /* first check direct superinterfaces */
     for (i=0; i<cls->interfacescount; ++i) {
@@ -822,7 +821,7 @@ classinfo_implements_interface(classinfo *cls,classinfo *interf)
 	TYPEINFO_ASSERT(interf);
 	TYPEINFO_ASSERT((interf->flags & ACC_INTERFACE) != 0);
 
-	if (!cls->linked)
+	if (!(cls->state & CLASS_LINKED))
 		if (!link_class(cls))
 			return typecheck_FAIL;
 
@@ -835,7 +834,7 @@ classinfo_implements_interface(classinfo *cls,classinfo *interf)
         return interface_extends_interface(cls,interf);
     }
 
-	TYPEINFO_ASSERT(cls->linked);
+	TYPEINFO_ASSERT(cls->state & CLASS_LINKED);
     return CLASSINFO_IMPLEMENTS_INTERFACE(cls,interf->index);
 }
 
@@ -972,8 +971,8 @@ merged_is_subclass(classinfo *typeclass,typeinfo_mergedlist *merged,
     if (typeclass == pseudo_class_Null)
         return typecheck_TRUE;
 
-	TYPEINFO_ASSERT(typeclass->loaded);
-	TYPEINFO_ASSERT(typeclass->linked);
+	TYPEINFO_ASSERT(typeclass->state & CLASS_LOADED);
+	TYPEINFO_ASSERT(typeclass->state & CLASS_LINKED);
 
     /* check if the common typeclass is a subclass of CLS. */
 	if (class_issubclass(typeclass,cls))
@@ -993,7 +992,7 @@ merged_is_subclass(classinfo *typeclass,typeinfo_mergedlist *merged,
 		if (IS_CLASSREF(*mlist)) {
 			return typecheck_MAYBE;
 		}
-		if (!mlist->cls->linked)
+		if (!(mlist->cls->state & CLASS_LINKED))
 			if (!link_class(mlist->cls))
 				return typecheck_FAIL;
 		if (!class_issubclass(mlist->cls,cls))
@@ -1090,20 +1089,20 @@ typeinfo_is_assignable_to_class(typeinfo *value,classref_or_classinfo dest)
 
 	cls = c.cls;
 	
-	TYPEINFO_ASSERT(cls->loaded);
-	TYPEINFO_ASSERT(dest.cls->loaded);
+	TYPEINFO_ASSERT(cls->state & CLASS_LOADED);
+	TYPEINFO_ASSERT(dest.cls->state & CLASS_LOADED);
 
 	/* maybe we need to link the classes */
-	if (!cls->linked)
+	if (!(cls->state & CLASS_LINKED))
 		if (!link_class(cls))
 			return typecheck_FAIL;
-	if (!dest.cls->linked)
+	if (!(dest.cls->state & CLASS_LINKED))
 		if (!link_class(dest.cls))
 			return typecheck_FAIL;
 
 	/* { we know that both c and dest are linked classes } */
-	TYPEINFO_ASSERT(cls->linked);
-	TYPEINFO_ASSERT(dest.cls->linked);
+	TYPEINFO_ASSERT(cls->state & CLASS_LINKED);
+	TYPEINFO_ASSERT(dest.cls->state & CLASS_LINKED);
 
     if (dest.cls->flags & ACC_INTERFACE) {
         /* We are assigning to an interface type. */
@@ -1283,7 +1282,7 @@ typeinfo_init_class(typeinfo *info,classref_or_classinfo c)
 	TYPEINFO_ASSERT(info);
 
 	/* if necessary, try to resolve lazily */
-	if (!resolve_classref_or_classinfo(NULL /* XXX should now method */,
+	if (!resolve_classref_or_classinfo(NULL /* XXX should know method */,
 				c,resolveLazy,false,true,&cls))
 	{
 		return false;
@@ -1605,7 +1604,7 @@ typeinfo_init_component(typeinfo *srcarray,typeinfo *dst)
 	else {
 		vftbl_t *comp;
 		
-		if (!srcarray->typeclass.cls->linked) {
+		if (!(srcarray->typeclass.cls->state & CLASS_LINKED)) {
 			if (!link_class(srcarray->typeclass.cls)) {
 				return false;
 			}
@@ -1973,8 +1972,8 @@ typeinfo_merge_nonarrays(typeinfo *dest,
 	{
 		typeinfo dbgx,dbgy;
 		fprintf(stderr,"merge_nonarrays:\n");
-		fprintf(stderr,"    ");if(IS_CLASSREF(x))fprintf(stderr,"<ref>");utf_fprint(stderr,xname);fprintf(stderr,"\n");
-		fprintf(stderr,"    ");if(IS_CLASSREF(y))fprintf(stderr,"<ref>");utf_fprint(stderr,yname);fprintf(stderr,"\n");
+		fprintf(stderr,"    ");if(IS_CLASSREF(x))fprintf(stderr,"<ref>");utf_fprint_printable_ascii(stderr,xname);fprintf(stderr,"\n");
+		fprintf(stderr,"    ");if(IS_CLASSREF(y))fprintf(stderr,"<ref>");utf_fprint_printable_ascii(stderr,yname);fprintf(stderr,"\n");
 		fflush(stderr);
 		typeinfo_init_class(&dbgx,x);
 		dbgx.merged = mergedx;
@@ -1986,8 +1985,8 @@ typeinfo_merge_nonarrays(typeinfo *dest,
 	}
 #endif
 
-	TYPEINFO_ASSERT(IS_CLASSREF(x) || x.cls->loaded);
-	TYPEINFO_ASSERT(IS_CLASSREF(y) || y.cls->loaded);
+	TYPEINFO_ASSERT(IS_CLASSREF(x) || (x.cls->state & CLASS_LOADED));
+	TYPEINFO_ASSERT(IS_CLASSREF(y) || (y.cls->state & CLASS_LOADED));
 
     /* If y is unresolved or an interface, swap x and y. */
     if (IS_CLASSREF(y) || (!IS_CLASSREF(x) && y.cls->flags & ACC_INTERFACE))
@@ -2016,8 +2015,8 @@ typeinfo_merge_nonarrays(typeinfo *dest,
     /* {We know: If only one of x,y is an interface it is x.} */
 
 	TYPEINFO_ASSERT(!IS_CLASSREF(x) && !IS_CLASSREF(y));
-	TYPEINFO_ASSERT(x.cls->loaded);
-	TYPEINFO_ASSERT(y.cls->loaded);
+	TYPEINFO_ASSERT(x.cls->state & CLASS_LOADED);
+	TYPEINFO_ASSERT(y.cls->state & CLASS_LOADED);
 
     /* Handle merging of interfaces: */
     if (x.cls->flags & ACC_INTERFACE) {
@@ -2045,15 +2044,15 @@ typeinfo_merge_nonarrays(typeinfo *dest,
          */
 
 		/* we may have to link the classes */
-		if (!x.cls->linked)
+		if (!(x.cls->state & CLASS_LINKED))
 			if (!link_class(x.cls))
 				return typecheck_FAIL;
-		if (!y.cls->linked)
+		if (!(y.cls->state & CLASS_LINKED))
 			if (!link_class(y.cls))
 				return typecheck_FAIL;
         
-		TYPEINFO_ASSERT(x.cls->linked);
-		TYPEINFO_ASSERT(y.cls->linked);
+		TYPEINFO_ASSERT(x.cls->state & CLASS_LINKED);
+		TYPEINFO_ASSERT(y.cls->state & CLASS_LINKED);
 
         if (CLASSINFO_IMPLEMENTS_INTERFACE(y.cls,x.cls->index))
 		{
@@ -2080,15 +2079,15 @@ typeinfo_merge_nonarrays(typeinfo *dest,
     /* {We know: x and y are classes (not interfaces).} */
     
 	/* we may have to link the classes */
-	if (!x.cls->linked)
+	if (!(x.cls->state & CLASS_LINKED))
 		if (!link_class(x.cls))
 			return typecheck_FAIL;
-	if (!y.cls->linked)
+	if (!(y.cls->state & CLASS_LINKED))
 		if (!link_class(y.cls))
 			return typecheck_FAIL;
         
-	TYPEINFO_ASSERT(x.cls->linked);
-	TYPEINFO_ASSERT(y.cls->linked);
+	TYPEINFO_ASSERT(x.cls->state & CLASS_LINKED);
+	TYPEINFO_ASSERT(y.cls->state & CLASS_LINKED);
 
     /* If *x is deeper in the inheritance hierarchy swap x and y. */
     if (x.cls->index > y.cls->index) {
@@ -2333,7 +2332,7 @@ return_simple:
 						return typecheck_FAIL;
 					}
 				}
-                /* DEBUG */ /* utf_display(common->name); printf("\n"); */
+                /* DEBUG */ /* utf_display_printable_ascii(common->name); printf("\n"); */
             }
 			else {
 				common.any = y->typeclass.any;
@@ -2618,10 +2617,10 @@ typeinfo_print_class(FILE *file,classref_or_classinfo c)
 	else {
 		if (IS_CLASSREF(c)) {
 			fprintf(file,"<ref>");
-			utf_fprint(file,c.ref->name);
+			utf_fprint_printable_ascii(file,c.ref->name);
 		}
 		else {
-			utf_fprint(file,c.cls->name);
+			utf_fprint_printable_ascii(file,c.cls->name);
 		}
 	}
 }
@@ -2658,7 +2657,7 @@ typeinfo_print(FILE *file,typeinfo *info,int indent)
         ins = (instruction *)TYPEINFO_NEWOBJECT_INSTRUCTION(*info);
         if (ins) {
             fprintf(file,"%sNEW(%p):",ind,(void*)ins);
-			typeinfo_print_class(file,CLASSREF_OR_CLASSINFO(ins[-1].target));
+			typeinfo_print_class(file,ICMD_ACONST_CLASSREF_OR_CLASSINFO(ins-1));
             fprintf(file,"\n");
         }
         else {
@@ -2737,7 +2736,7 @@ typeinfo_print_short(FILE *file,typeinfo *info)
         if (ins) {
 			/*fprintf(file,"<ins %p>",ins);*/
             fprintf(file,"NEW(%p):",(void*)ins);
-			typeinfo_print_class(file,CLASSREF_OR_CLASSINFO(ins[-1].target));
+			typeinfo_print_class(file,ICMD_ACONST_CLASSREF_OR_CLASSINFO(ins-1));
         }
         else
             fprintf(file,"NEW(this)");
