@@ -1,9 +1,9 @@
 /* typeinfo.h - type system used by the type checker
 
-   Copyright (C) 1996-2005 R. Grafl, A. Krall, C. Kruegel, C. Oates,
-   R. Obermaisser, M. Platter, M. Probst, S. Ring, E. Steiner,
-   C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich, J. Wenninger,
-   Institut f. Computersprachen - TU Wien
+   Copyright (C) 1996-2005, 2006 R. Grafl, A. Krall, C. Kruegel,
+   C. Oates, R. Obermaisser, M. Platter, M. Probst, S. Ring,
+   E. Steiner, C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich,
+   J. Wenninger, Institut f. Computersprachen - TU Wien
 
    This file is part of CACAO.
 
@@ -19,23 +19,27 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-   02111-1307, USA.
+   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
+   02110-1301, USA.
 
-   Contact: cacao@complang.tuwien.ac.at
+   Contact: cacao@cacaojvm.org
 
    Authors: Edwin Steiner
 
-   $Id: typeinfo.h,v 3.21 2005/11/08 19:01:17 edwin Exp $
+   $Id: typeinfo.h 4863 2006-04-30 16:17:44Z edwin $
 
 */
 
 #ifndef _TYPEINFO_H
 #define _TYPEINFO_H
 
+#include "config.h"
+#include "vm/types.h"
+
 #include "vm/global.h"
 #include "vm/references.h"
 
+
 /* configuration **************************************************************/
 
 /*
@@ -47,7 +51,7 @@
  * TYPECHECK_VERBOSE activates all debug messages
  * TYPEINFO_VERBOSE activates debug prints in typeinfo.c
  */
-#ifdef CACAO_TYPECHECK
+#ifdef ENABLE_VERIFIER
 #ifndef NDEBUG
 /*#define TYPECHECK_STATISTICS*/
 #define TYPEINFO_DEBUG
@@ -63,7 +67,7 @@
 #endif
 
 #ifdef TYPECHECK_VERBOSE_OPT
-extern bool typecheckverbose;
+extern bool opt_typecheckverbose;
 #endif
 
 /* resolve typedef cycles *****************************************************/
@@ -442,9 +446,9 @@ struct typevector {
              (info).dimension = 0;                      \
              (info).elementtype = 0;} while(0)
 
-#define TYPEINFO_INIT_JAVA_LANG_CLASS(info,cr)                  \
+#define TYPEINFO_INIT_JAVA_LANG_CLASS(info,c)                   \
          do {(info).typeclass.any = class_java_lang_Class;      \
-             (info).elementclass.ref = (cr);                    \
+             (info).elementclass = (c);                         \
              (info).merged = NULL;                              \
              (info).dimension = 0;                              \
              (info).elementtype = 0;} while(0)
