@@ -1,9 +1,9 @@
 /* src/vm/jit/x86_64/patcher.c - x86_64 code patching functions
 
-   Copyright (C) 1996-2005 R. Grafl, A. Krall, C. Kruegel, C. Oates,
-   R. Obermaisser, M. Platter, M. Probst, S. Ring, E. Steiner,
-   C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich, J. Wenninger,
-   Institut f. Computersprachen - TU Wien
+   Copyright (C) 1996-2005, 2006 R. Grafl, A. Krall, C. Kruegel,
+   C. Oates, R. Obermaisser, M. Platter, M. Probst, S. Ring,
+   E. Steiner, C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich,
+   J. Wenninger, Institut f. Computersprachen - TU Wien
 
    This file is part of CACAO.
 
@@ -19,16 +19,16 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-   02111-1307, USA.
+   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
+   02110-1301, USA.
 
-   Contact: cacao@complang.tuwien.ac.at
+   Contact: cacao@cacaojvm.org
 
    Authors: Christian Thalinger
 
    Changes:
 
-   $Id: patcher.c,v 1.33 2005/11/18 00:30:10 twisti Exp $
+   $Id: patcher.c 4530 2006-02-21 09:11:53Z twisti $
 
 */
 
@@ -39,12 +39,12 @@
 #include "mm/memory.h"
 #include "native/native.h"
 #include "vm/builtin.h"
-#include "vm/exceptions.h"
+#include "vm/class.h"
 #include "vm/field.h"
 #include "vm/initialize.h"
 #include "vm/options.h"
-#include "vm/resolve.h"
 #include "vm/references.h"
+#include "vm/resolve.h"
 #include "vm/jit/patcher.h"
 
 
@@ -93,10 +93,12 @@ bool patcher_get_putstatic(u1 *sp)
 
 	/* check if the field's class is initialized */
 
-	if (!initialize_class(fi->class)) {
-		PATCHER_MONITOREXIT;
+	if (!(fi->class->state & CLASS_INITIALIZED)) {
+		if (!initialize_class(fi->class)) {
+			PATCHER_MONITOREXIT;
 
-		return false;
+			return false;
+		}
 	}
 
 	/* patch back original code */
@@ -318,12 +320,6 @@ bool patcher_aconst(u1 *sp)
 		return false;
 	}
 
-	if (!use_class_as_object(c)) {
-		PATCHER_MONITOREXIT;
-
-		return false;
-	}
-
 	/* patch back original code */
 
 	*((u8 *) ra) = mcode;
@@ -937,10 +933,12 @@ bool patcher_clinit(u1 *sp)
 
 	/* check if the class is initialized */
 
-	if (!initialize_class(c)) {
-		PATCHER_MONITOREXIT;
+	if (!(c->state & CLASS_INITIALIZED)) {
+		if (!initialize_class(c)) {
+			PATCHER_MONITOREXIT;
 
-		return false;
+			return false;
+		}
 	}
 
 	/* patch back original code */
@@ -961,6 +959,7 @@ bool patcher_clinit(u1 *sp)
 
 *******************************************************************************/
 
+#ifdef ENABLE_VERIFIER
 bool patcher_athrow_areturn(u1 *sp)
 {
 	u1                *ra;
@@ -999,6 +998,7 @@ bool patcher_athrow_areturn(u1 *sp)
 
 	return true;
 }
+#endif /* ENABLE_VERIFIER */
 
 
 /* patcher_resolve_native ******************************************************
@@ -1011,7 +1011,7 @@ bool patcher_athrow_areturn(u1 *sp)
 
 *******************************************************************************/
 
-#if !defined(ENABLE_STATICVM)
+#if !defined(WITH_STATIC_CLASSPATH)
 bool patcher_resolve_native(u1 *sp)
 {
 	u1                *ra;
@@ -1059,7 +1059,7 @@ bool patcher_resolve_native(u1 *sp)
 
 	return true;
 }
-#endif /* !defined(ENABLE_STATICVM) */
+#endif /* !defined(WITH_STATIC_CLASSPATH) */
 
 
 /*
