@@ -1,9 +1,9 @@
 /* src/vm/linker.c - class linker functions
 
-   Copyright (C) 1996-2005 R. Grafl, A. Krall, C. Kruegel, C. Oates,
-   R. Obermaisser, M. Platter, M. Probst, S. Ring, E. Steiner,
-   C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich, J. Wenninger,
-   Institut f. Computersprachen - TU Wien
+   Copyright (C) 1996-2005, 2006 R. Grafl, A. Krall, C. Kruegel,
+   C. Oates, R. Obermaisser, M. Platter, M. Probst, S. Ring,
+   E. Steiner, C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich,
+   J. Wenninger, Institut f. Computersprachen - TU Wien
 
    This file is part of CACAO.
 
@@ -19,10 +19,10 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-   02111-1307, USA.
+   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
+   02110-1301, USA.
 
-   Contact: cacao@complang.tuwien.ac.at
+   Contact: cacao@cacaojvm.org
 
    Authors: Reinhard Grafl
 
@@ -32,14 +32,15 @@
             Edwin Steiner
             Christian Thalinger
 
-   $Id: linker.c,v 1.33 2005/11/16 13:28:59 twisti Exp $
+   $Id: linker.c 4973 2006-05-29 15:21:46Z edwin $
 
 */
 
 
+#include "config.h"
+
 #include <assert.h>
 
-#include "config.h"
 #include "vm/types.h"
 
 #include "mm/memory.h"
@@ -53,7 +54,8 @@
 #include "vm/resolve.h"
 #include "vm/statistics.h"
 #include "vm/stringlocal.h"
-#include "vm/jit/codegen.inc.h"
+#include "vm/access.h"
+#include "vm/rt-timing.h"
 
 
 /* global variables ***********************************************************/
@@ -96,7 +98,7 @@ static classinfo *link_class_intern(classinfo *c);
 static arraydescriptor *link_array(classinfo *c);
 static void linker_compute_class_values(classinfo *c);
 static void linker_compute_subclasses(classinfo *c);
-static void linker_addinterface(classinfo *c, classinfo *ic);
+static bool linker_addinterface(classinfo *c, classinfo *ic);
 static s4 class_highestinterface(classinfo *c);
 
 
@@ -112,6 +114,19 @@ bool linker_init(void)
 
 	interfaceindex = 0;
 
+	/* link java.lang.Class as first class of the system, because we
+       need it's vftbl for all other classes so we can use a class as
+       object */
+
+	if (!link_class(class_java_lang_Class))
+		return false;
+
+	/* now set the header.vftbl of all classes which were created
+       before java.lang.Class was linked */
+
+	class_postset_header_vftbl();
+
+
 	/* link important system classes */
 
 	if (!link_class(class_java_lang_Object))
@@ -159,9 +174,6 @@ bool linker_init(void)
 
 	/* load some other important classes */
 
-	if (!link_class(class_java_lang_Class))
-		return false;
-
 	if (!link_class(class_java_lang_ClassLoader))
 		return false;
 
@@ -209,14 +221,14 @@ bool linker_init(void)
 
     /* pseudo class for Arraystubs (extends java.lang.Object) */
     
-    pseudo_class_Arraystub =
+	pseudo_class_Arraystub =
 		class_create_classinfo(utf_new_char("$ARRAYSTUB$"));
-	pseudo_class_Arraystub->loaded = true;
-    pseudo_class_Arraystub->super.cls = class_java_lang_Object;
-    pseudo_class_Arraystub->interfacescount = 2;
-    pseudo_class_Arraystub->interfaces = MNEW(classref_or_classinfo, 2);
-    pseudo_class_Arraystub->interfaces[0].cls = class_java_lang_Cloneable;
-    pseudo_class_Arraystub->interfaces[1].cls = class_java_io_Serializable;
+	pseudo_class_Arraystub->state |= CLASS_LOADED;
+	pseudo_class_Arraystub->super.cls = class_java_lang_Object;
+	pseudo_class_Arraystub->interfacescount = 2;
+	pseudo_class_Arraystub->interfaces = MNEW(classref_or_classinfo, 2);
+	pseudo_class_Arraystub->interfaces[0].cls = class_java_lang_Cloneable;
+	pseudo_class_Arraystub->interfaces[1].cls = class_java_io_Serializable;
 
 	if (!classcache_store_unique(pseudo_class_Arraystub)) {
 		log_text("could not cache pseudo_class_Arraystub");
@@ -227,9 +239,9 @@ bool linker_init(void)
 		return false;
 
 	/* pseudo class representing the null type */
-    
+
 	pseudo_class_Null = class_create_classinfo(utf_new_char("$NULL$"));
-	pseudo_class_Null->loaded = true;
+	pseudo_class_Null->state |= CLASS_LOADED;
 	pseudo_class_Null->super.cls = class_java_lang_Object;
 
 	if (!classcache_store_unique(pseudo_class_Null)) {
@@ -243,8 +255,8 @@ bool linker_init(void)
 	/* pseudo class representing new uninitialized objects */
     
 	pseudo_class_New = class_create_classinfo(utf_new_char("$NEW$"));
-	pseudo_class_New->loaded = true;
-	pseudo_class_New->linked = true; /* XXX is this allright? */
+	pseudo_class_New->state |= CLASS_LOADED;
+	pseudo_class_New->state |= CLASS_LINKED; /* XXX is this allright? */
 	pseudo_class_New->super.cls = class_java_lang_Object;
 
 	if (!classcache_store_unique(pseudo_class_New)) {
@@ -290,12 +302,10 @@ static bool link_primitivetype_table(void)
 		c = class_create_classinfo(utf_new_char(primitivetype_table[i].name));
 
 		c->flags = ACC_PUBLIC | ACC_FINAL | ACC_ABSTRACT;
-		c->classUsed = NOTUSED; /* not used initially CO-RT */
-		c->impldBy = NULL;
 		
 		/* prevent loader from loading primitive class */
 
-		c->loaded = true;
+		c->state |= CLASS_LOADED;
 
 		/* INFO: don't put primitive classes into the classcache */
 
@@ -312,8 +322,6 @@ static bool link_primitivetype_table(void)
 			return false;
 
 		primitivetype_table[i].class_wrap = c;
-		primitivetype_table[i].class_wrap->classUsed = NOTUSED; /* not used initially CO-RT */
-		primitivetype_table[i].class_wrap->impldBy = NULL;
 
 		/* create the primitive array class */
 
@@ -326,8 +334,9 @@ static bool link_primitivetype_table(void)
 
 			primitivetype_table[i].arrayclass = c;
 
-			assert(c->loaded);
-			if (!c->linked)
+			assert(c->state & CLASS_LOADED);
+
+			if (!(c->state & CLASS_LINKED))
 				if (!link_class(c))
 					return false;
 
@@ -349,34 +358,40 @@ static bool link_primitivetype_table(void)
 classinfo *link_class(classinfo *c)
 {
 	classinfo *r;
+#if defined(ENABLE_RT_TIMING)
+	struct timespec time_start, time_end;
+#endif
+
+	RT_TIMING_GET_TIME(time_start);
 
 	if (!c) {
-		*exceptionptr = new_nullpointerexception();
+		exceptions_throw_nullpointerexception();
 		return NULL;
 	}
 
-#if defined(USE_THREADS)
+#if defined(ENABLE_THREADS)
 	/* enter a monitor on the class */
 
 	builtin_monitorenter((java_objectheader *) c);
 #endif
 
 	/* maybe the class is already linked */
-	if (c->linked) {
-#if defined(USE_THREADS)
+
+	if (c->state & CLASS_LINKED) {
+#if defined(ENABLE_THREADS)
 		builtin_monitorexit((java_objectheader *) c);
 #endif
 
 		return c;
 	}
 
-#if defined(STATISTICS)
+#if defined(ENABLE_STATISTICS)
 	/* measure time */
 
-	if (getcompilingtime)
+	if (opt_getcompilingtime)
 		compilingtime_stop();
 
-	if (getloadingtime)
+	if (opt_getloadingtime)
 		loadingtime_start();
 #endif
 
@@ -387,24 +402,28 @@ classinfo *link_class(classinfo *c)
 	/* if return value is NULL, we had a problem and the class is not linked */
 
 	if (!r)
-		c->linked = false;
+		c->state &= ~CLASS_LINKING;
 
-#if defined(STATISTICS)
+#if defined(ENABLE_STATISTICS)
 	/* measure time */
 
-	if (getloadingtime)
+	if (opt_getloadingtime)
 		loadingtime_stop();
 
-	if (getcompilingtime)
+	if (opt_getcompilingtime)
 		compilingtime_start();
 #endif
 
-#if defined(USE_THREADS)
+#if defined(ENABLE_THREADS)
 	/* leave the monitor */
 
 	builtin_monitorexit((java_objectheader *) c);
 #endif
 
+	RT_TIMING_GET_TIME(time_end);
+
+	RT_TIMING_TIME_DIFF(time_start,time_end,RT_TIMING_LINK_TOTAL);
+
 	return r;
 }
 
@@ -427,20 +446,29 @@ static classinfo *link_class_intern(classinfo *c)
 	vftbl_t *v;                   /* vftbl of current class                   */
 	s4 i,j;                       /* interface/method/field counter           */
 	arraydescriptor *arraydesc;   /* descriptor for array classes             */
+#if defined(ENABLE_RT_TIMING)
+	struct timespec time_start, time_resolving, time_compute_vftbl,
+					time_abstract, time_compute_iftbl, time_fill_vftbl,
+					time_offsets, time_fill_iftbl, time_finalizer,
+					time_exceptions, time_subclasses;
+#endif
 
-	/* maybe the class is already linked */
+	RT_TIMING_GET_TIME(time_start);
 
-	if (c->linked)
+	/* the class is already linked */
+
+	if (c->state & CLASS_LINKED)
 		return c;
 
+#if !defined(NDEBUG)
 	if (linkverbose)
 		log_message_class("Linking class: ", c);
+#endif
 
 	/* the class must be loaded */
 
-	if (!c->loaded)
-		throw_cacao_exception_exit(string_java_lang_InternalError,
-								   "Trying to link unloaded class");
+	/* XXX should this be a specific exception? */
+	assert(c->state & CLASS_LOADED);
 
 	/* cache the self-reference of this class                          */
 	/* we do this for cases where the defining loader of the class     */
@@ -452,9 +480,9 @@ static classinfo *link_class_intern(classinfo *c)
 	if (c->classloader)
 		classcache_store(c->classloader,c,false);
 
-	/* ok, this class is somewhat linked */
+	/* this class is currently linking */
 
-	c->linked = true;
+	c->state |= CLASS_LINKING;
 
 	arraydesc = NULL;
 
@@ -478,7 +506,7 @@ static classinfo *link_class_intern(classinfo *c)
 			return NULL;
 		}
 
-		assert(tc->loaded);
+		assert(tc->state & CLASS_LOADED);
 
 		if (!(tc->flags & ACC_INTERFACE)) {
 			*exceptionptr =
@@ -487,7 +515,7 @@ static classinfo *link_class_intern(classinfo *c)
 			return NULL;
 		}
 
-		if (!tc->linked)
+		if (!(tc->state & CLASS_LINKED))
 			if (!link_class(tc))
 				return NULL;
 	}
@@ -498,8 +526,6 @@ static classinfo *link_class_intern(classinfo *c)
 
 	if (c->super.any == NULL) {                     /* class java.lang.Object */
 		c->index = 0;
-        c->classUsed = USED;              /* Object class is always used CO-RT*/
-		c->impldBy = NULL;
 		c->instancesize = sizeof(java_objectheader);
 		
 		vftbllength = supervftbllength = 0;
@@ -523,7 +549,7 @@ static classinfo *link_class_intern(classinfo *c)
 			return NULL;
 		}
 
-		assert(super->loaded);
+		assert(super->state & CLASS_LOADED);
 
 		if (super->flags & ACC_INTERFACE) {
 			/* java.lang.IncompatibleClassChangeError: class a has interface java.lang.Cloneable as super class */
@@ -539,11 +565,17 @@ static classinfo *link_class_intern(classinfo *c)
 									  "Cannot inherit from final class");
 			return NULL;
 		}
+
+		/* link the superclass if necessary */
 		
-		if (!super->linked)
+		if (!(super->state & CLASS_LINKED))
 			if (!link_class(super))
 				return NULL;
 
+		/* OR the ACC_CLASS_HAS_POINTERS flag */
+
+		c->flags |= (super->flags & ACC_CLASS_HAS_POINTERS);
+
 		/* handle array classes */
 
 		if (c->name->text[0] == '[')
@@ -556,11 +588,12 @@ static classinfo *link_class_intern(classinfo *c)
 			c->index = super->index + 1;
 		
 		c->instancesize = super->instancesize;
-		
+
 		vftbllength = supervftbllength = super->vftbl->vftbllength;
 		
 		c->finalizer = super->finalizer;
 	}
+	RT_TIMING_GET_TIME(time_resolving);
 
 
 	/* compute vftbl length */
@@ -579,6 +612,15 @@ static classinfo *link_class_intern(classinfo *c)
 						if (tc->methods[j].flags & ACC_PRIVATE)
 							goto notfoundvftblindex;
 
+						/* package-private methods in other packages */
+						/* must not be overridden                    */
+						/* (see Java Language Specification 8.4.8.1) */
+						if ( !(tc->methods[j].flags & (ACC_PUBLIC | ACC_PROTECTED)) 
+							 && !SAME_PACKAGE(c,tc) ) 
+						{
+						    goto notfoundvftblindex;
+						}
+
 						if (tc->methods[j].flags & ACC_FINAL) {
 							/* class a overrides final method . */
 							*exceptionptr =
@@ -586,6 +628,21 @@ static classinfo *link_class_intern(classinfo *c)
 							return NULL;
 						}
 
+						/* method m overwrites method j of class tc */
+
+#if defined(ENABLE_VERIFIER)
+						/* Add loading constraints (for the more general    */
+						/* types of method tc->methods[j]). --              */
+						/* Not for <init>,  as it is not invoked virtually. */
+						if ((m->name != utf_init)
+							&& !classcache_add_constraints_for_params(
+									c->classloader, tc->classloader,
+									&(tc->methods[j])))
+						{
+							return NULL;
+						}
+#endif
+
 						m->vftblindex = tc->methods[j].vftblindex;
 						goto foundvftblindex;
 					}
@@ -599,10 +656,14 @@ static classinfo *link_class_intern(classinfo *c)
 		foundvftblindex:
 			;
 		}
-	}	
+	}
+	RT_TIMING_GET_TIME(time_compute_vftbl);
 
 
-	/* check interfaces of ABSTRACT class for unimplemented methods */
+	/* Check all interfaces of an abtract class (maybe be an interface
+	   too) for unimplemented methods.  Such methods are called
+	   miranda-methods and are marked with the ACC_MIRANDA flag.
+	   VMClass.getDeclaredMethods does not return such methods. */
 
 	if (c->flags & ACC_ABSTRACT) {
 		classinfo  *ic;
@@ -613,6 +674,8 @@ static classinfo *link_class_intern(classinfo *c)
 
 		abstractmethodscount = 0;
 
+		/* check all interfaces of the abtract class */
+
 		for (i = 0; i < c->interfacescount; i++) {
 			ic = c->interfaces[i].cls;
 
@@ -621,18 +684,14 @@ static classinfo *link_class_intern(classinfo *c)
 
 				/* skip `<clinit>' and `<init>' */
 
-				if (im->name == utf_clinit || im->name == utf_init)
+				if ((im->name == utf_clinit) || (im->name == utf_init))
 					continue;
 
-				tc = c;
-
-				while (tc) {
+				for (tc = c; tc != NULL; tc = tc->super.cls) {
 					for (k = 0; k < tc->methodscount; k++) {
 						if (method_canoverwrite(im, &(tc->methods[k])))
 							goto noabstractmethod;
 					}
-
-					tc = tc->super.cls;
 				}
 
 				abstractmethodscount++;
@@ -658,27 +717,27 @@ static classinfo *link_class_intern(classinfo *c)
 
 					/* skip `<clinit>' and `<init>' */
 
-					if (im->name == utf_clinit || im->name == utf_init)
+					if ((im->name == utf_clinit) || (im->name == utf_init))
 						continue;
 
-					tc = c;
-
-					while (tc) {
+					for (tc = c; tc != NULL; tc = tc->super.cls) {
 						for (k = 0; k < tc->methodscount; k++) {
 							if (method_canoverwrite(im, &(tc->methods[k])))
 								goto noabstractmethod2;
 						}
-
-						tc = tc->super.cls;
 					}
 
+					/* Copy the method found into the new c->methods
+					   array and tag it as miranda-method. */
+
 					am = &(c->methods[c->methodscount]);
 					c->methodscount++;
 
 					MCOPY(am, im, methodinfo, 1);
 
-					am->vftblindex = (vftbllength++);
-					am->class = c;
+					am->vftblindex  = (vftbllength++);
+					am->class       = c;
+					am->flags      |= ACC_MIRANDA;
 
 				noabstractmethod2:
 					;
@@ -686,9 +745,10 @@ static classinfo *link_class_intern(classinfo *c)
 			}
 		}
 	}
+	RT_TIMING_GET_TIME(time_abstract);
 
 
-#if defined(STATISTICS)
+#if defined(ENABLE_STATISTICS)
 	if (opt_stat)
 		count_vftbl_len +=
 			sizeof(vftbl_t) + (sizeof(methodptr) * (vftbllength - 1));
@@ -706,6 +766,7 @@ static classinfo *link_class_intern(classinfo *c)
 		}
 		tc = tc->super.cls;
 	}
+	RT_TIMING_GET_TIME(time_compute_iftbl);
 
 	/* allocate virtual function table */
 
@@ -714,7 +775,7 @@ static classinfo *link_class_intern(classinfo *c)
 							  sizeof(methodptr*) * (interfacetablelength - (interfacetablelength > 0)));
 	v = (vftbl_t *) (((methodptr *) v) +
 					 (interfacetablelength - 1) * (interfacetablelength > 1));
-	c->header.vftbl = c->vftbl = v;
+	c->vftbl = v;
 	v->class = c;
 	v->vftbllength = vftbllength;
 	v->interfacetablelength = interfacetablelength;
@@ -735,15 +796,26 @@ static classinfo *link_class_intern(classinfo *c)
 	for (i = 0; i < c->methodscount; i++) {
 		methodinfo *m = &(c->methods[i]);
 
-		/* Methods in ABSTRACT classes from interfaces maybe already have a   */
-		/* stubroutine.                                                       */
+		/* Methods in ABSTRACT classes from interfaces maybe already
+		   have a stubroutine. */
 
-		if (!m->stubroutine)
-			m->stubroutine = createcompilerstub(m);
+		if (!m->stubroutine) {
+#if defined(ENABLE_JIT)
+# if defined(ENABLE_INTRP)
+			if (opt_intrp)
+				m->stubroutine = intrp_createcompilerstub(m);
+			else
+#endif
+				m->stubroutine = createcompilerstub(m);
+#else
+			m->stubroutine = intrp_createcompilerstub(m);
+#endif
+		}
 
 		if (!(m->flags & ACC_STATIC))
 			v->table[m->vftblindex] = (methodptr) (ptrint) m->stubroutine;
 	}
+	RT_TIMING_GET_TIME(time_fill_vftbl);
 
 	/* compute instance size and offset of each field */
 	
@@ -752,18 +824,30 @@ static classinfo *link_class_intern(classinfo *c)
 		fieldinfo *f = &(c->fields[i]);
 		
 		if (!(f->flags & ACC_STATIC)) {
-			dsize = desc_typesize(f->descriptor);
+			dsize = descriptor_typesize(f->parseddesc);
+
+			/* On i386 we only align to 4 bytes even for double and s8.    */
+			/* This matches what gcc does for struct members. We must      */
+			/* do the same as gcc here because the offsets in native       */
+			/* header structs like java_lang_Double must match the offsets */
+			/* of the Java fields (eg. java.lang.Double.value).            */
+#if defined(__I386__)
+			c->instancesize = ALIGN(c->instancesize, 4);
+#else
 			c->instancesize = ALIGN(c->instancesize, dsize);
+#endif
+
 			f->offset = c->instancesize;
 			c->instancesize += dsize;
 		}
 	}
+	RT_TIMING_GET_TIME(time_offsets);
 
 	/* initialize interfacetable and interfacevftbllength */
-	
+
 	v->interfacevftbllength = MNEW(s4, interfacetablelength);
 
-#if defined(STATISTICS)
+#if defined(ENABLE_STATISTICS)
 	if (opt_stat)
 		count_vftbl_len += (4 + sizeof(s4)) * v->interfacetablelength;
 #endif
@@ -772,12 +856,15 @@ static classinfo *link_class_intern(classinfo *c)
 		v->interfacevftbllength[i] = 0;
 		v->interfacetable[-i] = NULL;
 	}
-	
+
 	/* add interfaces */
-	
+
 	for (tc = c; tc != NULL; tc = tc->super.cls)
 		for (i = 0; i < tc->interfacescount; i++)
-			linker_addinterface(c, tc->interfaces[i].cls);
+			if (!linker_addinterface(c, tc->interfaces[i].cls))
+				return NULL;
+
+	RT_TIMING_GET_TIME(time_fill_iftbl);
 
 	/* add finalizer method (not for java.lang.Object) */
 
@@ -790,6 +877,7 @@ static classinfo *link_class_intern(classinfo *c)
 			if (!(fi->flags & ACC_STATIC))
 				c->finalizer = fi;
 	}
+	RT_TIMING_GET_TIME(time_finalizer);
 
 	/* resolve exception class references */
 
@@ -806,13 +894,33 @@ static classinfo *link_class_intern(classinfo *c)
 				return NULL;
 		}
 	}
+	RT_TIMING_GET_TIME(time_exceptions);
 	
 	/* final tasks */
 
 	linker_compute_subclasses(c);
 
+	RT_TIMING_GET_TIME(time_subclasses);
+
+	/* revert the linking state and class is linked */
+
+	c->state = (c->state & ~CLASS_LINKING) | CLASS_LINKED;
+
+#if !defined(NDEBUG)
 	if (linkverbose)
 		log_message_class("Linking done class: ", c);
+#endif
+
+	RT_TIMING_TIME_DIFF(time_start        ,time_resolving    ,RT_TIMING_LINK_RESOLVE);
+	RT_TIMING_TIME_DIFF(time_resolving    ,time_compute_vftbl,RT_TIMING_LINK_C_VFTBL);
+	RT_TIMING_TIME_DIFF(time_compute_vftbl,time_abstract     ,RT_TIMING_LINK_ABSTRACT);
+	RT_TIMING_TIME_DIFF(time_abstract     ,time_compute_iftbl,RT_TIMING_LINK_C_IFTBL);
+	RT_TIMING_TIME_DIFF(time_compute_iftbl,time_fill_vftbl   ,RT_TIMING_LINK_F_VFTBL);
+	RT_TIMING_TIME_DIFF(time_fill_vftbl   ,time_offsets      ,RT_TIMING_LINK_OFFSETS);
+	RT_TIMING_TIME_DIFF(time_offsets      ,time_fill_iftbl   ,RT_TIMING_LINK_F_IFTBL);
+	RT_TIMING_TIME_DIFF(time_fill_iftbl   ,time_finalizer    ,RT_TIMING_LINK_FINALIZER);
+	RT_TIMING_TIME_DIFF(time_finalizer    ,time_exceptions   ,RT_TIMING_LINK_EXCEPTS);
+	RT_TIMING_TIME_DIFF(time_exceptions   ,time_subclasses   ,RT_TIMING_LINK_SUBCLASS);
 
 	/* just return c to show that we didn't had a problem */
 
@@ -860,14 +968,15 @@ static arraydescriptor *link_array(classinfo *c)
 	}
 
 	/* If the component type has not been linked, link it now */
-	assert(!comp || comp->loaded);
-	if (comp && !comp->linked) {
 
+	assert(!comp || (comp->state & CLASS_LOADED));
+
+	if (comp && !(comp->state & CLASS_LINKED))
 		if (!link_class(comp))
 			return NULL;
-	}
 
 	/* Allocate the arraydescriptor */
+
 	desc = NEW(arraydescriptor);
 
 	if (comp) {
@@ -976,12 +1085,8 @@ static arraydescriptor *link_array(classinfo *c)
 
 static void linker_compute_subclasses(classinfo *c)
 {
-#if defined(USE_THREADS)
-#if defined(NATIVE_THREADS)
+#if defined(ENABLE_THREADS)
 	compiler_lock();
-#else
-	intsDisable();
-#endif
 #endif
 
 	if (!(c->flags & ACC_INTERFACE)) {
@@ -1000,12 +1105,8 @@ static void linker_compute_subclasses(classinfo *c)
 
 	linker_compute_class_values(class_java_lang_Object);
 
-#if defined(USE_THREADS)
-#if defined(NATIVE_THREADS)
+#if defined(ENABLE_THREADS)
 	compiler_unlock();
-#else
-	intsRestore();
-#endif
 #endif
 }
 
@@ -1039,9 +1140,13 @@ static void linker_compute_class_values(classinfo *c)
    Is needed by link_class for adding a VTBL to a class. All
    interfaces implemented by ic are added as well.
 
+   RETURN VALUE:
+      true.........everything ok
+	  false........an exception has been thrown
+
 *******************************************************************************/
 
-static void linker_addinterface(classinfo *c, classinfo *ic)
+static bool linker_addinterface(classinfo *c, classinfo *ic)
 {
 	s4     j, m;
 	s4     i   = ic->index;
@@ -1052,19 +1157,22 @@ static void linker_addinterface(classinfo *c, classinfo *ic)
 		assert(0);
 	}
 
+	/* if this interface has already been added, return immediately */
+
 	if (v->interfacetable[-i])
-		return;
+		return true;
 
 	if (ic->methodscount == 0) {  /* fake entry needed for subtype test */
 		v->interfacevftbllength[i] = 1;
 		v->interfacetable[-i] = MNEW(methodptr, 1);
 		v->interfacetable[-i][0] = NULL;
 
-	} else {
+	}
+	else {
 		v->interfacevftbllength[i] = ic->methodscount;
 		v->interfacetable[-i] = MNEW(methodptr, ic->methodscount);
 
-#if defined(STATISTICS)
+#if defined(ENABLE_STATISTICS)
 		if (opt_stat)
 			count_vftbl_len += sizeof(methodptr) *
 				(ic->methodscount + (ic->methodscount == 0));
@@ -1078,6 +1186,17 @@ static void linker_addinterface(classinfo *c, classinfo *ic)
 					methodinfo *mi = &(sc->methods[m]);
 
 					if (method_canoverwrite(mi, &(ic->methods[j]))) {
+						/* method mi overwrites the (abstract) method    */
+#if defined(ENABLE_VERIFIER)
+						/* Add loading constraints (for the more general */
+						/* types of the method ic->methods[j]).          */
+						if (!classcache_add_constraints_for_params(
+									c->classloader, ic->classloader,
+									&(ic->methods[j])))
+						{
+							return false;
+						}
+#endif
 						v->interfacetable[-i][j] = v->table[mi->vftblindex];
 						goto foundmethod;
 					}
@@ -1089,8 +1208,14 @@ static void linker_addinterface(classinfo *c, classinfo *ic)
 		}
 	}
 
-	for (j = 0; j < ic->interfacescount; j++) 
-		linker_addinterface(c, ic->interfaces[j].cls);
+	/* add superinterfaces of this interface */
+
+	for (j = 0; j < ic->interfacescount; j++)
+		if (!linker_addinterface(c, ic->interfaces[j].cls))
+			return false;
+
+	/* everything ok */
+	return true;
 }
 
 
@@ -1133,4 +1258,5 @@ static s4 class_highestinterface(classinfo *c)
  * c-basic-offset: 4
  * tab-width: 4
  * End:
+ * vim:noexpandtab:sw=4:ts=4:
  */
