@@ -1,9 +1,9 @@
 /* src/vm/statistics.c - global varables for statistics
 
-   Copyright (C) 1996-2005 R. Grafl, A. Krall, C. Kruegel, C. Oates,
-   R. Obermaisser, M. Platter, M. Probst, S. Ring, E. Steiner,
-   C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich, J. Wenninger,
-   Institut f. Computersprachen - TU Wien
+   Copyright (C) 1996-2005, 2006 R. Grafl, A. Krall, C. Kruegel,
+   C. Oates, R. Obermaisser, M. Platter, M. Probst, S. Ring,
+   E. Steiner, C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich,
+   J. Wenninger, Institut f. Computersprachen - TU Wien
 
    This file is part of CACAO.
 
@@ -19,29 +19,35 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-   02111-1307, USA.
+   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
+   02110-1301, USA.
 
-   Contact: cacao@complang.tuwien.ac.at
+   Contact: cacao@cacaojvm.org
 
    Authors: Christian Thalinger
 
-   $Id: statistics.c,v 1.16 2005/10/28 11:34:23 twisti Exp $
+   Changes:
+
+   $Id: statistics.c 4397 2006-01-31 23:29:59Z twisti $
 
 */
 
 
+#include "config.h"
+
 #include <string.h> 
 #include <sys/time.h>
 #include <sys/resource.h>
 
+#include "vm/types.h"
+
 #include "toolbox/logging.h"
 #include "vm/global.h"
 #include "vm/options.h"
 #include "vm/statistics.h"
 
 
-/* global variables */
+/* global variables ***********************************************************/
 
 static s8 loadingtime = 0;              /* accumulated loading time           */
 static s8 loadingstarttime = 0;
@@ -136,29 +142,11 @@ int count_cstub_len = 0;
 int count_nstub_len = 0;
 int count_max_new_stack = 0;
 int count_upper_bound_new_stack = 0;
-                                /* in_  inline statistics */
-int count_in = 0;
-int count_in_uniqVirt = 0;
-int count_in_uniqIntf = 0;
-int count_in_rejected = 0;
-int count_in_rejected_mult = 0;
-int count_in_outsiders = 0;
-int count_in_uniqueVirt_not_inlined = 0;
-int count_in_uniqueInterface_not_inlined = 0;
-int count_in_maxDepth = 0;
-int count_in_maxMethods = 0;
 
 u8 count_native_function_calls=0;
-u8 count_compiled_function_calls=0;
 u8 count_jni_callXmethod_calls=0;
 u8 count_jni_calls=0;
 
-u2 count_in_not   [512];
-/***
-int count_no_in[12] = {0,0,0,0, 0,0,0,0, 0,0,0,0};
-***/
-
-
 
 static int count_block_stack_init[11] = {
 	0, 0, 0, 0, 0, 
@@ -207,45 +195,41 @@ s4 count_schedule_max_leaders = 0;
 s4 count_schedule_critical_path = 0;
 
 
-
 /* nativeinvokation ***********************************************************
 
    increments the native invokation count by one
 	
 *******************************************************************************/
-void nativeinvokation() {
-	/*do locking here */
+
+void nativeinvokation(void)
+{
+	/* XXX do locking here */
 	count_native_function_calls++;
 }
 
-/* compiledinvokation *********************************************************
-
-   increments the compiled invokation count by one
-	
-*******************************************************************************/
-void compiledinvokation() {
-	/* do locking here*/
-	count_compiled_function_calls++;
-}
-
 
 /* jnicallXmethodinvokation ***************************************************
 
    increments the jni CallXMethod invokation count by one
 	
 *******************************************************************************/
-void jnicallXmethodnvokation() {
-	/*do locking here */
+
+void jnicallXmethodnvokation(void)
+{
+	/* XXX do locking here */
 	count_jni_callXmethod_calls++;
 }
 
+
 /* jniinvokation *************************************************************
 
    increments the jni overall  invokation count by one
 	
 *******************************************************************************/
-void jniinvokation() {
-	/*do locking here */
+
+void jniinvokation(void)
+{
+	/* XXX do locking here */
 	count_jni_calls++;
 }
 
@@ -369,12 +353,14 @@ void print_times(void)
 
 void print_stats(void)
 {
+	s4    i;
+	float f;
+	s4    sum;
+
+
 	dolog("Number of JIT compiler calls: %6d", count_jit_calls);
 	dolog("Number of compiled methods:   %6d", count_methods);
 
-	if (opt_rt)
-		dolog("Number of Methods marked Used: %d", count_methods_marked_used);
-
 	dolog("Number of compiled basic blocks:               %6d",
 		  count_basic_blocks);
 	dolog("Number of max. basic blocks per method:        %6d",
@@ -451,26 +437,67 @@ void print_stats(void)
 		  count_analyse_iterations[0], count_analyse_iterations[1],
 		  count_analyse_iterations[2], count_analyse_iterations[3],
 		  count_analyse_iterations[4]);
-	dolog("Distribution of basic blocks per method");
-	dolog("   <=5  <=10  <=15  <=20  <=30  <=40  <=50  <=75   >75");
-	dolog("%6d%6d%6d%6d%6d%6d%6d%6d%6d",
-		  count_method_bb_distribution[0], count_method_bb_distribution[1],
-		  count_method_bb_distribution[2], count_method_bb_distribution[3],
-		  count_method_bb_distribution[4], count_method_bb_distribution[5],
-		  count_method_bb_distribution[6], count_method_bb_distribution[7],
-		  count_method_bb_distribution[8]);
-	dolog("Distribution of basic block sizes");
-	dolog("     0     1     2     3     4    5    6    7    8    9  <13  <15  <17  <19  <21  <26  <31  >30");
-	dolog("%6d%6d%6d%6d%6d%5d%5d%5d%5d%5d%5d%5d%5d%5d%5d%5d%5d%5d",
-		  count_block_size_distribution[0], count_block_size_distribution[1],
-		  count_block_size_distribution[2], count_block_size_distribution[3],
-		  count_block_size_distribution[4], count_block_size_distribution[5],
-		  count_block_size_distribution[6], count_block_size_distribution[7],
-		  count_block_size_distribution[8], count_block_size_distribution[9],
-		  count_block_size_distribution[10], count_block_size_distribution[11],
-		  count_block_size_distribution[12], count_block_size_distribution[13],
-		  count_block_size_distribution[14], count_block_size_distribution[15],
-		  count_block_size_distribution[16], count_block_size_distribution[17]);
+
+
+	/* Distribution of basic blocks per method ********************************/
+
+	log_println("Distribution of basic blocks per method:");
+	log_println("   <=5  <=10  <=15  <=20  <=30  <=40  <=50  <=75   >75");
+
+	log_start();
+	for (i = 0; i <= 8; i++)
+		log_print("%6d", count_method_bb_distribution[i]);
+	log_finish();
+
+	/* print ratio */
+
+	f = (float) count_methods;
+
+	log_start();
+	for (i = 0; i <= 8; i++)
+		log_print("%6.2f", (float) count_method_bb_distribution[i] / f);
+	log_finish();
+
+	/* print cumulated ratio */
+
+	log_start();
+	for (i = 0, sum = 0; i <= 8; i++) {
+		sum += count_method_bb_distribution[i];
+		log_print("%6.2f", (float) sum / f);
+	}
+	log_finish();
+
+
+	/* Distribution of basic block sizes **************************************/
+
+	log_println("Distribution of basic block sizes:");
+	log_println("     0     1     2     3     4     5     6     7     8     9   <13   <15   <17   <19   <21   <26   <31   >30");
+
+	/* print block sizes */
+
+	log_start();
+	for (i = 0; i <= 17; i++)
+		log_print("%6d", count_block_size_distribution[i]);
+	log_finish();
+
+	/* print ratio */
+
+	f = (float) count_basic_blocks;
+
+	log_start();
+	for (i = 0; i <= 17; i++)
+		log_print("%6.2f", (float) count_block_size_distribution[i] / f);
+	log_finish();
+
+	/* print cumulated ratio */
+
+	log_start();
+	for (i = 0, sum = 0; i <= 17; i++) {
+		sum += count_block_size_distribution[i];
+		log_print("%6.2f", (float) sum / f);
+	}
+	log_finish();
+
 	dolog("Size of Code Area:        %10.3f kB", (float) count_code_len / 1024);
 	dolog("Size of Data Area:        %10.3f kB", (float) count_data_len / 1024);
 	dolog("Size of Class Infos:      %10.3f kB", (float) count_class_infos / 1024);
@@ -517,116 +544,35 @@ void print_stats(void)
 	dolog("Number of Methods kept in registers:         %6d\n",
 		  count_method_in_register);
 
-		
-	/****if (useinlining)  ***/
-	 {
-		u2 i;
-		char * in_not_reasons[IN_MAX] = {
-        					"unqVirt    | ",
-					        "unqIntf    | ",
-					        "outsider   | ",
-					        "maxDepth   | ",
-					        "maxcode    | ",
-					        "maxlen     | ",
-					        "exception  | ",
-					        "notUnqVirt | ",
-					        "notUnqIntf |"
-					        };
 
-		dolog("Number of Methods Inlined :                              \t%6d",
-			  count_in);
-
-		if (inlinevirtuals) {
-			dolog("Number of Unique Virtual Methods inlined:                \t%6d",
-				  count_in_uniqVirt);
-			dolog("Number of Unique Implemented Interface Methods inlined:    %6d",
-				  count_in_uniqIntf);
-		}
-
-		dolog("Number of Methods Inlines (total) rejected:              \t%6d",
-			  count_in_rejected);
-		dolog("Number of Methods Inlined rejected for multiple reasons: \t%6d",
-			  count_in_rejected_mult);
-		dolog("Number of Methods where Inline max depth hit:            \t%6d",
-			  count_in_maxDepth);
-		dolog("Number of Methods Inlined rejected for max methods       \t%6d\n",
-			  count_in_maxMethods);
-		dolog("Number of Methods calls fom Outsider Class not inlined:  \t%6d",
-			  count_in_outsiders);
-
-		dolog("Number of Unique Virtual Methods not inlined:            \t%6d",
-			  count_in_uniqueVirt_not_inlined);
-		dolog("Number of Unique Implemented Interface Methods not inlined:%6d\n",
-			  count_in_uniqueInterface_not_inlined);
-
-#define INLINEDETAILS
-#ifdef  INLINEDETAILS
-		dolog("Details about Not Inlined Reasons:");
-
-		for (i = 0; i < 512; i++) {
-			if (count_in_not[i] > 0) {
-				/* XXX Please reimplement me! I'm ugly and insecure! */
-				char logtext2[1024]="\t";
-
-				if (inlinevirtuals) {
-					if (i & IN_UNIQUEVIRT)
-						strcat(logtext2, in_not_reasons[N_UNIQUEVIRT]);
-
-					if (i & IN_UNIQUE_INTERFACE)
-						strcat(logtext2, in_not_reasons[N_UNIQUE_INTERFACE]);
-
-					if (i & IN_NOT_UNIQUE_VIRT)
-						strcat(logtext2, in_not_reasons[N_NOT_UNIQUE_VIRT]);
-
-					if (i & IN_NOT_UNIQUE_INTERFACE)
-						strcat(logtext2, in_not_reasons[N_NOT_UNIQUE_INTERFACE]);
-				}
-
-				if (i & IN_OUTSIDERS)
-					strcat(logtext2, in_not_reasons[N_OUTSIDERS]);
-
-				if (i & IN_MAXDEPTH)
-					strcat(logtext2, in_not_reasons[N_MAXDEPTH]);
-
-				if (i & IN_MAXCODE)
-					strcat(logtext2, in_not_reasons[N_MAXCODE]);
-
-				if (i & IN_JCODELENGTH)
-					strcat(logtext2, in_not_reasons[N_JCODELENGTH]);
-
-				if (i & IN_EXCEPTION)
-					strcat(logtext2, in_not_reasons[N_EXCEPTION]);
-
-				dolog("  [%X]=%6d    %s", i, count_in_not[i], logtext2);
-			}
-		}
-#endif
-	 }
-
-
-	 /* instruction scheduler statistics **************************************/
+	/* instruction scheduler statistics ***************************************/
 
 #if defined(USE_SCHEDULER)
-	 dolog("Instruction scheduler statistics:");
-	 dolog("Number of basic blocks:       %7d", count_schedule_basic_blocks);
-	 dolog("Number of nodes:              %7d", count_schedule_nodes);
-	 dolog("Number of leaders nodes:      %7d", count_schedule_leaders);
-	 dolog("Number of max. leaders nodes: %7d", count_schedule_max_leaders);
-	 dolog("Length of critical path:      %7d\n", count_schedule_critical_path);
+	dolog("Instruction scheduler statistics:");
+	dolog("Number of basic blocks:       %7d", count_schedule_basic_blocks);
+	dolog("Number of nodes:              %7d", count_schedule_nodes);
+	dolog("Number of leaders nodes:      %7d", count_schedule_leaders);
+	dolog("Number of max. leaders nodes: %7d", count_schedule_max_leaders);
+	dolog("Length of critical path:      %7d\n", count_schedule_critical_path);
 #endif
 
 
 	/* call statistics ********************************************************/
 
-	 dolog("Function call statistics:");
-	 dolog("Number of native function invokations:           %ld",
-		   count_native_function_calls);
-	 dolog("Number of compiled function invokations:         %ld",
-		   count_compiled_function_calls);
-	 dolog("Number of jni->CallXMethod function invokations: %ld",
-		   count_jni_callXmethod_calls);
-	 dolog("Overall number of jni invokations:               %ld",
-		   count_jni_calls);
+	dolog("Function call statistics:");
+	dolog("Number of native function invokations:           %ld",
+		  count_native_function_calls);
+	dolog("Number of jni->CallXMethod function invokations: %ld",
+		  count_jni_callXmethod_calls);
+	dolog("Overall number of jni invokations:               %ld",
+		  count_jni_calls);
+
+
+	/* now print other statistics ********************************************/
+
+#if defined(ENABLE_INTRP)
+	print_dynamic_super_statistics();
+#endif
 }
 
 
