@@ -1,9 +1,9 @@
 /* src/vm/string.c - java.lang.String related functions
 
-   Copyright (C) 1996-2005 R. Grafl, A. Krall, C. Kruegel, C. Oates,
-   R. Obermaisser, M. Platter, M. Probst, S. Ring, E. Steiner,
-   C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich, J. Wenninger,
-   Institut f. Computersprachen - TU Wien
+   Copyright (C) 1996-2005, 2006 R. Grafl, A. Krall, C. Kruegel,
+   C. Oates, R. Obermaisser, M. Platter, M. Probst, S. Ring,
+   E. Steiner, C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich,
+   J. Wenninger, Institut f. Computersprachen - TU Wien
 
    This file is part of CACAO.
 
@@ -19,25 +19,27 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-   02111-1307, USA.
+   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
+   02110-1301, USA.
 
-   Contact: cacao@complang.tuwien.ac.at
+   Contact: cacao@cacaojvm.org
 
    Authors: Reinhard Grafl
             Roman Obermaisser
             Andreas Krall
 
    Changes: Christian Thalinger
+   			Edwin Steiner
 
-   $Id: string.c,v 1.12 2005/11/17 13:47:58 twisti Exp $
+   $Id: string.c 4921 2006-05-15 14:24:36Z twisti $
 
 */
 
 
+#include "config.h"
+
 #include <assert.h>
 
-#include "config.h"
 #include "vm/types.h"
 
 #include "vm/global.h"
@@ -53,8 +55,14 @@
 
 /* global variables ***********************************************************/
 
-#if defined(USE_THREADS)
-static java_objectheader *lock_string_hashtable;
+/* hashsize must be power of 2 */
+
+#define HASHTABLE_STRING_SIZE    2048   /* initial size of javastring-hash    */
+
+hashtable hashtable_string;             /* hashtable for javastrings          */
+
+#if defined(ENABLE_THREADS)
+static java_objectheader *lock_hashtable_string;
 #endif
 
 
@@ -125,6 +133,9 @@ const char *string_java_lang_NoSuchMethodException =
 const char *string_java_lang_NullPointerException =
     "java/lang/NullPointerException";
 
+const char *string_java_lang_StringIndexOutOfBoundsException =
+    "java/lang/StringIndexOutOfBoundsException";
+
 const char *string_java_lang_reflect_InvocationTargetException =
     "java/lang/reflect/InvocationTargetException";
 
@@ -191,14 +202,16 @@ const char *string_java_lang_VirtualMachineError =
 
 bool string_init(void)
 {
-#if defined(USE_THREADS)
+	/* create string (javastring) hashtable */
+
+	hashtable_create(&hashtable_string, HASHTABLE_STRING_SIZE);
+
+#if defined(ENABLE_THREADS)
 	/* create string hashtable lock object */
 
-	lock_string_hashtable = NEW(java_objectheader);
+	lock_hashtable_string = NEW(java_objectheader);
 
-# if defined(NATIVE_THREADS)
-	initObjectLock(lock_string_hashtable);
-# endif
+	lock_init_object_lock(lock_hashtable_string);
 #endif
 
 	/* everything's ok */
@@ -222,8 +235,8 @@ void stringtable_update(void)
 	literalstring *s;       /* hashtable entry */
 	int i;
 
-	for (i = 0; i < string_hash.size; i++) {
-		s = string_hash.ptr[i];
+	for (i = 0; i < hashtable_string.size; i++) {
+		s = hashtable_string.ptr[i];
 		if (s) {
 			while (s) {
                                                                
@@ -253,6 +266,77 @@ void stringtable_update(void)
 }
 
 
+/* javastring_new_from_utf_buffer **********************************************
+
+   Create a new object of type java/lang/String with the text from
+   the specified utf8 buffer.
+
+   IN:
+      buffer.......points to first char in the buffer
+	  blength......number of bytes to read from the buffer
+
+   RETURN VALUE:
+      the java.lang.String object, or
+      NULL if an exception has been thrown
+
+*******************************************************************************/
+
+java_lang_String *javastring_new_from_utf_buffer(const char *buffer, u4 blength)
+{
+	const char *utf_ptr;            /* current utf character in utf string    */
+	u4 utflength;                   /* length of utf-string if uncompressed   */
+	java_lang_String *s;            /* result-string                          */
+	java_chararray *a;
+	u4 i;
+
+	assert(buffer);
+
+	utflength = utf_get_number_of_u2s_for_buffer(buffer,blength);
+
+	s = (java_lang_String *) builtin_new(class_java_lang_String);
+	a = builtin_newarray_char(utflength);
+
+	/* javastring or character-array could not be created */
+	if (!a || !s)
+		return NULL;
+
+	/* decompress utf-string */
+	utf_ptr = buffer;
+	for (i = 0; i < utflength; i++)
+		a->data[i] = utf_nextu2((char **)&utf_ptr);
+	
+	/* set fields of the javastring-object */
+	s->value  = a;
+	s->offset = 0;
+	s->count  = utflength;
+
+	return s;
+}
+
+
+/* javastring_new_from_utf_string **********************************************
+
+   Create a new object of type java/lang/String with the text from
+   the specified zero-terminated utf8 string.
+
+   IN:
+      buffer.......points to first char in the buffer
+	  blength......number of bytes to read from the buffer
+
+   RETURN VALUE:
+      the java.lang.String object, or
+      NULL if an exception has been thrown
+
+*******************************************************************************/
+
+java_lang_String *javastring_new_from_utf_string(const char *utfstr)
+{
+	assert(utfstr);
+
+	return javastring_new_from_utf_buffer(utfstr, strlen(utfstr));
+}
+
+
 /* javastring_new **************************************************************
 
    creates a new object of type java/lang/String with the text of 
@@ -271,12 +355,12 @@ java_lang_String *javastring_new(utf *u)
 	s4 i;
 
 	if (!u) {
-		*exceptionptr = new_nullpointerexception();
+		exceptions_throw_nullpointerexception();
 		return NULL;
 	}
 
 	utf_ptr = u->text;
-	utflength = utf_strlen(u);
+	utflength = utf_get_number_of_u2s(u);
 
 	s = (java_lang_String *) builtin_new(class_java_lang_String);
 	a = builtin_newarray_char(utflength);
@@ -316,12 +400,12 @@ java_lang_String *javastring_new_slash_to_dot(utf *u)
 	u2 ch;
 
 	if (!u) {
-		*exceptionptr = new_nullpointerexception();
+		exceptions_throw_nullpointerexception();
 		return NULL;
 	}
 
 	utf_ptr = u->text;
-	utflength = utf_strlen(u);
+	utflength = utf_get_number_of_u2s(u);
 
 	s = (java_lang_String *) builtin_new(class_java_lang_String);
 	a = builtin_newarray_char(utflength);
@@ -347,16 +431,21 @@ java_lang_String *javastring_new_slash_to_dot(utf *u)
 }
 
 
-/* javastring_new_char *********************************************************
+/* javastring_new_from_ascii ***************************************************
 
-   creates a new java/lang/String object which contains the convertet
-   C-string passed via text.
+   creates a new java/lang/String object which contains the given ASCII
+   C-string converted to UTF-16.
 
-   return: the object pointer or NULL if memory is exhausted.
+   IN:
+      text.........string of ASCII characters
+
+   RETURN VALUE:
+      the java.lang.String object, or 
+      NULL if an exception has been thrown.
 
 *******************************************************************************/
 
-java_lang_String *javastring_new_char(const char *text)
+java_lang_String *javastring_new_from_ascii(const char *text)
 {
 	s4 i;
 	s4 len;                             /* length of the string               */
@@ -364,7 +453,7 @@ java_lang_String *javastring_new_char(const char *text)
 	java_chararray *a;
 
 	if (!text) {
-		*exceptionptr = new_nullpointerexception();
+		exceptions_throw_nullpointerexception();
 		return NULL;
 	}
 
@@ -446,10 +535,8 @@ utf *javastring_toutf(java_lang_String *string, bool isclassname)
 	
 *******************************************************************************/
 
-s4 javastring_strlen(java_objectheader *so)
+s4 javastring_strlen(java_lang_String *s)
 {
-	java_lang_String *s = (java_lang_String *) so;
-	
 	if (!s)
 		return 0;
 
@@ -476,15 +563,15 @@ java_objectheader *literalstring_u2(java_chararray *a, u4 length, u4 offset,
     u4                slot;
     u2                i;
 
-#if defined(USE_THREADS)
-	builtin_monitorenter(lock_string_hashtable);
+#if defined(ENABLE_THREADS)
+	builtin_monitorenter(lock_hashtable_string);
 #endif
 
     /* find location in hashtable */
 
     key  = unicode_hashkey(a->data + offset, length);
-    slot = key & (string_hash.size - 1);
-    s    = string_hash.ptr[slot];
+    slot = key & (hashtable_string.size - 1);
+    s    = hashtable_string.ptr[slot];
 
     while (s) {
 		js = (java_lang_String *) s->string;
@@ -501,8 +588,8 @@ java_objectheader *literalstring_u2(java_chararray *a, u4 length, u4 offset,
 			if (!copymode)
 				mem_free(a, sizeof(java_chararray) + sizeof(u2) * (length - 1) + 10);
 
-#if defined(USE_THREADS)
-			builtin_monitorexit(lock_string_hashtable);
+#if defined(ENABLE_THREADS)
+			builtin_monitorexit(lock_hashtable_string);
 #endif
 
 			return (java_objectheader *) js;
@@ -531,12 +618,8 @@ java_objectheader *literalstring_u2(java_chararray *a, u4 length, u4 offset,
 		primitivetype_table[ARRAYTYPE_CHAR].arrayvftbl;
     stringdata->header.size = length;
 
-	/* XXX TWISTI: is this necessary? */
-	if (!class_java_lang_String)
-		class_java_lang_String = load_class_bootstrap(utf_java_lang_String);
-
 	assert(class_java_lang_String);
-	assert(class_java_lang_String->loaded);
+	assert(class_java_lang_String->state & CLASS_LOADED);
 
 	/* if we use eager loading, we have to check loaded String class */
 
@@ -547,8 +630,8 @@ java_objectheader *literalstring_u2(java_chararray *a, u4 length, u4 offset,
 
 	js = NEW(java_lang_String);
 
-#if defined(USE_THREADS) && defined(NATIVE_THREADS)
-	initObjectLock(&js->header);
+#if defined(ENABLE_THREADS)
+	lock_init_object_lock(&js->header);
 #endif
 
 	js->header.vftbl = class_java_lang_String->vftbl;
@@ -559,17 +642,17 @@ java_objectheader *literalstring_u2(java_chararray *a, u4 length, u4 offset,
 	/* create new literalstring */
 
 	s = NEW(literalstring);
-	s->hashlink = string_hash.ptr[slot];
+	s->hashlink = hashtable_string.ptr[slot];
 	s->string   = (java_objectheader *) js;
-	string_hash.ptr[slot] = s;
+	hashtable_string.ptr[slot] = s;
 
 	/* update number of hashtable entries */
 
-	string_hash.entries++;
+	hashtable_string.entries++;
 
 	/* reorganization of hashtable */       
 
-	if (string_hash.entries > (string_hash.size * 2)) {
+	if (hashtable_string.entries > (hashtable_string.size * 2)) {
 		/* reorganization of hashtable, average length of the external
 		   chains is approx. 2 */
 
@@ -577,17 +660,17 @@ java_objectheader *literalstring_u2(java_chararray *a, u4 length, u4 offset,
 		literalstring    *s;
 		literalstring    *nexts;
 		java_lang_String *tmpjs;
-		hashtable         newhash; /* the new hashtable */
+		hashtable         newhash;                       /* the new hashtable */
       
 		/* create new hashtable, double the size */
 
-		init_hashtable(&newhash, string_hash.size * 2);
-		newhash.entries = string_hash.entries;
+		hashtable_create(&newhash, hashtable_string.size * 2);
+		newhash.entries = hashtable_string.entries;
       
 		/* transfer elements to new hashtable */
 
-		for (i = 0; i < string_hash.size; i++) {
-			s = string_hash.ptr[i];
+		for (i = 0; i < hashtable_string.size; i++) {
+			s = hashtable_string.ptr[i];
 
 			while (s) {
 				nexts = s->hashlink;
@@ -604,12 +687,12 @@ java_objectheader *literalstring_u2(java_chararray *a, u4 length, u4 offset,
 	
 		/* dispose old table */
 
-		MFREE(string_hash.ptr, void*, string_hash.size);
-		string_hash = newhash;
+		MFREE(hashtable_string.ptr, void*, hashtable_string.size);
+		hashtable_string = newhash;
 	}
 
-#if defined(USE_THREADS)
-	builtin_monitorexit(lock_string_hashtable);
+#if defined(ENABLE_THREADS)
+	builtin_monitorexit(lock_hashtable_string);
 #endif
 
 	return (java_objectheader *) js;
@@ -632,7 +715,7 @@ java_objectheader *literalstring_new(utf *u)
     u4              i;
 
 	utf_ptr = u->text;
-	utflength = utf_strlen(u);
+	utflength = utf_get_number_of_u2s(u);
 
     /* allocate memory */ 
     a = mem_alloc(sizeof(java_chararray) + sizeof(u2) * (utflength - 1) + 10);
@@ -678,4 +761,5 @@ void literalstring_free(java_objectheader* sobj)
  * c-basic-offset: 4
  * tab-width: 4
  * End:
+ * vim:noexpandtab:sw=4:ts=4:
  */
