@@ -1,9 +1,9 @@
-/* src/vm/utf.c - utf functions
+/* src/vm/utf8.c - utf8 string functions
 
-   Copyright (C) 1996-2005 R. Grafl, A. Krall, C. Kruegel, C. Oates,
-   R. Obermaisser, M. Platter, M. Probst, S. Ring, E. Steiner,
-   C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich, J. Wenninger,
-   Institut f. Computersprachen - TU Wien
+   Copyright (C) 1996-2005, 2006 R. Grafl, A. Krall, C. Kruegel,
+   C. Oates, R. Obermaisser, M. Platter, M. Probst, S. Ring,
+   E. Steiner, C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich,
+   J. Wenninger, Institut f. Computersprachen - TU Wien
 
    This file is part of CACAO.
 
@@ -19,45 +19,57 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-   02111-1307, USA.
+   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
+   02110-1301, USA.
 
-   Contact: cacao@complang.tuwien.ac.at
+   Contact: cacao@cacaojvm.org
 
    Authors: Reinhard Grafl
 
    Changes: Mark Probst
             Andreas Krall
             Christian Thalinger
+			Edwin Steiner
 
-   $Id: utf8.c,v 1.27 2005/11/16 19:13:37 edwin Exp $
+   $Id: utf8.c 4921 2006-05-15 14:24:36Z twisti $
 
 */
 
 
+#include "config.h"
+
 #include <string.h>
 #include <assert.h>
 
+#include "vm/types.h"
+
 #include "mm/memory.h"
+
+#if defined(ENABLE_THREADS)
+# include "threads/native/threads.h"
+#endif
+
+#include "vm/builtin.h"
 #include "vm/exceptions.h"
+#include "vm/hashtable.h"
 #include "vm/options.h"
 #include "vm/statistics.h"
 #include "vm/stringlocal.h"
-#include "vm/tables.h"
 #include "vm/utf8.h"
 
+
 /* global variables ***********************************************************/
 
-#if defined(USE_THREADS)
-static java_objectheader *lock_utf_hashtable;
-#endif
+/* hashsize must be power of 2 */
 
-hashtable utf_hash;                     /* hashtable for utf8-symbols         */
+#define HASHTABLE_UTF_SIZE    16384     /* initial size of utf-hash           */
+
+hashtable *hashtable_utf;               /* hashtable for utf8-symbols         */
 
 
 /* utf-symbols for pointer comparison of frequently used strings **************/
 
-utf *utf_java_lang_Object;              /* java/lang/Object                   */
+utf *utf_java_lang_Object;
 
 utf *utf_java_lang_Class;
 utf *utf_java_lang_ClassLoader;
@@ -72,6 +84,7 @@ utf *utf_java_lang_Throwable;
 utf *utf_java_lang_VMThrowable;
 utf *utf_java_lang_Error;
 utf *utf_java_lang_NoClassDefFoundError;
+utf *utf_java_lang_LinkageError;
 utf *utf_java_lang_NoSuchMethodError;
 utf *utf_java_lang_OutOfMemoryError;
 
@@ -162,14 +175,15 @@ utf *array_packagename;
 
 bool utf8_init(void)
 {
-#if defined(USE_THREADS)
-	/* create utf hashtable lock object */
+	/* create utf8 hashtable */
 
-	lock_utf_hashtable = NEW(java_objectheader);
+	hashtable_utf = NEW(hashtable);
 
-# if defined(NATIVE_THREADS)
-	initObjectLock(lock_utf_hashtable);
-# endif
+	hashtable_create(hashtable_utf, HASHTABLE_UTF_SIZE);
+
+#if defined(ENABLE_STATISTICS)
+	if (opt_stat)
+		count_utf_len += sizeof(utf*) * hashtable_utf.size;
 #endif
 
 	/* create utf-symbols for pointer comparison of frequently used strings */
@@ -192,6 +206,9 @@ bool utf8_init(void)
 	utf_java_lang_NoClassDefFoundError =
 		utf_new_char(string_java_lang_NoClassDefFoundError);
 
+	utf_java_lang_LinkageError =
+		utf_new_char(string_java_lang_LinkageError);
+
 	utf_java_lang_NoSuchMethodError =
 		utf_new_char(string_java_lang_NoSuchMethodError);
 
@@ -423,8 +440,36 @@ u4 utf_hashkey(const char *text, u4 length)
     }
 }
 
+/* utf_full_hashkey ************************************************************
 
-/* utf_hashkey *****************************************************************
+   This function computes a hash value using all bytes in the string.
+
+   The algorithm is the "One-at-a-time" algorithm as published
+   by Bob Jenkins on http://burtleburtle.net/bob/hash/doobs.html.
+
+*******************************************************************************/
+
+u4 utf_full_hashkey(const char *text, u4 length)
+{
+	register const unsigned char *p = (const unsigned char *) text;
+	register u4 hash;
+	register u4 i;
+
+	hash = 0;
+	for (i=length; i--;)
+	{
+	    hash += *p++;
+	    hash += (hash << 10);
+	    hash ^= (hash >> 6);
+	}
+	hash += (hash << 3);
+	hash ^= (hash >> 11);
+	hash += (hash << 15);
+
+	return hash;
+}
+
+/* unicode_hashkey *************************************************************
 
    Compute the hashkey of a unicode string.
 
@@ -456,27 +501,18 @@ utf *utf_new(const char *text, u2 length)
 	utf *u;                             /* hashtable element                  */
 	u2 i;
 
-	/* XXX REMOVE ME! after testing of course ;-) */
-	static int running = 0;
-	/* XXX REMOVE ME! */
-
-#if defined(USE_THREADS)
-	builtin_monitorenter(lock_utf_hashtable);
+#if defined(ENABLE_THREADS)
+	builtin_monitorenter(hashtable_utf->header);
 #endif
 
-	/* XXX REMOVE ME! after testing of course ;-) */
-	assert(running == 0);
-	running = 1;
-	/* XXX REMOVE ME! */
-
-#ifdef STATISTICS
+#if defined(ENABLE_STATISTICS)
 	if (opt_stat)
 		count_utf_new++;
 #endif
 
 	key  = utf_hashkey(text, length);
-	slot = key & (utf_hash.size - 1);
-	u    = utf_hash.ptr[slot];
+	slot = key & (hashtable_utf->size - 1);
+	u    = hashtable_utf->ptr[slot];
 
 	/* search external hash chain for utf-symbol */
 
@@ -488,19 +524,15 @@ utf *utf_new(const char *text, u2 length)
 				if (text[i] != u->text[i])
 					goto nomatch;
 			
-#if defined(STATISTICS)
+#if defined(ENABLE_STATISTICS)
 			if (opt_stat)
 				count_utf_new_found++;
 #endif
 
 			/* symbol found in hashtable */
 
-			/* XXX REMOVE ME! */
-			running = 0;
-			/* XXX REMOVE ME! */
-
-#if defined(USE_THREADS)
-			builtin_monitorexit(lock_utf_hashtable);
+#if defined(ENABLE_THREADS)
+			builtin_monitorexit(hashtable_utf->header);
 #endif
 
 			return u;
@@ -510,7 +542,7 @@ utf *utf_new(const char *text, u2 length)
 		u = u->hashlink; /* next element in external chain */
 	}
 
-#if defined(STATISTICS)
+#if defined(ENABLE_STATISTICS)
 	if (opt_stat)
 		count_utf_len += sizeof(utf) + length + 1;
 #endif
@@ -518,58 +550,62 @@ utf *utf_new(const char *text, u2 length)
 	/* location in hashtable found, create new utf element */
 	u = NEW(utf);
 	u->blength  = length;               /* length in bytes of utfstring       */
-	u->hashlink = utf_hash.ptr[slot];   /* link in external hashchain         */
+	u->hashlink = hashtable_utf->ptr[slot]; /* link in external hashchain     */
 	u->text     = mem_alloc(length + 1);/* allocate memory for utf-text       */
+
 	memcpy(u->text, text, length);      /* copy utf-text                      */
 	u->text[length] = '\0';
-	utf_hash.ptr[slot] = u;             /* insert symbol into table           */
 
-	utf_hash.entries++;                 /* update number of entries           */
+	hashtable_utf->ptr[slot] = u;       /* insert symbol into table           */
+	hashtable_utf->entries++;           /* update number of entries           */
 
-	if (utf_hash.entries > (utf_hash.size * 2)) {
+	if (hashtable_utf->entries > (hashtable_utf->size * 2)) {
 
-        /* reorganization of hashtable, average length of 
-           the external chains is approx. 2                */  
+        /* reorganization of hashtable, average length of the external
+           chains is approx. 2 */
 
-		u4 i;
-		utf *u;
-		hashtable newhash; /* the new hashtable */
+		hashtable *newhash;                              /* the new hashtable */
+		u4         i;
+		utf       *u;
+		utf       *nextu;
+		u4         slot;
 
 		/* create new hashtable, double the size */
-		init_hashtable(&newhash, utf_hash.size * 2);
-		newhash.entries = utf_hash.entries;
 
-#ifdef STATISTICS
+		newhash = hashtable_resize(hashtable_utf, hashtable_utf->size * 2);
+
+#if defined(ENABLE_STATISTICS)
 		if (opt_stat)
-			count_utf_len += sizeof(utf*) * utf_hash.size;
+			count_utf_len += sizeof(utf*) * hashtable_utf->size;
 #endif
 
 		/* transfer elements to new hashtable */
-		for (i = 0; i < utf_hash.size; i++) {
-			u = (utf *) utf_hash.ptr[i];
+
+		for (i = 0; i < hashtable_utf->size; i++) {
+			u = hashtable_utf->ptr[i];
+
 			while (u) {
-				utf *nextu = u->hashlink;
-				u4 slot = utf_hashkey(u->text, u->blength) & (newhash.size - 1);
+				nextu = u->hashlink;
+				slot  = utf_hashkey(u->text, u->blength) & (newhash->size - 1);
 						
-				u->hashlink = (utf *) newhash.ptr[slot];
-				newhash.ptr[slot] = u;
+				u->hashlink = (utf *) newhash->ptr[slot];
+				newhash->ptr[slot] = u;
 
 				/* follow link in external hash chain */
+
 				u = nextu;
 			}
 		}
 	
 		/* dispose old table */
-		MFREE(utf_hash.ptr, void*, utf_hash.size);
-		utf_hash = newhash;
+
+		hashtable_free(hashtable_utf);
+
+		hashtable_utf = newhash;
 	}
 
-	/* XXX REMOVE ME! */
-	running = 0;
-	/* XXX REMOVE ME! */
-
-#if defined(USE_THREADS)
-	builtin_monitorexit(lock_utf_hashtable);
+#if defined(ENABLE_THREADS)
+	builtin_monitorexit(hashtable_utf->header);
 #endif
 
 	return u;
@@ -741,20 +777,94 @@ u2 utf_nextu2(char **utf_ptr)
 }
 
 
-/* utf_strlen ******************************************************************
+/* utf_bytes *******************************************************************
 
-   Determine number of unicode characters in the utf string.
+   Determine number of bytes (aka. octets) in the utf string.
+
+   IN:
+      u............utf string
+
+   OUT:
+      The number of octets of this utf string.
+	  There is _no_ terminating zero included in this count.
 
 *******************************************************************************/
 
-u4 utf_strlen(utf *u)
+u4 utf_bytes(utf *u)
+{
+	return u->blength;
+}
+
+/* utf_get_number_of_u2s_for_buffer ********************************************
+
+   Determine number of UTF-16 u2s in the given UTF-8 buffer
+
+   CAUTION: Use this function *only* when you want to convert an UTF-8 buffer
+   to an array of u2s (UTF-16) and want to know how many of them you will get.
+   All other uses of this function are probably wrong.
+
+   IN:
+      buffer........points to first char in buffer
+	  blength.......number of _bytes_ in the buffer
+
+   OUT:
+      the number of u2s needed to hold this string in UTF-16 encoding.
+	  There is _no_ terminating zero included in this count.
+
+   NOTE: Unlike utf_get_number_of_u2s, this function never throws an
+   exception.
+
+*******************************************************************************/
+
+u4 utf_get_number_of_u2s_for_buffer(const char *buffer, u4 blength)
+{
+	const char *endpos;                 /* points behind utf string           */
+	const char *utf_ptr;                /* current position in utf text       */
+	u4 len = 0;                         /* number of unicode characters       */
+
+	utf_ptr = buffer;
+	endpos = utf_ptr + blength;
+
+	while (utf_ptr < endpos) {
+		len++;
+		/* next unicode character */
+		utf_nextu2((char **)&utf_ptr);
+	}
+
+	assert(utf_ptr == endpos);
+
+	return len;
+}
+
+
+/* utf_get_number_of_u2s *******************************************************
+
+   Determine number of UTF-16 u2s in the utf string.
+
+   CAUTION: Use this function *only* when you want to convert a utf string
+   to an array of u2s and want to know how many of them you will get.
+   All other uses of this function are probably wrong.
+
+   IN:
+      u............utf string
+
+   OUT:
+      the number of u2s needed to hold this string in UTF-16 encoding.
+	  There is _no_ terminating zero included in this count.
+	  XXX 0 if a NullPointerException has been thrown (see below)
+
+*******************************************************************************/
+
+u4 utf_get_number_of_u2s(utf *u)
 {
 	char *endpos;                       /* points behind utf string           */
 	char *utf_ptr;                      /* current position in utf text       */
 	u4 len = 0;                         /* number of unicode characters       */
 
+	/* XXX this is probably not checked by most callers! Review this after */
+	/* the invalid uses of this function have been eliminated */
 	if (!u) {
-		*exceptionptr = new_nullpointerexception();
+		exceptions_throw_nullpointerexception();
 		return 0;
 	}
 
@@ -805,18 +915,100 @@ u4 u2_utflength(u2 *text, u4 u2_length)
 }
 
 
-/* utf_display *****************************************************************
+/* utf_copy ********************************************************************
 
-   Write utf symbol to stdout (for debugging purposes).
+   Copy the given utf string byte-for-byte to a buffer.
+
+   IN:
+      buffer.......the buffer
+	  u............the utf string
 
 *******************************************************************************/
 
-void utf_display(utf *u)
+void utf_copy(char *buffer, utf *u)
+{
+	/* our utf strings are zero-terminated (done by utf_new) */
+	MCOPY(buffer, u->text, char, u->blength + 1);
+}
+
+
+/* utf_cat *********************************************************************
+
+   Append the given utf string byte-for-byte to a buffer.
+
+   IN:
+      buffer.......the buffer
+	  u............the utf string
+
+*******************************************************************************/
+
+void utf_cat(char *buffer, utf *u)
+{
+	/* our utf strings are zero-terminated (done by utf_new) */
+	MCOPY(buffer + strlen(buffer), u->text, char, u->blength + 1);
+}
+
+
+/* utf_copy_classname **********************************************************
+
+   Copy the given utf classname byte-for-byte to a buffer.
+   '/' is replaced by '.'
+
+   IN:
+      buffer.......the buffer
+	  u............the utf string
+
+*******************************************************************************/
+
+void utf_copy_classname(char *buffer, utf *u)
+{
+	char *bufptr;
+	char *srcptr;
+	char *endptr;
+	char ch;
+
+	bufptr = buffer;
+	srcptr = u->text;
+	endptr = UTF_END(u) + 1; /* utfs are zero-terminared by utf_new */
+
+	while (srcptr != endptr) {
+		ch = *srcptr++;
+		if (ch == '/')
+			ch = '.';
+		*bufptr++ = ch;
+	}
+}
+
+
+/* utf_cat *********************************************************************
+
+   Append the given utf classname byte-for-byte to a buffer.
+   '/' is replaced by '.'
+
+   IN:
+      buffer.......the buffer
+	  u............the utf string
+
+*******************************************************************************/
+
+void utf_cat_classname(char *buffer, utf *u)
+{
+	utf_copy_classname(buffer + strlen(buffer), u);
+}
+
+/* utf_display_printable_ascii *************************************************
+
+   Write utf symbol to stdout (for debugging purposes).
+   Non-printable and non-ASCII characters are printed as '?'.
+
+*******************************************************************************/
+
+void utf_display_printable_ascii(utf *u)
 {
 	char *endpos;                       /* points behind utf string           */
 	char *utf_ptr;                      /* current position in utf text       */
 
-	if (!u) {
+	if (u == NULL) {
 		printf("NULL");
 		fflush(stdout);
 		return;
@@ -826,29 +1018,34 @@ void utf_display(utf *u)
 	utf_ptr = u->text;
 
 	while (utf_ptr < endpos) {
-		/* read next unicode character */                
+		/* read next unicode character */
+
 		u2 c = utf_nextu2(&utf_ptr);
-		if (c >= 32 && c <= 127) printf("%c", c);
-		else printf("?");
+
+		if ((c >= 32) && (c <= 127))
+			printf("%c", c);
+		else
+			printf("?");
 	}
 
 	fflush(stdout);
 }
 
 
-/* utf_display_classname *******************************************************
+/* utf_display_printable_ascii_classname ***************************************
 
    Write utf symbol to stdout with `/' converted to `.' (for debugging
    purposes).
+   Non-printable and non-ASCII characters are printed as '?'.
 
 *******************************************************************************/
 
-void utf_display_classname(utf *u)
+void utf_display_printable_ascii_classname(utf *u)
 {
 	char *endpos;                       /* points behind utf string           */
 	char *utf_ptr;                      /* current position in utf text       */
 
-	if (!u) {
+	if (u == NULL) {
 		printf("NULL");
 		fflush(stdout);
 		return;
@@ -858,24 +1055,32 @@ void utf_display_classname(utf *u)
 	utf_ptr = u->text;
 
 	while (utf_ptr < endpos) {
-		/* read next unicode character */                
+		/* read next unicode character */
+
 		u2 c = utf_nextu2(&utf_ptr);
-		if (c == '/') c = '.';
-		if (c >= 32 && c <= 127) printf("%c", c);
-		else printf("?");
+
+		if (c == '/')
+			c = '.';
+
+		if ((c >= 32) && (c <= 127))
+			printf("%c", c);
+		else
+			printf("?");
 	}
 
 	fflush(stdout);
 }
 
 
-/* utf_sprint ******************************************************************
+/* utf_sprint_convert_to_latin1 ************************************************
 	
    Write utf symbol into c-string (for debugging purposes).
+   Characters are converted to 8-bit Latin-1, non-Latin-1 characters yield
+   invalid results.
 
 *******************************************************************************/
 
-void utf_sprint(char *buffer, utf *u)
+void utf_sprint_convert_to_latin1(char *buffer, utf *u)
 {
 	char *endpos;                       /* points behind utf string           */
 	char *utf_ptr;                      /* current position in utf text       */
@@ -898,14 +1103,16 @@ void utf_sprint(char *buffer, utf *u)
 }
 
 
-/* utf_sprint_classname ********************************************************
+/* utf_sprint_convert_to_latin1_classname **************************************
 	
    Write utf symbol into c-string with `/' converted to `.' (for debugging
    purposes).
+   Characters are converted to 8-bit Latin-1, non-Latin-1 characters yield
+   invalid results.
 
 *******************************************************************************/
 
-void utf_sprint_classname(char *buffer, utf *u)
+void utf_sprint_convert_to_latin1_classname(char *buffer, utf *u)
 {
 	char *endpos;                       /* points behind utf string           */
 	char *utf_ptr;                      /* current position in utf text       */
@@ -931,37 +1138,42 @@ void utf_sprint_classname(char *buffer, utf *u)
 }
 
 
-/* utf_strcat ******************************************************************
+/* utf_strcat_convert_to_latin1 ************************************************
 	
    Like libc strcat, but uses an utf8 string.
+   Characters are converted to 8-bit Latin-1, non-Latin-1 characters yield
+   invalid results.
 
 *******************************************************************************/
 
-void utf_strcat(char *buffer, utf *u)
+void utf_strcat_convert_to_latin1(char *buffer, utf *u)
 {
-	utf_sprint(buffer + strlen(buffer), u);
+	utf_sprint_convert_to_latin1(buffer + strlen(buffer), u);
 }
 
 
-/* utf_strcat_classname ********************************************************
+/* utf_strcat_convert_to_latin1_classname **************************************
 	
    Like libc strcat, but uses an utf8 string.
+   Characters are converted to 8-bit Latin-1, non-Latin-1 characters yield
+   invalid results.
 
 *******************************************************************************/
 
-void utf_strcat_classname(char *buffer, utf *u)
+void utf_strcat_convert_to_latin1_classname(char *buffer, utf *u)
 {
-	utf_sprint_classname(buffer + strlen(buffer), u);
+	utf_sprint_convert_to_latin1_classname(buffer + strlen(buffer), u);
 }
 
 
-/* utf_fprint ******************************************************************
+/* utf_fprint_printable_ascii **************************************************
 	
    Write utf symbol into file.
+   Non-printable and non-ASCII characters are printed as '?'.
 
 *******************************************************************************/
 
-void utf_fprint(FILE *file, utf *u)
+void utf_fprint_printable_ascii(FILE *file, utf *u)
 {
 	char *endpos;                       /* points behind utf string           */
 	char *utf_ptr;                      /* current position in utf text       */
@@ -982,13 +1194,14 @@ void utf_fprint(FILE *file, utf *u)
 }
 
 
-/* utf_fprint_classname ********************************************************
+/* utf_fprint_printable_ascii_classname ****************************************
 	
    Write utf symbol into file with `/' converted to `.'.
+   Non-printable and non-ASCII characters are printed as '?'.
 
 *******************************************************************************/
 
-void utf_fprint_classname(FILE *file, utf *u)
+void utf_fprint_printable_ascii_classname(FILE *file, utf *u)
 {
 	char *endpos;                       /* points behind utf string           */
 	char *utf_ptr;                      /* current position in utf text       */
@@ -1119,6 +1332,7 @@ bool is_valid_name_utf(utf *u)
 
 *******************************************************************************/
 
+#if !defined(NDEBUG)
 void utf_show(void)
 {
 
@@ -1130,29 +1344,30 @@ void utf_show(void)
 	u4 beyond_limit = 0;         /* number of utf-symbols in chains with length>=CHAIN_LIMIT-1 */
 	u4 i;
 
-	printf ("UTF-HASH:\n");
+	printf("UTF-HASH:\n");
 
 	/* show element of utf-hashtable */
-	for (i=0; i<utf_hash.size; i++) {
-		utf *u = utf_hash.ptr[i];
+
+	for (i = 0; i < hashtable_utf->size; i++) {
+		utf *u = hashtable_utf->ptr[i];
+
 		if (u) {
-			printf ("SLOT %d: ", (int) i);
+			printf("SLOT %d: ", (int) i);
+
 			while (u) {
-				printf ("'");
-				utf_display (u);
-				printf ("' ");
+				printf("'");
+				utf_display_printable_ascii(u);
+				printf("' ");
 				u = u->hashlink;
 			}	
-			printf ("\n");
+			printf("\n");
 		}
-		
 	}
 
-	printf ("UTF-HASH: %d slots for %d entries\n", 
-			(int) utf_hash.size, (int) utf_hash.entries );
+	printf("UTF-HASH: %d slots for %d entries\n", 
+		   (int) hashtable_utf->size, (int) hashtable_utf->entries );
 
-
-	if (utf_hash.entries == 0)
+	if (hashtable_utf->entries == 0)
 		return;
 
 	printf("chains:\n  chainlength    number of chains    %% of utfstrings\n");
@@ -1161,9 +1376,9 @@ void utf_show(void)
 		chain_count[i]=0;
 
 	/* count numbers of hashchains according to their length */
-	for (i=0; i<utf_hash.size; i++) {
+	for (i=0; i<hashtable_utf->size; i++) {
 		  
-		utf *u = (utf*) utf_hash.ptr[i];
+		utf *u = (utf*) hashtable_utf->ptr[i];
 		u4 chain_length = 0;
 
 		/* determine chainlength */
@@ -1191,16 +1406,17 @@ void utf_show(void)
 
 	/* display results */  
 	for (i=1;i<CHAIN_LIMIT-1;i++) 
-		printf("       %2d %17d %18.2f%%\n",i,chain_count[i],(((float) chain_count[i]*i*100)/utf_hash.entries));
+		printf("       %2d %17d %18.2f%%\n",i,chain_count[i],(((float) chain_count[i]*i*100)/hashtable_utf->entries));
 	  
-	printf("     >=%2d %17d %18.2f%%\n",CHAIN_LIMIT-1,chain_count[CHAIN_LIMIT-1],((float) beyond_limit*100)/utf_hash.entries);
+	printf("     >=%2d %17d %18.2f%%\n",CHAIN_LIMIT-1,chain_count[CHAIN_LIMIT-1],((float) beyond_limit*100)/hashtable_utf->entries);
 
 
 	printf("max. chainlength:%5d\n",max_chainlength);
 
 	/* avg. chainlength = sum of chainlengths / number of chains */
-	printf("avg. chainlength:%5.2f\n",(float) sum_chainlength / (utf_hash.size-chain_count[0]));
+	printf("avg. chainlength:%5.2f\n",(float) sum_chainlength / (hashtable_utf->size-chain_count[0]));
 }
+#endif /* !defined(NDEBUG) */
 
 
 /*
@@ -1214,4 +1430,5 @@ void utf_show(void)
  * c-basic-offset: 4
  * tab-width: 4
  * End:
+ * vim:noexpandtab:sw=4:ts=4:
  */
